/* Warning: Do not edit this file.
   Auto-generated by 'make_compiler_flags'.*/
const char* erts_build_flags_CFLAGS = "-Werror=undef -Werror=implicit -Werror=return-type  -fno-common -fstrict-flex-arrays=3 -fno-strict-aliasing -fno-delete-null-pointer-checks -fno-strict-overflow -fexceptions -fcf-protection=full -fstack-protector-strong -fstack-clash-protection -U_FORTIFY_SOURCE -D_FORTIFY_SOURCE=3 -march=x86-64 -mtune=generic -O2 -pipe -fno-plt -fexceptions -Wp,-D_FORTIFY_SOURCE=3 -Wformat -Werror=format-security -fstack-clash-protection -fcf-protection -fno-omit-frame-pointer -mno-omit-leaf-frame-pointer -g -ffile-prefix-map=/build/erlang/src=/usr/src/debug/erlang -flto=auto -ffat-lto-objects -I/build/erlang/src/otp/erts/x86_64-pc-linux-gnu   -D_GNU_SOURCE  -DHAVE_CONFIG_H -Wall -Wstrict-prototypes -Wpointer-arith -Wmissing-prototypes -Wdeclaration-after-statement -DUSE_THREADS -D_THREAD_SAFE -D_REENTRANT -DPOSIX_THREADS -D_POSIX_THREAD_SAFE_FUNCTIONS   -DBEAMASM=1";
const char* erts_build_flags_CONFIG_H = "/* x86_64-pc-linux-gnu/config.h.  Generated from config.h.in by configure.  */\n\
/* config.h.in.  Generated from configure.ac by autoheader.  */\n\
\n\
\n\
#ifndef __ERTS_CONFIG_H__\n\
#define __ERTS_CONFIG_H__\n\
\n\
#define GHBN_R_SOLARIS  2\n\
#define GHBN_R_AIX      3\n\
#define GHBN_R_GLIBC    4\n\
\n\
\n\
/* Define if building universal (internal helper macro) */\n\
/* #undef AC_APPLE_UNIVERSAL_BUILD */\n\
\n\
/* Assumed cache-line size (in bytes) */\n\
#define ASSUMED_CACHE_LINE_SIZE 64\n\
\n\
/* Define the brk() argument type. */\n\
#define BRK_ARG_TYPE void *\n\
\n\
/* Define the brk() return type. */\n\
#define BRK_RET_TYPE int\n\
\n\
/* Define if you need to include rpc/types.h to get INADDR_LOOPBACK defined */\n\
/* #undef DEF_INADDR_LOOPBACK_IN_RPC_TYPES_H */\n\
\n\
/* Define if you need to include winsock2.h to get INADDR_LOOPBACK defined */\n\
/* #undef DEF_INADDR_LOOPBACK_IN_WINSOCK2_H */\n\
\n\
/* Define to 1 if your processor stores the words in a double in middle-endian\n\
   format (like some ARMs). */\n\
/* #undef DOUBLE_MIDDLE_ENDIAN */\n\
\n\
/* Define if we need frame pointers on the Erlang stack */\n\
#define ERLANG_FRAME_POINTERS /**/\n\
\n\
/* Define if sbrk()/brk() wrappers can track malloc()s core memory use */\n\
/* #undef ERTS_BRK_WRAPPERS_CAN_TRACK_MALLOC */\n\
\n\
/* Define to override the default number of write_concurrency locks */\n\
/* #undef ERTS_DB_HASH_LOCK_CNT */\n\
\n\
/* The only reason ERTS_EMU_CMDLINE_FLAGS exists is to force modification of\n\
   config.h when the emulator command line flags are modified by configure */\n\
#define ERTS_EMU_CMDLINE_FLAGS \" -Werror=undef -Werror=implicit -Werror=return-type  -fno-common -fstrict-flex-arrays=3 -fno-strict-aliasing -fno-delete-null-pointer-checks -fno-strict-overflow -fexceptions -fcf-protection=full -fstack-protector-strong -fstack-clash-protection -U_FORTIFY_SOURCE -D_FORTIFY_SOURCE=3 -march=x86-64 -mtune=generic -O2 -pipe -fno-plt -fexceptions -Wp,-D_FORTIFY_SOURCE=3 -Wformat -Werror=format-security -fstack-clash-protection -fcf-protection -fno-omit-frame-pointer -mno-omit-leaf-frame-pointer -g -ffile-prefix-map=/build/erlang/src=/usr/src/debug/erlang -flto=auto -ffat-lto-objects -I/build/erlang/src/otp/erts/x86_64-pc-linux-gnu   -D_GNU_SOURCE -D_GLIBCXX_ASSERTIONS -march=x86-64 -mtune=generic -O2 -pipe -fno-plt -fexceptions         -Wp,-D_FORTIFY_SOURCE=3 -Wformat -Werror=format-security         -fstack-clash-protection -fcf-protection         -fno-omit-frame-pointer -mno-omit-leaf-frame-pointer -Wp,-D_GLIBCXX_ASSERTIONS -g -ffile-prefix-map=/build/erlang/src=/usr/src/debug/erlang -flto=auto -std=c++17 -g -Og  -Werror=undef -Werror=implicit -Werror=return-type  -fno-common -fstrict-flex-arrays=3 -fno-strict-aliasing -fno-delete-null-pointer-checks -fno-strict-overflow -fexceptions -fcf-protection=full -fstack-protector-strong -fstack-clash-protection -U_FORTIFY_SOURCE -D_FORTIFY_SOURCE=3 -march=x86-64 -mtune=generic  -pipe -fno-plt -fexceptions -Wp,-D_FORTIFY_SOURCE=3 -Wformat -Werror=format-security -fstack-clash-protection -fcf-protection -fno-omit-frame-pointer -mno-omit-leaf-frame-pointer -g -ffile-prefix-map=/build/erlang/src=/usr/src/debug/erlang -flto=auto -ffat-lto-objects -I/build/erlang/src/otp/erts/x86_64-pc-linux-gnu   -D_GNU_SOURCE -g -Og  -D_GLIBCXX_ASSERTIONS -march=x86-64 -mtune=generic  -pipe -fno-plt -fexceptions         -Wp,-D_FORTIFY_SOURCE=3 -Wformat -Werror=format-security         -fstack-clash-protection -fcf-protection         -fno-omit-frame-pointer -mno-omit-leaf-frame-pointer -Wp,-D_GLIBCXX_ASSERTIONS -g -ffile-prefix-map=/build/erlang/src=/usr/src/debug/erlang -flto=auto -std=c++17  -DUSE_THREADS -D_THREAD_SAFE -D_REENTRANT -DPOSIX_THREADS -D_POSIX_THREAD_SAFE_FUNCTIONS  -Werror=undef -Werror=implicit -Werror=return-type  -Wall -Wstrict-prototypes -Wpointer-arith -Wmissing-prototypes -Wdeclaration-after-statement\"\n\
\n\
/* Define if you have kernel poll and want to use it */\n\
#define ERTS_ENABLE_KERNEL_POLL 1\n\
\n\
/* Define as 1 if you want to enable microstate accounting, 2 if you want\n\
   extra states */\n\
#define ERTS_ENABLE_MSACC 1\n\
\n\
/* Define > 0 if big-endian < 0 if little-endian, or 0 if unknown */\n\
#define ERTS_ENDIANNESS -1\n\
\n\
/* Define if ensurance of the monotonicity of OS monotonic timestamps should\n\
   be enabled */\n\
/* #undef ERTS_ENSURE_OS_MONOTONIC_TIME */\n\
\n\
/* Define if OS monotonic clock is corrected */\n\
#define ERTS_HAVE_CORRECTED_OS_MONOTONIC_TIME 1\n\
\n\
/* Define if you have a low resolution OS monotonic clock */\n\
/* #undef ERTS_HAVE_LOW_RESOLUTION_OS_MONOTONIC_LOW */\n\
\n\
/* Define if dlopen() needs to be called before first call to dlerror() */\n\
/* #undef ERTS_NEED_DLOPEN_BEFORE_DLERROR */\n\
\n\
/* Per-function attribute for disabling retpoline. This is *only* defined when\n\
   --with-spectre-mitigation=incomplete and has no effects otherwise */\n\
#define ERTS_NO_RETPOLINE \n\
\n\
/* Type qualifier restrict */\n\
#define ERTS_RESTRICT restrict\n\
\n\
/* Define if structure alignment is enough for allocators. If not defined,\n\
   64-bit alignment will be forced. */\n\
#define ERTS_STRUCTURE_ALIGNED_ALLOC 1\n\
\n\
/* Define if poll() should be used instead of select() */\n\
#define ERTS_USE_POLL 1\n\
\n\
/* Define if __after_morecore_hook can track malloc()s core memory use. */\n\
/* #undef ERTS___AFTER_MORECORE_HOOK_CAN_TRACK_MALLOC */\n\
\n\
/* ESOCK counter size */\n\
#define ESOCK_COUNTER_SIZE 64\n\
\n\
/* Enable esock */\n\
#define ESOCK_ENABLE 1\n\
\n\
/* Socket address dl length */\n\
/* #undef ESOCK_SDL_LEN */\n\
\n\
/* Interface hwaddr supported */\n\
/* #undef ESOCK_USE_ENADDR */\n\
\n\
/* Use extended error info */\n\
#define ESOCK_USE_EXTENDED_ERROR_INFO 1\n\
\n\
/* Interface hwaddr supported */\n\
#define ESOCK_USE_HWADDR /**/\n\
\n\
/* Interface ifindex supported */\n\
#define ESOCK_USE_IFINDEX /**/\n\
\n\
/* Interface map supported */\n\
#define ESOCK_USE_IFMAP /**/\n\
\n\
/* Interface index supported */\n\
/* #undef ESOCK_USE_INDEX */\n\
\n\
/* Use SO_[RCV|SND]TMIEO */\n\
/* #undef ESOCK_USE_RCVSNDTIMEO */\n\
\n\
/* Use socket registry by default */\n\
#define ESOCK_USE_SOCKET_REGISTRY 1\n\
\n\
/* Use verbose mutex names */\n\
#define ESOCK_VERBOSE_MTX_NAMES 1\n\
\n\
/* Define if bigendian */\n\
/* #undef ETHR_BIGENDIAN */\n\
\n\
/* Define if gcc won't let you clobber ebx with cmpxchg8b and position\n\
   independent code */\n\
/* #undef ETHR_CMPXCHG8B_PIC_NO_CLOBBER_EBX */\n\
\n\
/* Define if you get a register shortage with cmpxchg8b and position\n\
   independent code */\n\
/* #undef ETHR_CMPXCHG8B_REGISTER_SHORTAGE */\n\
\n\
/* Define if you want to disable native ethread implementations */\n\
/* #undef ETHR_DISABLE_NATIVE_IMPLS */\n\
\n\
/* Define if you want to force usage of pthread rwlocks */\n\
/* #undef ETHR_FORCE_PTHREAD_RWLOCK */\n\
\n\
/* Define if you use a gcc that supports the double word cmpxchg instruction\n\
   */\n\
#define ETHR_GCC_HAVE_DW_CMPXCHG_ASM_SUPPORT 1\n\
\n\
/* Define if you use a gcc that supports -msse2 and understand sse2 specific\n\
   asm statements */\n\
/* #undef ETHR_GCC_HAVE_SSE2_ASM_SUPPORT */\n\
\n\
/* Define if you have a clock_gettime() with a monotonic clock */\n\
#define ETHR_HAVE_CLOCK_GETTIME_MONOTONIC 1\n\
\n\
/* Define if you have all ethread defines */\n\
#define ETHR_HAVE_ETHREAD_DEFINES 1\n\
\n\
/* Define as a boolean indicating whether you have a gcc compatible compiler\n\
   capable of generating the ARM 'dc cvau' instruction, and are compiling for\n\
   an ARM processor with ARM DC instruction support, or not */\n\
#define ETHR_HAVE_GCC_ASM_ARM_DC_CVAU_INSTRUCTION 0\n\
\n\
/* Define as a boolean indicating whether you have a gcc compatible compiler\n\
   capable of generating the ARM 'dmb sy' instruction, and are compiling for\n\
   an ARM processor with ARM DMB instruction support, or not */\n\
#define ETHR_HAVE_GCC_ASM_ARM_DMB_INSTRUCTION 0\n\
\n\
/* Define as a boolean indicating whether you have a gcc compatible compiler\n\
   capable of generating the ARM 'dmb ld' instruction, and are compiling for\n\
   an ARM processor with ARM DMB instruction support, or not */\n\
#define ETHR_HAVE_GCC_ASM_ARM_DMB_LD_INSTRUCTION 0\n\
\n\
/* Define as a boolean indicating whether you have a gcc compatible compiler\n\
   capable of generating the ARM 'dmb st' instruction, and are compiling for\n\
   an ARM processor with ARM DMB instruction support, or not */\n\
#define ETHR_HAVE_GCC_ASM_ARM_DMB_ST_INSTRUCTION 0\n\
\n\
/* Define as a boolean indicating whether you have a gcc compatible compiler\n\
   capable of generating the ARM 'ic ivau' instruction, and are compiling for\n\
   an ARM processor with ARM IC instruction support, or not */\n\
#define ETHR_HAVE_GCC_ASM_ARM_IC_IVAU_INSTRUCTION 0\n\
\n\
/* Define as a boolean indicating whether you have a gcc compatible compiler\n\
   capable of generating the ARM 'isb sy' instruction, and are compiling for\n\
   an ARM processor with ARM ISB instruction support, or not */\n\
#define ETHR_HAVE_GCC_ASM_ARM_ISB_SY_INSTRUCTION 0\n\
\n\
/* Define as a boolean indicating whether you have a gcc __atomic builtins or\n\
   not */\n\
#define ETHR_HAVE_GCC___ATOMIC_BUILTINS 1\n\
\n\
/* Define if you have a monotonic gethrtime() */\n\
/* #undef ETHR_HAVE_GETHRTIME */\n\
\n\
/* Define if you have libatomic_ops atomic operations */\n\
/* #undef ETHR_HAVE_LIBATOMIC_OPS */\n\
\n\
/* Define if you have a linux futex implementation. */\n\
#define ETHR_HAVE_LINUX_FUTEX 1\n\
\n\
/* Define if you have a mach clock_get_time() with a monotonic clock */\n\
/* #undef ETHR_HAVE_MACH_CLOCK_GET_TIME */\n\
\n\
/* Define if the pthread.h header file is in pthread/mit directory. */\n\
/* #undef ETHR_HAVE_MIT_PTHREAD_H */\n\
\n\
/* Define if you have the pthread_attr_setguardsize function. */\n\
#define ETHR_HAVE_PTHREAD_ATTR_SETGUARDSIZE 1\n\
\n\
/* Define if pthread_cond_timedwait() can be used with a monotonic clock */\n\
#define ETHR_HAVE_PTHREAD_COND_TIMEDWAIT_MONOTONIC 1\n\
\n\
/* Define if you have ibm style pthread_getname_np */\n\
/* #undef ETHR_HAVE_PTHREAD_GETNAME_NP_2 */\n\
\n\
/* Define if you have linux style pthread_getname_np */\n\
#define ETHR_HAVE_PTHREAD_GETNAME_NP_3 1\n\
\n\
/* Define if you have the <pthread.h> header file. */\n\
#define ETHR_HAVE_PTHREAD_H 1\n\
\n\
/* Define if you have the pthread_rwlockattr_setkind_np() function. */\n\
#define ETHR_HAVE_PTHREAD_RWLOCKATTR_SETKIND_NP 1\n\
\n\
/* Define if you have the PTHREAD_RWLOCK_PREFER_WRITER_NONRECURSIVE_NP rwlock\n\
   attribute. */\n\
#define ETHR_HAVE_PTHREAD_RWLOCK_PREFER_WRITER_NONRECURSIVE_NP 1\n\
\n\
/* Define if you have darwin style pthread_setname_np */\n\
/* #undef ETHR_HAVE_PTHREAD_SETNAME_NP_1 */\n\
\n\
/* Define if you have linux style pthread_setname_np with 2 args */\n\
#define ETHR_HAVE_PTHREAD_SETNAME_NP_2 1\n\
\n\
/* Define if you have netbsd style pthread_setname_np with 3 args */\n\
/* #undef ETHR_HAVE_PTHREAD_SETNAME_NP_3 */\n\
\n\
/* Define if you have bsd style pthread_set_name_np */\n\
/* #undef ETHR_HAVE_PTHREAD_SET_NAME_NP_2 */\n\
\n\
/* Define if you have the pthread_spin_lock function. */\n\
#define ETHR_HAVE_PTHREAD_SPIN_LOCK 1\n\
\n\
/* Define if you have the pthread_yield() function. */\n\
/* #undef ETHR_HAVE_PTHREAD_YIELD */\n\
\n\
/* Define if you have the <sched.h> header file. */\n\
#define ETHR_HAVE_SCHED_H 1\n\
\n\
/* Define if you have the sched_yield() function. */\n\
#define ETHR_HAVE_SCHED_YIELD 1\n\
\n\
/* Define if you have the <sys/time.h> header file. */\n\
#define ETHR_HAVE_SYS_TIME_H 1\n\
\n\
/* Define if you can use PTHREAD_STACK_MIN */\n\
#define ETHR_HAVE_USABLE_PTHREAD_STACK_MIN 1\n\
\n\
/* Define if you have _InterlockedAnd() */\n\
/* #undef ETHR_HAVE__INTERLOCKEDAND */\n\
\n\
/* Define if you have _InterlockedAnd64() */\n\
/* #undef ETHR_HAVE__INTERLOCKEDAND64 */\n\
\n\
/* Define if you have _InterlockedCompareExchange() */\n\
/* #undef ETHR_HAVE__INTERLOCKEDCOMPAREEXCHANGE */\n\
\n\
/* Define if you have _InterlockedCompareExchange128() */\n\
/* #undef ETHR_HAVE__INTERLOCKEDCOMPAREEXCHANGE128 */\n\
\n\
/* Define if you have _InterlockedCompareExchange64() */\n\
/* #undef ETHR_HAVE__INTERLOCKEDCOMPAREEXCHANGE64 */\n\
\n\
/* Define if you have _InterlockedCompareExchange64_acq() */\n\
/* #undef ETHR_HAVE__INTERLOCKEDCOMPAREEXCHANGE64_ACQ */\n\
\n\
/* Define if you have _InterlockedCompareExchange64_rel() */\n\
/* #undef ETHR_HAVE__INTERLOCKEDCOMPAREEXCHANGE64_REL */\n\
\n\
/* Define if you have _InterlockedCompareExchange_acq() */\n\
/* #undef ETHR_HAVE__INTERLOCKEDCOMPAREEXCHANGE_ACQ */\n\
\n\
/* Define if you have _InterlockedCompareExchange_rel() */\n\
/* #undef ETHR_HAVE__INTERLOCKEDCOMPAREEXCHANGE_REL */\n\
\n\
/* Define if you have _InterlockedDecrement() */\n\
/* #undef ETHR_HAVE__INTERLOCKEDDECREMENT */\n\
\n\
/* Define if you have _InterlockedDecrement64() */\n\
/* #undef ETHR_HAVE__INTERLOCKEDDECREMENT64 */\n\
\n\
/* Define if you have _InterlockedDecrement64_rel() */\n\
/* #undef ETHR_HAVE__INTERLOCKEDDECREMENT64_REL */\n\
\n\
/* Define if you have _InterlockedDecrement_rel() */\n\
/* #undef ETHR_HAVE__INTERLOCKEDDECREMENT_REL */\n\
\n\
/* Define if you have _InterlockedExchange() */\n\
/* #undef ETHR_HAVE__INTERLOCKEDEXCHANGE */\n\
\n\
/* Define if you have _InterlockedExchange64() */\n\
/* #undef ETHR_HAVE__INTERLOCKEDEXCHANGE64 */\n\
\n\
/* Define if you have _InterlockedExchangeAdd() */\n\
/* #undef ETHR_HAVE__INTERLOCKEDEXCHANGEADD */\n\
\n\
/* Define if you have _InterlockedExchangeAdd64() */\n\
/* #undef ETHR_HAVE__INTERLOCKEDEXCHANGEADD64 */\n\
\n\
/* Define if you have _InterlockedExchangeAdd64_acq() */\n\
/* #undef ETHR_HAVE__INTERLOCKEDEXCHANGEADD64_ACQ */\n\
\n\
/* Define if you have _InterlockedExchangeAdd_acq() */\n\
/* #undef ETHR_HAVE__INTERLOCKEDEXCHANGEADD_ACQ */\n\
\n\
/* Define if you have _InterlockedIncrement() */\n\
/* #undef ETHR_HAVE__INTERLOCKEDINCREMENT */\n\
\n\
/* Define if you have _InterlockedIncrement64() */\n\
/* #undef ETHR_HAVE__INTERLOCKEDINCREMENT64 */\n\
\n\
/* Define if you have _InterlockedIncrement64_acq() */\n\
/* #undef ETHR_HAVE__INTERLOCKEDINCREMENT64_ACQ */\n\
\n\
/* Define if you have _InterlockedIncrement_acq() */\n\
/* #undef ETHR_HAVE__INTERLOCKEDINCREMENT_ACQ */\n\
\n\
/* Define if you have _InterlockedOr() */\n\
/* #undef ETHR_HAVE__INTERLOCKEDOR */\n\
\n\
/* Define if you have _InterlockedOr64() */\n\
/* #undef ETHR_HAVE__INTERLOCKEDOR64 */\n\
\n\
/* Define as a bitmask corresponding to the word sizes that\n\
   __atomic_add_fetch() can handle on your system */\n\
#define ETHR_HAVE___atomic_add_fetch 12\n\
\n\
/* Define as a bitmask corresponding to the word sizes that\n\
   __atomic_compare_exchange_n() can handle on your system */\n\
#define ETHR_HAVE___atomic_compare_exchange_n 12\n\
\n\
/* Define as a bitmask corresponding to the word sizes that\n\
   __atomic_fetch_and() can handle on your system */\n\
#define ETHR_HAVE___atomic_fetch_and 12\n\
\n\
/* Define as a bitmask corresponding to the word sizes that\n\
   __atomic_fetch_or() can handle on your system */\n\
#define ETHR_HAVE___atomic_fetch_or 12\n\
\n\
/* Define as a bitmask corresponding to the word sizes that __atomic_load_n()\n\
   can handle on your system */\n\
#define ETHR_HAVE___atomic_load_n 12\n\
\n\
/* Define as a bitmask corresponding to the word sizes that __atomic_store_n()\n\
   can handle on your system */\n\
#define ETHR_HAVE___atomic_store_n 12\n\
\n\
/* Define as a bitmask corresponding to the word sizes that\n\
   __sync_add_and_fetch() can handle on your system */\n\
#define ETHR_HAVE___sync_add_and_fetch 12\n\
\n\
/* Define as a bitmask corresponding to the word sizes that\n\
   __sync_fetch_and_and() can handle on your system */\n\
#define ETHR_HAVE___sync_fetch_and_and 12\n\
\n\
/* Define as a bitmask corresponding to the word sizes that\n\
   __sync_fetch_and_or() can handle on your system */\n\
#define ETHR_HAVE___sync_fetch_and_or 12\n\
\n\
/* Define as a bitmask corresponding to the word sizes that\n\
   __sync_synchronize() can handle on your system */\n\
#define ETHR_HAVE___sync_synchronize ~0\n\
\n\
/* Define as a bitmask corresponding to the word sizes that\n\
   __sync_val_compare_and_swap() can handle on your system */\n\
#define ETHR_HAVE___sync_val_compare_and_swap 12\n\
\n\
/* Define if you want to modify the default stack size */\n\
/* #undef ETHR_MODIFIED_DEFAULT_STACK_SIZE */\n\
\n\
/* Define to the monotonic clock id to use */\n\
#define ETHR_MONOTONIC_CLOCK_ID CLOCK_MONOTONIC\n\
\n\
/* Define if you need the <nptl/pthread.h> header file. */\n\
/* #undef ETHR_NEED_NPTL_PTHREAD_H */\n\
\n\
/* Define if you have the powerpc lwsync instruction */\n\
/* #undef ETHR_PPC_HAVE_LWSYNC */\n\
\n\
/* Define if you do not have the powerpc lwsync instruction */\n\
/* #undef ETHR_PPC_HAVE_NO_LWSYNC */\n\
\n\
/* Define if you prefer gcc native ethread implementations */\n\
/* #undef ETHR_PREFER_GCC_NATIVE_IMPLS */\n\
\n\
/* Define if you prefer libatomic_ops native ethread implementations */\n\
/* #undef ETHR_PREFER_LIBATOMIC_OPS_NATIVE_IMPLS */\n\
\n\
/* Define if you have pthreads */\n\
#define ETHR_PTHREADS 1\n\
\n\
/* Define if pthread_yield() returns an int. */\n\
/* #undef ETHR_PTHREAD_YIELD_RET_INT */\n\
\n\
/* Define if sched_yield() returns an int. */\n\
#define ETHR_SCHED_YIELD_RET_INT 1\n\
\n\
/* Define to the size of AO_t if libatomic_ops is used */\n\
/* #undef ETHR_SIZEOF_AO_T */\n\
\n\
/* Define to the size of int */\n\
#define ETHR_SIZEOF_INT 4\n\
\n\
/* Define to the size of long */\n\
#define ETHR_SIZEOF_LONG 8\n\
\n\
/* Define to the size of long long */\n\
#define ETHR_SIZEOF_LONG_LONG 8\n\
\n\
/* Define to the size of pointers */\n\
#define ETHR_SIZEOF_PTR 8\n\
\n\
/* Define to the size of __int128_t */\n\
#define ETHR_SIZEOF___INT128_T 16\n\
\n\
/* Define to the size of __int64 */\n\
#define ETHR_SIZEOF___INT64 0\n\
\n\
/* Define if you want to enable check for native ethread implementations */\n\
#define ETHR_SMP_REQUIRE_NATIVE_IMPLS 1\n\
\n\
/* Define if only run in Sparc PSO, or TSO mode */\n\
/* #undef ETHR_SPARC_PSO */\n\
\n\
/* Define if run in Sparc RMO, PSO, or TSO mode */\n\
/* #undef ETHR_SPARC_RMO */\n\
\n\
/* Define if only run in Sparc TSO mode */\n\
/* #undef ETHR_SPARC_TSO */\n\
\n\
/* Define as a boolean indicating whether you trust gcc's __atomic_* builtins\n\
   memory barrier implementations, or not */\n\
#define ETHR_TRUST_GCC_ATOMIC_BUILTINS_MEMORY_BARRIERS 0\n\
\n\
/* Define if you have win32 threads */\n\
/* #undef ETHR_WIN32_THREADS */\n\
\n\
/* Define if x86/x86_64 out of order instructions should be synchronized */\n\
/* #undef ETHR_X86_OUT_OF_ORDER */\n\
\n\
/* Define to 1 if _Float16 can be converted from/to double. */\n\
#define FLOAT16_IS_CONVERTIBLE 1\n\
\n\
/* Define to 1 if you have the <arpa/nameser.h> header file. */\n\
#define HAVE_ARPA_NAMESER_H 1\n\
\n\
/* Define to 1 if you have the 'brk' function. */\n\
#define HAVE_BRK 1\n\
\n\
/* Define to 1 if you have the 'clock_getres' function. */\n\
#define HAVE_CLOCK_GETRES 1\n\
\n\
/* define if clock_gettime() works for getting thread time */\n\
#define HAVE_CLOCK_GETTIME_CPU_TIME /**/\n\
\n\
/* Define if you have clock_gettime(CLOCK_MONOTONIC_RAW, _) */\n\
#define HAVE_CLOCK_GETTIME_MONOTONIC_RAW 1\n\
\n\
/* Define to 1 if you have the 'clock_get_attributes' function. */\n\
/* #undef HAVE_CLOCK_GET_ATTRIBUTES */\n\
\n\
/* Define to 1 if you have the 'closefrom' function. */\n\
#define HAVE_CLOSEFROM 1\n\
\n\
/* Define if you have a decl of fread that conflicts with int fread */\n\
#define HAVE_CONFLICTING_FREAD_DECLARATION 1\n\
\n\
/* Define if you have a putenv() that stores a copy of the key-value pair */\n\
/* #undef HAVE_COPYING_PUTENV */\n\
\n\
/* Define if you have cpuset_getaffinity/cpuset_setaffinity */\n\
/* #undef HAVE_CPUSET_xETAFFINITY */\n\
\n\
/* Define to 1 if you have the <curses.h> header file. */\n\
#define HAVE_CURSES_H 1\n\
\n\
/* Define to 1 if you have the declaration of 'daylight', and to 0 if you\n\
   don't. */\n\
#define HAVE_DECL_DAYLIGHT 1\n\
\n\
/* Define to 1 if you have the declaration of 'getrlimit', and to 0 if you\n\
   don't. */\n\
#define HAVE_DECL_GETRLIMIT 1\n\
\n\
/* Define to 1 if you have the declaration of 'IN6ADDR_ANY_INIT', and to 0 if\n\
   you don't. */\n\
#define HAVE_DECL_IN6ADDR_ANY_INIT 1\n\
\n\
/* Define to 1 if you have the declaration of 'IN6ADDR_LOOPBACK_INIT', and to\n\
   0 if you don't. */\n\
#define HAVE_DECL_IN6ADDR_LOOPBACK_INIT 1\n\
\n\
/* Define to 1 if you have the declaration of 'IPV6_V6ONLY', and to 0 if you\n\
   don't. */\n\
#define HAVE_DECL_IPV6_V6ONLY 1\n\
\n\
/* Define to 1 if you have the declaration of 'posix2time', and to 0 if you\n\
   don't. */\n\
#define HAVE_DECL_POSIX2TIME 0\n\
\n\
/* Define to 1 if you have the declaration of 'RLIMIT_STACK', and to 0 if you\n\
   don't. */\n\
#define HAVE_DECL_RLIMIT_STACK 1\n\
\n\
/* Define to 1 if you have the declaration of 'SCTPS_BOUND', and to 0 if you\n\
   don't. */\n\
#define HAVE_DECL_SCTPS_BOUND 0\n\
\n\
/* Define to 1 if you have the declaration of 'SCTPS_COOKIE_ECHOED', and to 0\n\
   if you don't. */\n\
#define HAVE_DECL_SCTPS_COOKIE_ECHOED 0\n\
\n\
/* Define to 1 if you have the declaration of 'SCTPS_COOKIE_WAIT', and to 0 if\n\
   you don't. */\n\
#define HAVE_DECL_SCTPS_COOKIE_WAIT 0\n\
\n\
/* Define to 1 if you have the declaration of 'SCTPS_ESTABLISHED', and to 0 if\n\
   you don't. */\n\
#define HAVE_DECL_SCTPS_ESTABLISHED 0\n\
\n\
/* Define to 1 if you have the declaration of 'SCTPS_IDLE', and to 0 if you\n\
   don't. */\n\
#define HAVE_DECL_SCTPS_IDLE 0\n\
\n\
/* Define to 1 if you have the declaration of 'SCTPS_LISTEN', and to 0 if you\n\
   don't. */\n\
#define HAVE_DECL_SCTPS_LISTEN 0\n\
\n\
/* Define to 1 if you have the declaration of 'SCTPS_SHUTDOWN_ACK_SENT', and\n\
   to 0 if you don't. */\n\
#define HAVE_DECL_SCTPS_SHUTDOWN_ACK_SENT 0\n\
\n\
/* Define to 1 if you have the declaration of 'SCTPS_SHUTDOWN_PENDING', and to\n\
   0 if you don't. */\n\
#define HAVE_DECL_SCTPS_SHUTDOWN_PENDING 0\n\
\n\
/* Define to 1 if you have the declaration of 'SCTPS_SHUTDOWN_RECEIVED', and\n\
   to 0 if you don't. */\n\
#define HAVE_DECL_SCTPS_SHUTDOWN_RECEIVED 0\n\
\n\
/* Define to 1 if you have the declaration of 'SCTPS_SHUTDOWN_SENT', and to 0\n\
   if you don't. */\n\
#define HAVE_DECL_SCTPS_SHUTDOWN_SENT 0\n\
\n\
/* Define to 1 if you have the declaration of 'SCTP_ABORT', and to 0 if you\n\
   don't. */\n\
#define HAVE_DECL_SCTP_ABORT 1\n\
\n\
/* Define to 1 if you have the declaration of 'SCTP_ADDR_CONFIRMED', and to 0\n\
   if you don't. */\n\
#define HAVE_DECL_SCTP_ADDR_CONFIRMED 1\n\
\n\
/* Define to 1 if you have the declaration of 'SCTP_ADDR_OVER', and to 0 if\n\
   you don't. */\n\
#define HAVE_DECL_SCTP_ADDR_OVER 1\n\
\n\
/* Define to 1 if you have the declaration of 'SCTP_BOUND', and to 0 if you\n\
   don't. */\n\
#define HAVE_DECL_SCTP_BOUND 0\n\
\n\
/* Define to 1 if you have the declaration of 'SCTP_CLOSED', and to 0 if you\n\
   don't. */\n\
#define HAVE_DECL_SCTP_CLOSED 1\n\
\n\
/* Define to 1 if you have the declaration of 'SCTP_COOKIE_ECHOED', and to 0\n\
   if you don't. */\n\
#define HAVE_DECL_SCTP_COOKIE_ECHOED 1\n\
\n\
/* Define to 1 if you have the declaration of 'SCTP_COOKIE_WAIT', and to 0 if\n\
   you don't. */\n\
#define HAVE_DECL_SCTP_COOKIE_WAIT 1\n\
\n\
/* Define to 1 if you have the declaration of 'SCTP_DELAYED_ACK_TIME', and to\n\
   0 if you don't. */\n\
#define HAVE_DECL_SCTP_DELAYED_ACK_TIME 1\n\
\n\
/* Define to 1 if you have the declaration of 'SCTP_EMPTY', and to 0 if you\n\
   don't. */\n\
#define HAVE_DECL_SCTP_EMPTY 1\n\
\n\
/* Define to 1 if you have the declaration of 'SCTP_EOF', and to 0 if you\n\
   don't. */\n\
#define HAVE_DECL_SCTP_EOF 1\n\
\n\
/* Define to 1 if you have the declaration of 'SCTP_ESTABLISHED', and to 0 if\n\
   you don't. */\n\
#define HAVE_DECL_SCTP_ESTABLISHED 1\n\
\n\
/* Define to 1 if you have the declaration of 'SCTP_LISTEN', and to 0 if you\n\
   don't. */\n\
#define HAVE_DECL_SCTP_LISTEN 0\n\
\n\
/* Define to 1 if you have the declaration of 'SCTP_SENDALL', and to 0 if you\n\
   don't. */\n\
#define HAVE_DECL_SCTP_SENDALL 1\n\
\n\
/* Define to 1 if you have the declaration of 'SCTP_SHUTDOWN_ACK_SENT', and to\n\
   0 if you don't. */\n\
#define HAVE_DECL_SCTP_SHUTDOWN_ACK_SENT 1\n\
\n\
/* Define to 1 if you have the declaration of 'SCTP_SHUTDOWN_PENDING', and to\n\
   0 if you don't. */\n\
#define HAVE_DECL_SCTP_SHUTDOWN_PENDING 1\n\
\n\
/* Define to 1 if you have the declaration of 'SCTP_SHUTDOWN_RECEIVED', and to\n\
   0 if you don't. */\n\
#define HAVE_DECL_SCTP_SHUTDOWN_RECEIVED 1\n\
\n\
/* Define to 1 if you have the declaration of 'SCTP_SHUTDOWN_SENT', and to 0\n\
   if you don't. */\n\
#define HAVE_DECL_SCTP_SHUTDOWN_SENT 1\n\
\n\
/* Define to 1 if you have the declaration of 'SCTP_UNCONFIRMED', and to 0 if\n\
   you don't. */\n\
#define HAVE_DECL_SCTP_UNCONFIRMED 1\n\
\n\
/* Define to 1 if you have the declaration of 'SCTP_UNORDERED', and to 0 if\n\
   you don't. */\n\
#define HAVE_DECL_SCTP_UNORDERED 1\n\
\n\
/* Define to 1 if you have the declaration of 'setrlimit', and to 0 if you\n\
   don't. */\n\
#define HAVE_DECL_SETRLIMIT 1\n\
\n\
/* Define to 1 if you have the declaration of 'time2posix', and to 0 if you\n\
   don't. */\n\
#define HAVE_DECL_TIME2POSIX 0\n\
\n\
/* Define to 1 if you have the <dirent.h> header file, and it defines 'DIR'.\n\
   */\n\
#define HAVE_DIRENT_H 1\n\
\n\
/* Define to 1 if you have the <dlfcn.h> header file. */\n\
#define HAVE_DLFCN_H 1\n\
\n\
/* Define to 1 if you have the 'dlopen' function. */\n\
#define HAVE_DLOPEN 1\n\
\n\
/* Define to 1 if you have the 'dlvsym' function. */\n\
#define HAVE_DLVSYM 1\n\
\n\
/* Define to 1 if you don't have 'vprintf' but do have '_doprnt.' */\n\
/* #undef HAVE_DOPRNT */\n\
\n\
/* Define to 1 if you have the <elf.h> header file. */\n\
#define HAVE_ELF_H 1\n\
\n\
/* Define to 1 if you have the 'endprotoent' function. */\n\
#define HAVE_ENDPROTOENT 1\n\
\n\
/* Define if you have the 'end' symbol */\n\
#define HAVE_END_SYMBOL 1\n\
\n\
/* Define if you have a working fallocate() */\n\
#define HAVE_FALLOCATE 1\n\
\n\
/* Define to 1 if you have the <fcntl.h> header file. */\n\
#define HAVE_FCNTL_H 1\n\
\n\
/* Define to 1 if you have the 'fdatasync' function. */\n\
#define HAVE_FDATASYNC 1\n\
\n\
/* Define to 1 if you have the 'finite' function. */\n\
#define HAVE_FINITE 1\n\
\n\
/* Define to 1 if you have the 'flockfile' function. */\n\
#define HAVE_FLOCKFILE 1\n\
\n\
/* Define to 1 if you have the 'fpsetmask' function. */\n\
/* #undef HAVE_FPSETMASK */\n\
\n\
/* Define to 1 if you have the 'fstat' function. */\n\
#define HAVE_FSTAT 1\n\
\n\
/* Define if you have fwrite_unlocked */\n\
#define HAVE_FWRITE_UNLOCKED 1\n\
\n\
/* define if compiler support _Pragma('GCC diagnostic ignored\n\
   '-Waddress-of-packed-member'') */\n\
#define HAVE_GCC_DIAG_IGNORE_WADDRESS_OF_PACKED_MEMBER 1\n\
\n\
/* define if compiler support _Pragma('GCC diagnostic ignored\n\
   '-Wformat-nonliteral'') */\n\
#define HAVE_GCC_DIAG_IGNORE_WFORMAT_NONLITERAL 1\n\
\n\
/* Define to 1 if you have a good `getaddrinfo' function. */\n\
#define HAVE_GETADDRINFO 1\n\
\n\
/* Define to 1 if you have the 'gethostbyname2' function. */\n\
#define HAVE_GETHOSTBYNAME2 1\n\
\n\
/* Define to flavour of gethostbyname_r */\n\
#define HAVE_GETHOSTBYNAME_R GHBN_R_GLIBC\n\
\n\
/* Define as 1 if function exists */\n\
#define HAVE_GETHOSTNAME 1\n\
\n\
/* Define to 1 if you have the 'gethrtime' function. */\n\
/* #undef HAVE_GETHRTIME */\n\
\n\
/* define if gethrvtime() works and uses ioctl() to /proc/self */\n\
/* #undef HAVE_GETHRVTIME_PROCFS_IOCTL */\n\
\n\
/* Define to 1 if you have the 'getifaddrs' function. */\n\
#define HAVE_GETIFADDRS 1\n\
\n\
/* Define to 1 if you have the 'getipnodebyaddr' function. */\n\
/* #undef HAVE_GETIPNODEBYADDR */\n\
\n\
/* Define to 1 if you have the 'getipnodebyname' function. */\n\
/* #undef HAVE_GETIPNODEBYNAME */\n\
\n\
/* Define to 1 if you have a good `getnameinfo' function. */\n\
#define HAVE_GETNAMEINFO 1\n\
\n\
/* Define to 1 if you have the 'getprotoent' function. */\n\
#define HAVE_GETPROTOENT 1\n\
\n\
/* Define to 1 if you have the 'getrusage' function. */\n\
#define HAVE_GETRUSAGE 1\n\
\n\
/* Define to 1 if you have the 'gettimeofday' function. */\n\
#define HAVE_GETTIMEOFDAY 1\n\
\n\
/* Define to 1 if you have the 'gmtime_r' function. */\n\
#define HAVE_GMTIME_R 1\n\
\n\
/* Define to 1 if you have the <ieeefp.h> header file. */\n\
/* #undef HAVE_IEEEFP_H */\n\
\n\
/* Define to 1 if you have the 'ieee_handler' function. */\n\
/* #undef HAVE_IEEE_HANDLER */\n\
\n\
/* Define to 1 if you have the <ifaddrs.h> header file. */\n\
#define HAVE_IFADDRS_H 1\n\
\n\
/* Define as 1 if function exists */\n\
#define HAVE_IF_FREENAMEINDEX 1\n\
\n\
/* Define as 1 if function exists */\n\
#define HAVE_IF_INDEXTONAME 1\n\
\n\
/* Define as 1 if function exists */\n\
#define HAVE_IF_NAMEINDEX 1\n\
\n\
/* Define as 1 if function exists */\n\
#define HAVE_IF_NAMETOINDEX 1\n\
\n\
/* Define if ipv6 is present */\n\
#define HAVE_IN6 1\n\
\n\
/* Define to 1 if you have the variable in6addr_any declared. */\n\
#define HAVE_IN6ADDR_ANY 1\n\
\n\
/* Define to 1 if you have the variable in6addr_loopback declared. */\n\
#define HAVE_IN6ADDR_LOOPBACK 1\n\
\n\
/* Define to 1 if you have the `inet_pton' function. */\n\
#define HAVE_INET_PTON 1\n\
\n\
/* Define to 1 if you have the <inttypes.h> header file. */\n\
#define HAVE_INTTYPES_H 1\n\
\n\
/* Early linux used in_addr6 instead of in6_addr, define if you have this */\n\
/* #undef HAVE_IN_ADDR6_STRUCT */\n\
\n\
/* Define to 1 if you have the `isfinite' function. */\n\
/* #undef HAVE_ISFINITE */\n\
\n\
/* Define to 1 if you have the 'isinf' function. */\n\
#define HAVE_ISINF 1\n\
\n\
/* Define to 1 if you have the 'isnan' function. */\n\
#define HAVE_ISNAN 1\n\
\n\
/* Define if you have kstat */\n\
/* #undef HAVE_KSTAT */\n\
\n\
/* Define to 1 if you have the <langinfo.h> header file. */\n\
#define HAVE_LANGINFO_H 1\n\
\n\
/* Define to 1 if you have the 'dl' library (-ldl). */\n\
#define HAVE_LIBDL 1\n\
\n\
/* Define to 1 if you have the 'dlpi' library (-ldlpi). */\n\
/* #undef HAVE_LIBDLPI */\n\
\n\
/* Define to 1 if you have the <libdlpi.h> header file. */\n\
/* #undef HAVE_LIBDLPI_H */\n\
\n\
/* Define to 1 if you have the 'inet' library (-linet). */\n\
/* #undef HAVE_LIBINET */\n\
\n\
/* Define to 1 if you have the 'm' library (-lm). */\n\
#define HAVE_LIBM 1\n\
\n\
/* Define to 1 if you have the 'sctp' library (-lsctp). */\n\
/* #undef HAVE_LIBSCTP */\n\
\n\
/* Define to 1 if you have the 'util' library (-lutil). */\n\
#define HAVE_LIBUTIL 1\n\
\n\
/* Define to 1 if you have the <libutil.h> header file. */\n\
/* #undef HAVE_LIBUTIL_H */\n\
\n\
/* Define to 1 if you have the `z' library (-lz). */\n\
#define HAVE_LIBZ 1\n\
\n\
/* Define to 1 if you have the <limits.h> header file. */\n\
#define HAVE_LIMITS_H 1\n\
\n\
/* Define to 1 if you have the <linux/errqueue.h> header file. */\n\
#define HAVE_LINUX_ERRQUEUE_H 1\n\
\n\
/* Define to 1 if you have the <linux/falloc.h> header file. */\n\
#define HAVE_LINUX_FALLOC_H 1\n\
\n\
/* Define if the targeted system supports the `perf` profiler */\n\
#define HAVE_LINUX_PERF_SUPPORT 1\n\
\n\
/* Define if the target system is Linux and THP is available */\n\
#define HAVE_LINUX_THP 1\n\
\n\
/* Define to 1 if you have the <linux/types.h> header file. */\n\
#define HAVE_LINUX_TYPES_H 1\n\
\n\
/* Define to 1 if you have the 'localtime_r' function. */\n\
#define HAVE_LOCALTIME_R 1\n\
\n\
/* Define to 1 if you have the 'log2' function. */\n\
#define HAVE_LOG2 1\n\
\n\
/* Define to 1 if you have the <lttng/tracepoint-event.h> header file. */\n\
/* #undef HAVE_LTTNG_TRACEPOINT_EVENT_H */\n\
\n\
/* Define to 1 if you have the <lttng/tracepoint.h> header file. */\n\
/* #undef HAVE_LTTNG_TRACEPOINT_H */\n\
\n\
/* Define to 1 if you have the 'madvise' function. */\n\
#define HAVE_MADVISE 1\n\
\n\
/* Define to 1 if you have the <malloc.h> header file. */\n\
#define HAVE_MALLOC_H 1\n\
\n\
/* Define to 1 if you have the 'mallopt' function. */\n\
#define HAVE_MALLOPT 1\n\
\n\
/* Define to 1 if you have the 'memcpy' function. */\n\
#define HAVE_MEMCPY 1\n\
\n\
/* Define to 1 if you have the 'memmove' function. */\n\
#define HAVE_MEMMOVE 1\n\
\n\
/* Define to 1 if you have the 'memrchr' function. */\n\
#define HAVE_MEMRCHR 1\n\
\n\
/* Define if the pthread.h header file is in pthread/mit directory. */\n\
/* #undef HAVE_MIT_PTHREAD_H */\n\
\n\
/* Define to 1 if you have the 'mlockall' function. */\n\
#define HAVE_MLOCKALL 1\n\
\n\
/* Define to 1 if you have the 'mmap' function. */\n\
#define HAVE_MMAP 1\n\
\n\
/* Define if you have a monotonic erts_os_hrtime() implementation */\n\
#define HAVE_MONOTONIC_ERTS_SYS_HRTIME 1\n\
\n\
/* Define to 1 if you have the 'mprotect' function. */\n\
#define HAVE_MPROTECT 1\n\
\n\
/* Define to 1 if you have the 'mremap' function. */\n\
#define HAVE_MREMAP 1\n\
\n\
/* Define if setsockopt() accepts multicast options */\n\
#define HAVE_MULTICAST_SUPPORT 1\n\
\n\
/* Define to 1 if you have the <ndir.h> header file, and it defines 'DIR'. */\n\
/* #undef HAVE_NDIR_H */\n\
\n\
/* Define to 1 if you have the <netpacket/packet.h> header file. */\n\
#define HAVE_NETPACKET_PACKET_H 1\n\
\n\
/* Define to 1 if you have the <net/errno.h> header file. */\n\
/* #undef HAVE_NET_ERRNO_H */\n\
\n\
/* Define to 1 if you have the <net/if_dl.h> header file. */\n\
/* #undef HAVE_NET_IF_DL_H */\n\
\n\
/* Define to 1 if you have the 'nl_langinfo' function. */\n\
#define HAVE_NL_LANGINFO 1\n\
\n\
/* Define if you don't have a definition of INADDR_LOOPBACK */\n\
/* #undef HAVE_NO_INADDR_LOOPBACK */\n\
\n\
/* Define to 1 if you have the 'openpty' function. */\n\
#define HAVE_OPENPTY 1\n\
\n\
/* Define to 1 if you have the 'poll' function. */\n\
#define HAVE_POLL 1\n\
\n\
/* Define to 1 if you have the <poll.h> header file. */\n\
#define HAVE_POLL_H 1\n\
\n\
/* Define to 1 if you have the 'posix2time' function. */\n\
/* #undef HAVE_POSIX2TIME */\n\
\n\
/* Define to 1 if you have the 'posix_fadvise' function. */\n\
#define HAVE_POSIX_FADVISE 1\n\
\n\
/* Define if you have a working posix_fallocate() */\n\
#define HAVE_POSIX_FALLOCATE /**/\n\
\n\
/* Define to 1 if you have the 'posix_madvise' function. */\n\
#define HAVE_POSIX_MADVISE 1\n\
\n\
/* Define to 1 if you have the `posix_memalign' function. */\n\
#define HAVE_POSIX_MEMALIGN 1\n\
\n\
/* Define to 1 if you have the 'ppoll' function. */\n\
#define HAVE_PPOLL 1\n\
\n\
/* Define to 1 if you have the 'pread' function. */\n\
#define HAVE_PREAD 1\n\
\n\
/* Define if you have processor_bind functionality */\n\
/* #undef HAVE_PROCESSOR_BIND */\n\
\n\
/* Define if you have pset functionality */\n\
/* #undef HAVE_PSET */\n\
\n\
/* Define if you have the <pthread.h> header file. */\n\
#define HAVE_PTHREAD_H 1\n\
\n\
/* Define to 1 if you have the <pty.h> header file. */\n\
#define HAVE_PTY_H 1\n\
\n\
/* Define if you have putc_unlocked */\n\
#define HAVE_PUTC_UNLOCKED 1\n\
\n\
/* Define to 1 if you have the 'pwrite' function. */\n\
#define HAVE_PWRITE 1\n\
\n\
/* Define to 1 if you have the 'res_gethostbyname' function. */\n\
/* #undef HAVE_RES_GETHOSTBYNAME */\n\
\n\
/* Define to 1 if you have the 'sbrk' function. */\n\
#define HAVE_SBRK 1\n\
\n\
/* Define to 1 if you have the <sched.h> header file. */\n\
#define HAVE_SCHED_H 1\n\
\n\
/* Define if you have sched_getaffinity/sched_setaffinity */\n\
#define HAVE_SCHED_xETAFFINITY 1\n\
\n\
/* Define to 1 if you have the 'sctp_bindx' function. */\n\
/* #undef HAVE_SCTP_BINDX */\n\
\n\
/* Define to 1 if you have the 'sctp_connectx' function. */\n\
/* #undef HAVE_SCTP_CONNECTX */\n\
\n\
/* Define to 1 if you have the 'sctp_freeladdrs' function. */\n\
/* #undef HAVE_SCTP_FREELADDRS */\n\
\n\
/* Define to 1 if you have the 'sctp_freepaddrs' function. */\n\
/* #undef HAVE_SCTP_FREEPADDRS */\n\
\n\
/* Define to 1 if you have the 'sctp_getladdrs' function. */\n\
/* #undef HAVE_SCTP_GETLADDRS */\n\
\n\
/* Define to 1 if you have the 'sctp_getpaddrs' function. */\n\
/* #undef HAVE_SCTP_GETPADDRS */\n\
\n\
/* Define to 1 if you have the <netinet/sctp.h> header file */\n\
#define HAVE_SCTP_H 1\n\
\n\
/* Define to 1 if you have the 'sctp_peeloff' function. */\n\
/* #undef HAVE_SCTP_PEELOFF */\n\
\n\
/* Define to 1 if you have the <sdkddkver.h> header file. */\n\
/* #undef HAVE_SDKDDKVER_H */\n\
\n\
/* Define to 1 if you have the 'sendfile' function. */\n\
#define HAVE_SENDFILE 1\n\
\n\
/* Define to 1 if you have the `sendfilev' function. */\n\
/* #undef HAVE_SENDFILEV */\n\
\n\
/* Define to 1 if you have the 'setlocale' function. */\n\
#define HAVE_SETLOCALE 1\n\
\n\
/* Define to 1 if you have the 'setns' function. */\n\
#define HAVE_SETNS 1\n\
\n\
/* Define to 1 if you have the <setns.h> header file. */\n\
/* #undef HAVE_SETNS_H */\n\
\n\
/* Define to 1 if you have the 'setprotoent' function. */\n\
#define HAVE_SETPROTOENT 1\n\
\n\
/* Define to 1 if you have the 'setsid' function. */\n\
#define HAVE_SETSID 1\n\
\n\
/* Define if we have socklen_t */\n\
#define HAVE_SOCKLEN_T 1\n\
\n\
/* define if you have the Solaris/ultrasparc /dev/perfmon interface */\n\
/* #undef HAVE_SOLARIS_SPARC_PERFMON */\n\
\n\
/* Define if you have SO_BSDCOMPAT flag on sockets */\n\
#define HAVE_SO_BSDCOMPAT /**/\n\
\n\
/* Define to 1 if you have the <stdint.h> header file. */\n\
#define HAVE_STDINT_H 1\n\
\n\
/* Define to 1 if you have the <stdio.h> header file. */\n\
#define HAVE_STDIO_H 1\n\
\n\
/* Define to 1 if you have the <stdlib.h> header file. */\n\
#define HAVE_STDLIB_H 1\n\
\n\
/* Define to 1 if you have the 'strerror' function. */\n\
#define HAVE_STRERROR 1\n\
\n\
/* Define to 1 if you have the 'strerror_r' function. */\n\
#define HAVE_STRERROR_R 1\n\
\n\
/* Define to 1 if you have the 'strftime' function. */\n\
#define HAVE_STRFTIME 1\n\
\n\
/* Define to 1 if you have the <strings.h> header file. */\n\
#define HAVE_STRINGS_H 1\n\
\n\
/* Define to 1 if you have the <string.h> header file. */\n\
#define HAVE_STRING_H 1\n\
\n\
/* Define to 1 if you have the 'strlcat' function. */\n\
#define HAVE_STRLCAT 1\n\
\n\
/* Define to 1 if you have the 'strlcpy' function. */\n\
#define HAVE_STRLCPY 1\n\
\n\
/* Define to 1 if you have the 'strncasecmp' function. */\n\
#define HAVE_STRNCASECMP 1\n\
\n\
/* Define to 1 if 'ifr_enaddr' is a member of 'struct ifreq'. */\n\
/* #undef HAVE_STRUCT_IFREQ_IFR_ENADDR */\n\
\n\
/* Define to 1 if 'ifr_hwaddr' is a member of 'struct ifreq'. */\n\
#define HAVE_STRUCT_IFREQ_IFR_HWADDR 1\n\
\n\
/* Define to 1 if 'ifr_ifindex' is a member of 'struct ifreq'. */\n\
#define HAVE_STRUCT_IFREQ_IFR_IFINDEX 1\n\
\n\
/* Define to 1 if 'ifr_index' is a member of 'struct ifreq'. */\n\
/* #undef HAVE_STRUCT_IFREQ_IFR_INDEX */\n\
\n\
/* Define to 1 if 'ifr_map' is a member of 'struct ifreq'. */\n\
#define HAVE_STRUCT_IFREQ_IFR_MAP 1\n\
\n\
/* Define to 1 if the system has the type 'struct ip_mreqn'. */\n\
#define HAVE_STRUCT_IP_MREQN 1\n\
\n\
/* Define to 1 if 'assoc_id' is a member of 'struct sctp_accoc_value'. */\n\
/* #undef HAVE_STRUCT_SCTP_ACCOC_VALUE_ASSOC_ID */\n\
\n\
/* Define to 1 if 'sctp_authentication_event' is a member of 'struct\n\
   sctp_event_subscribe'. */\n\
#define HAVE_STRUCT_SCTP_EVENT_SUBSCRIBE_SCTP_AUTHENTICATION_EVENT 1\n\
\n\
/* Define to 1 if 'sctp_sender_dry_event' is a member of 'struct\n\
   sctp_event_subscribe'. */\n\
#define HAVE_STRUCT_SCTP_EVENT_SUBSCRIBE_SCTP_SENDER_DRY_EVENT 1\n\
\n\
/* Define to 1 if 'spp_flags' is a member of 'struct sctp_paddrparams'. */\n\
#define HAVE_STRUCT_SCTP_PADDRPARAMS_SPP_FLAGS 1\n\
\n\
/* Define to 1 if 'spp_pathmtu' is a member of 'struct sctp_paddrparams'. */\n\
#define HAVE_STRUCT_SCTP_PADDRPARAMS_SPP_PATHMTU 1\n\
\n\
/* Define to 1 if 'spp_sackdelay' is a member of 'struct sctp_paddrparams'. */\n\
#define HAVE_STRUCT_SCTP_PADDRPARAMS_SPP_SACKDELAY 1\n\
\n\
/* Define to 1 if 'sre_data' is a member of 'struct sctp_remote_error'. */\n\
#define HAVE_STRUCT_SCTP_REMOTE_ERROR_SRE_DATA 1\n\
\n\
/* Define to 1 if 'ssf_data' is a member of 'struct sctp_send_failed'. */\n\
#define HAVE_STRUCT_SCTP_SEND_FAILED_SSF_DATA 1\n\
\n\
/* Define to 1 if 'sdl_len' is a member of 'struct sockaddr_dl'. */\n\
/* #undef HAVE_STRUCT_SOCKADDR_DL_SDL_LEN */\n\
\n\
/* Define to 1 if 'sun_path' is a member of 'struct sockaddr_un'. */\n\
#define HAVE_STRUCT_SOCKADDR_UN_SUN_PATH 1\n\
\n\
/* Define to 1 if you have the <syslog.h> header file. */\n\
#define HAVE_SYSLOG_H 1\n\
\n\
/* Define if you have systemd daemon */\n\
/* #undef HAVE_SYSTEMD_DAEMON */\n\
\n\
/* Define to 1 if you have the <systemd/sd-daemon.h> header file. */\n\
/* #undef HAVE_SYSTEMD_SD_DAEMON_H */\n\
\n\
/* Define if you have <sys/devpoll.h> header file. */\n\
/* #undef HAVE_SYS_DEVPOLL_H */\n\
\n\
/* Define to 1 if you have the <sys/dir.h> header file, and it defines 'DIR'.\n\
   */\n\
/* #undef HAVE_SYS_DIR_H */\n\
\n\
/* Define if you have the <sys/epoll.h> header file. */\n\
#define HAVE_SYS_EPOLL_H 1\n\
\n\
/* Define if you have <sys/event.h> header file. */\n\
/* #undef HAVE_SYS_EVENT_H */\n\
\n\
/* Define to 1 if you have the <sys/ioctl.h> header file. */\n\
#define HAVE_SYS_IOCTL_H 1\n\
\n\
/* Define to 1 if you have the <sys/mman.h> header file. */\n\
#define HAVE_SYS_MMAN_H 1\n\
\n\
/* Define to 1 if you have the <sys/ndir.h> header file, and it defines 'DIR'.\n\
   */\n\
/* #undef HAVE_SYS_NDIR_H */\n\
\n\
/* Define to 1 if you have the <sys/resource.h> header file */\n\
#define HAVE_SYS_RESOURCE_H 1\n\
\n\
/* Define to 1 if you have the <sys/sdt.h> header file. */\n\
/* #undef HAVE_SYS_SDT_H */\n\
\n\
/* Define to 1 if you have the <sys/socketio.h> header file. */\n\
/* #undef HAVE_SYS_SOCKETIO_H */\n\
\n\
/* Define to 1 if you have the <sys/socket.h> header file. */\n\
#define HAVE_SYS_SOCKET_H 1\n\
\n\
/* Define to 1 if you have the <sys/sockio.h> header file. */\n\
/* #undef HAVE_SYS_SOCKIO_H */\n\
\n\
/* Define to 1 if you have the <sys/stat.h> header file. */\n\
#define HAVE_SYS_STAT_H 1\n\
\n\
/* Define to 1 if you have the <sys/stropts.h> header file. */\n\
/* #undef HAVE_SYS_STROPTS_H */\n\
\n\
/* Define to 1 if you have the <sys/sysctl.h> header file. */\n\
/* #undef HAVE_SYS_SYSCTL_H */\n\
\n\
/* Define to 1 if you have the <sys/timerfd.h> header file. */\n\
#define HAVE_SYS_TIMERFD_H 1\n\
\n\
/* Define to 1 if you have the <sys/time.h> header file. */\n\
#define HAVE_SYS_TIME_H 1\n\
\n\
/* Define to 1 if you have the <sys/types.h> header file. */\n\
#define HAVE_SYS_TYPES_H 1\n\
\n\
/* Define to 1 if you have the <sys/uio.h> header file. */\n\
#define HAVE_SYS_UIO_H 1\n\
\n\
/* Define to 1 if you have the <sys/un.h> header file. */\n\
#define HAVE_SYS_UN_H 1\n\
\n\
/* Define to 1 if you have <sys/wait.h> that is POSIX.1 compatible. */\n\
#define HAVE_SYS_WAIT_H 1\n\
\n\
/* Define if termcap functions exists */\n\
#define HAVE_TERMCAP 1\n\
\n\
/* Define to 1 if you have the <termcap.h> header file. */\n\
#define HAVE_TERMCAP_H 1\n\
\n\
/* Define to 1 if you have the <term.h> header file. */\n\
#define HAVE_TERM_H 1\n\
\n\
/* Define to 1 if you have the 'time2posix' function. */\n\
/* #undef HAVE_TIME2POSIX */\n\
\n\
/* Define to 1 if you have the <unistd.h> header file. */\n\
#define HAVE_UNISTD_H 1\n\
\n\
/* Define to 1 if you have the <util.h> header file. */\n\
/* #undef HAVE_UTIL_H */\n\
\n\
/* Define to 1 if you have the <utmp.h> header file. */\n\
#define HAVE_UTMP_H 1\n\
\n\
/* Define to 1 if you have the 'vprintf' function. */\n\
#define HAVE_VPRINTF 1\n\
\n\
/* Define to 1 if you have the 'vsyslog' function. */\n\
#define HAVE_VSYSLOG 1\n\
\n\
/* Define to 1 if you have a `wcwidth' function. */\n\
#define HAVE_WCWIDTH 1\n\
\n\
/* Define to 1 if you have the <windows.h> header file. */\n\
/* #undef HAVE_WINDOWS_H */\n\
\n\
/* Define to 1 if you have the <winsock2.h> header file. */\n\
/* #undef HAVE_WINSOCK2_H */\n\
\n\
/* Define if you have a working posix_openpt implementation */\n\
#define HAVE_WORKING_POSIX_OPENPT 1\n\
\n\
/* Define to 1 if you have the 'writev' function. */\n\
#define HAVE_WRITEV 1\n\
\n\
/* Define to 1 if you have the <ws2tcpip.h> header file. */\n\
/* #undef HAVE_WS2TCPIP_H */\n\
\n\
/* Define if your zlib version defines inflateGetDictionary. */\n\
#define HAVE_ZLIB_INFLATEGETDICTIONARY 1\n\
\n\
/* Define to 1 if you have the '_brk' function. */\n\
/* #undef HAVE__BRK */\n\
\n\
/* Define if you have the '_end' symbol */\n\
#define HAVE__END_SYMBOL 1\n\
\n\
/* Define to 1 if you have the '_sbrk' function. */\n\
/* #undef HAVE__SBRK */\n\
\n\
/* Define to 1 if you have the '__brk' function. */\n\
/* #undef HAVE___BRK */\n\
\n\
/* Define to 1 if you have the '__sbrk' function. */\n\
#define HAVE___SBRK 1\n\
\n\
/* Define to monotonic clock id to use */\n\
#define HRTIME_CLOCK_ID CLOCK_MONOTONIC\n\
\n\
/* Define as a string of monotonic clock id to use */\n\
#define HRTIME_CLOCK_ID_STR \"CLOCK_MONOTONIC\"\n\
\n\
/* define if h_errno is declared (in some way) in a system header file */\n\
#define H_ERRNO_DECLARED 1\n\
\n\
/* Define to monotonic clock id to use */\n\
#define MONOTONIC_CLOCK_ID CLOCK_MONOTONIC\n\
\n\
/* Define as a string of monotonic clock id to use */\n\
#define MONOTONIC_CLOCK_ID_STR \"CLOCK_MONOTONIC\"\n\
\n\
/* Define if we can use the native stack for Erlang code */\n\
#define NATIVE_ERLANG_STACK /**/\n\
\n\
/* Define if netdb.h needs struct sockaddr_in ans in.h CAN be included before\n\
   */\n\
#define NETDB_H_NEEDS_IN_H 1\n\
\n\
/* Defined if no found C compiler can handle jump tables */\n\
/* #undef NO_JUMP_TABLE */\n\
\n\
/* Define if you dont have salen */\n\
#define NO_SA_LEN 1\n\
\n\
/* Define if you want to implement erts_os_monotonic_time() using\n\
   clock_gettime() */\n\
#define OS_MONOTONIC_TIME_USING_CLOCK_GETTIME 1\n\
\n\
/* Define if you want to implement erts_os_monotonic_time() using gethrtime()\n\
   */\n\
/* #undef OS_MONOTONIC_TIME_USING_GETHRTIME */\n\
\n\
/* Define if you want to implement erts_os_monotonic_time() using mach\n\
   clock_get_time() */\n\
/* #undef OS_MONOTONIC_TIME_USING_MACH_CLOCK_GET_TIME */\n\
\n\
/* Define if you want to implement erts_os_monotonic_time() using times() */\n\
/* #undef OS_MONOTONIC_TIME_USING_TIMES */\n\
\n\
/* Define if you want to implement erts_os_system_time() using gettimeofday()\n\
   */\n\
/* #undef OS_SYSTEM_TIME_GETTIMEOFDAY */\n\
\n\
/* Define if you want to implement erts_os_system_time() using clock_gettime()\n\
   */\n\
#define OS_SYSTEM_TIME_USING_CLOCK_GETTIME 1\n\
\n\
/* Define if you want to implement erts_os_system_time() using mach\n\
   clock_get_time() */\n\
/* #undef OS_SYSTEM_TIME_USING_MACH_CLOCK_GET_TIME */\n\
\n\
/* Define to the address where bug reports for this package should be sent. */\n\
#define PACKAGE_BUGREPORT \"\"\n\
\n\
/* Define to the full name of this package. */\n\
#define PACKAGE_NAME \"\"\n\
\n\
/* Define to the full name and version of this package. */\n\
#define PACKAGE_STRING \"\"\n\
\n\
/* Define to the one symbol short name of this package. */\n\
#define PACKAGE_TARNAME \"\"\n\
\n\
/* Define to the home page for this package. */\n\
#define PACKAGE_URL \"\"\n\
\n\
/* Define to the version of this package. */\n\
#define PACKAGE_VERSION \"\"\n\
\n\
/* Define if you wish to redefine FD_SETSIZE to be able to select on more fd\n\
   */\n\
/* #undef REDEFINE_FD_SETSIZE */\n\
\n\
/* Define the sbrk() argument type. */\n\
#define SBRK_ARG_TYPE intptr_t\n\
\n\
/* Define the sbrk() return type. */\n\
#define SBRK_RET_TYPE void *\n\
\n\
/* Define if building a sharing-preserving emulator */\n\
/* #undef SHCOPY */\n\
\n\
/* The size of 'AO_t', as computed by sizeof. */\n\
/* #undef SIZEOF_AO_T */\n\
\n\
/* The size of 'char', as computed by sizeof. */\n\
#define SIZEOF_CHAR 1\n\
\n\
/* The size of 'int', as computed by sizeof. */\n\
#define SIZEOF_INT 4\n\
\n\
/* The size of 'long', as computed by sizeof. */\n\
#define SIZEOF_LONG 8\n\
\n\
/* The size of 'long long', as computed by sizeof. */\n\
#define SIZEOF_LONG_LONG 8\n\
\n\
/* The size of 'off_t', as computed by sizeof. */\n\
#define SIZEOF_OFF_T 8\n\
\n\
/* The size of 'short', as computed by sizeof. */\n\
#define SIZEOF_SHORT 2\n\
\n\
/* The size of 'size_t', as computed by sizeof. */\n\
#define SIZEOF_SIZE_T 8\n\
\n\
/* The size of 'suseconds_t', as computed by sizeof. */\n\
#define SIZEOF_SUSECONDS_T 8\n\
\n\
/* The size of 'time_t', as computed by sizeof. */\n\
#define SIZEOF_TIME_T 8\n\
\n\
/* The size of 'void *', as computed by sizeof. */\n\
#define SIZEOF_VOID_P 8\n\
\n\
/* The size of '_Float16', as computed by sizeof. */\n\
#define SIZEOF__FLOAT16 2\n\
\n\
/* The size of '__int128_t', as computed by sizeof. */\n\
#define SIZEOF___INT128_T 16\n\
\n\
/* The size of '__int64', as computed by sizeof. */\n\
#define SIZEOF___INT64 0\n\
\n\
/* Define to 1 if all of the C89 standard headers exist (not just the ones\n\
   required in a freestanding environment). This macro is provided for\n\
   backward compatibility; new code need not use it. */\n\
#define STDC_HEADERS 1\n\
\n\
/* define if the variable sys_errlist is declared in a system header file */\n\
/* #undef SYS_ERRLIST_DECLARED */\n\
\n\
/* Define if you want to implement erts_os_hrtime() using clock_gettime() */\n\
#define SYS_HRTIME_USING_CLOCK_GETTIME 1\n\
\n\
/* Define if you want to implement erts_os_hrtime() using gethrtime() */\n\
/* #undef SYS_HRTIME_USING_GETHRTIME */\n\
\n\
/* Define if you want to implement erts_os_hrtime() using mach\n\
   clock_get_time() */\n\
/* #undef SYS_HRTIME_USING_MACH_CLOCK_GET_TIME */\n\
\n\
/* Define to 1 if your <sys/time.h> declares 'struct tm'. */\n\
/* #undef TM_IN_SYS_TIME */\n\
\n\
/* Define if you want to use dtrace for dynamic tracing */\n\
/* #undef USE_DTRACE */\n\
\n\
/* Define if you want to use dynamic tracing */\n\
/* #undef USE_DYNAMIC_TRACE */\n\
\n\
/* Define if you want to use lttng for dynamic tracing */\n\
/* #undef USE_LTTNG */\n\
\n\
/* Define if select() should be used instead of poll() */\n\
/* #undef USE_SELECT */\n\
\n\
/* Define if you want to use systemtap for dynamic tracing */\n\
/* #undef USE_SYSTEMTAP */\n\
\n\
/* Define to enable VM dynamic trace probes */\n\
/* #undef USE_VM_PROBES */\n\
\n\
/* Define to wall clock id to use */\n\
#define WALL_CLOCK_ID CLOCK_REALTIME\n\
\n\
/* Define as a string of wall clock id to use */\n\
#define WALL_CLOCK_ID_STR \"CLOCK_REALTIME\"\n\
\n\
/* Define if windows.h includes winsock2.h */\n\
/* #undef WINDOWS_H_INCLUDES_WINSOCK2_H */\n\
\n\
/* Define if big-endian */\n\
/* #undef WORDS_BIGENDIAN */\n\
\n\
/* Number of bits in a file offset, on hosts where this is settable. */\n\
/* #undef _FILE_OFFSET_BITS */\n\
\n\
/* Define to 1 on platforms where this makes off_t a 64-bit type. */\n\
/* #undef _LARGE_FILES */\n\
\n\
/* Number of bits in time_t, on hosts where this is settable. */\n\
/* #undef _TIME_BITS */\n\
\n\
/* Define to 1 on platforms where this makes time_t a 64-bit type. */\n\
/* #undef __MINGW_USE_VC2005_COMPAT */\n\
\n\
/* Define to empty if 'const' does not conform to ANSI C. */\n\
/* #undef const */\n\
\n\
/* Define to 'long int' if <sys/types.h> does not define. */\n\
/* #undef off_t */\n\
\n\
/* Define as a signed integer type capable of holding a process identifier. */\n\
/* #undef pid_t */\n\
\n\
/* Define as 'unsigned int' if <stddef.h> doesn't define. */\n\
/* #undef size_t */\n\
\n\
\n\
/* Redefine in6_addr. XXX this should be moved to the files where it's used? */\n\
#ifdef HAVE_IN_ADDR6_STRUCT\n\
#define in6_addr in_addr6\n\
#endif\n\
\n\
/* Define a reasonable default for INADDR_LOOPBACK */\n\
/* XXX this should be moved to the files where it's used? */\n\
#ifdef HAVE_NO_INADDR_LOOPBACK\n\
#define INADDR_LOOPBACK (u_long)0x7F000001\n\
#endif\n\
\n\
#ifdef REDEFINE_FD_SETSIZE\n\
#define FD_SETSIZE 1024\n\
#endif\n\
 \n\
#ifdef HAVE_GETHRVTIME_PROCFS_IOCTL\n\
#define HAVE_GETHRVTIME\n\
#endif\n\
\n\
#if !defined(HAVE_ISFINITE) && !defined(HAVE_FINITE)\n\
# if defined(HAVE_ISINF) && defined(HAVE_ISNAN)\n\
#  define USE_ISINF_ISNAN\n\
# endif\n\
#endif\n\
\n\
#if defined(DEBUG) && !defined(ERTS_ENABLE_LOCK_CHECK)\n\
#define ERTS_ENABLE_LOCK_CHECK 1\n\
#endif\n\
\n\
#endif /* __ERTS_CONFIG_H__ */\n\
\n\
";
const char* erts_build_flags_LDFLAGS = "-Wl,--no-copy-dt-needed-entries -Wl,--as-needed -Wl,-z,now -Wl,-z,relro -Wl,-z,noexecstack -fstack-protector-strong -Wl,-O1 -Wl,--sort-common -Wl,--as-needed -Wl,-z,relro -Wl,-z,now -Wl,-z,pack-relative-relocs -flto=auto";
