/*
 *  Warning: Do *not* edit this file. It was automatically
 *  generated by 'make_tables'.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif /* HAVE_CONFIG_H */
#include "export.h"
#include "sys.h"
#include "erl_vm.h"
#include "erl_process.h"
#include "bif.h"
#include "erl_bif_table.h"
#include "erl_atom_table.h"


Export bif_trap_exports__[BIF_SIZE];
BifEntry bif_table[] = {
  {am_erlang, am_abs, 1, abs_1, abs_1, BIF_KIND_GUARD},
  {am_erlang, am_adler32, 1, adler32_1, adler32_1, BIF_KIND_REGULAR},
  {am_erlang, am_adler32, 2, adler32_2, adler32_2, BIF_KIND_REGULAR},
  {am_erlang, am_adler32_combine, 3, adler32_combine_3, adler32_combine_3, BIF_KIND_REGULAR},
  {am_erlang, am_atom_to_list, 1, atom_to_list_1, atom_to_list_1, BIF_KIND_REGULAR},
  {am_erlang, am_binary_to_list, 1, binary_to_list_1, binary_to_list_1, BIF_KIND_REGULAR},
  {am_erlang, am_binary_to_list, 3, binary_to_list_3, binary_to_list_3, BIF_KIND_REGULAR},
  {am_erlang, am_binary_to_term, 1, binary_to_term_1, binary_to_term_1, BIF_KIND_REGULAR},
  {am_erlang, am_crc32, 1, crc32_1, crc32_1, BIF_KIND_REGULAR},
  {am_erlang, am_crc32, 2, crc32_2, crc32_2, BIF_KIND_REGULAR},
  {am_erlang, am_crc32_combine, 3, crc32_combine_3, crc32_combine_3, BIF_KIND_REGULAR},
  {am_erlang, am_date, 0, date_0, date_0, BIF_KIND_REGULAR},
  {am_erlang, am_delete_module, 1, delete_module_1, delete_module_1, BIF_KIND_REGULAR},
  {am_erlang, am_display, 1, display_1, display_1, BIF_KIND_REGULAR},
  {am_erlang, am_display_string, 2, schedule_dirty_cpu_display_string_2, display_string_2, BIF_KIND_REGULAR},
  {am_erlang, am_element, 2, element_2, element_2, BIF_KIND_GUARD},
  {am_erlang, am_erase, 0, erase_0, erase_0, BIF_KIND_REGULAR},
  {am_erlang, am_erase, 1, erase_1, erase_1, BIF_KIND_HEAVY},
  {am_erlang, am_exit, 1, exit_1, exit_1, BIF_KIND_REGULAR},
  {am_erlang, am_exit, 2, exit_2, exit_2, BIF_KIND_REGULAR},
  {am_erlang, am_exit_signal, 2, exit_signal_2, exit_signal_2, BIF_KIND_REGULAR},
  {am_erlang, am_external_size, 1, external_size_1, external_size_1, BIF_KIND_REGULAR},
  {am_erlang, am_external_size, 2, external_size_2, external_size_2, BIF_KIND_REGULAR},
  {am_erlang, am_float, 1, float_1, float_1, BIF_KIND_GUARD},
  {am_erlang, am_float_to_list, 1, float_to_list_1, float_to_list_1, BIF_KIND_REGULAR},
  {am_erlang, am_float_to_list, 2, float_to_list_2, float_to_list_2, BIF_KIND_REGULAR},
  {am_erlang, am_fun_info, 2, fun_info_2, fun_info_2, BIF_KIND_REGULAR},
  {am_erts_internal, am_garbage_collect, 1, erts_internal_garbage_collect_1, erts_internal_garbage_collect_1, BIF_KIND_HEAVY},
  {am_erlang, am_get, 0, get_0, get_0, BIF_KIND_REGULAR},
  {am_erlang, am_get, 1, get_1, get_1, BIF_KIND_REGULAR},
  {am_erlang, am_get_keys, 1, get_keys_1, get_keys_1, BIF_KIND_REGULAR},
  {am_erlang, am_group_leader, 0, group_leader_0, group_leader_0, BIF_KIND_REGULAR},
  {am_erts_internal, am_group_leader, 2, erts_internal_group_leader_2, erts_internal_group_leader_2, BIF_KIND_REGULAR},
  {am_erts_internal, am_group_leader, 3, erts_internal_group_leader_3, erts_internal_group_leader_3, BIF_KIND_REGULAR},
  {am_erlang, am_halt, 2, halt_2, halt_2, BIF_KIND_REGULAR},
  {am_erlang, am_phash, 2, phash_2, phash_2, BIF_KIND_REGULAR},
  {am_erlang, am_phash2, 1, phash2_1, phash2_1, BIF_KIND_REGULAR},
  {am_erlang, am_phash2, 2, phash2_2, phash2_2, BIF_KIND_REGULAR},
  {am_erlang, am_hd, 1, hd_1, hd_1, BIF_KIND_GUARD},
  {am_erlang, am_integer_to_list, 1, integer_to_list_1, integer_to_list_1, BIF_KIND_REGULAR},
  {am_erlang, am_length, 1, length_1, length_1, BIF_KIND_GUARD},
  {am_erlang, am_link, 1, link_1, link_1, BIF_KIND_REGULAR},
  {am_erlang, am_list_to_atom, 1, list_to_atom_1, list_to_atom_1, BIF_KIND_GUARD},
  {am_erlang, am_list_to_binary, 1, list_to_binary_1, list_to_binary_1, BIF_KIND_REGULAR},
  {am_erlang, am_list_to_float, 1, list_to_float_1, list_to_float_1, BIF_KIND_REGULAR},
  {am_erlang, am_list_to_pid, 1, list_to_pid_1, list_to_pid_1, BIF_KIND_REGULAR},
  {am_erlang, am_list_to_port, 1, list_to_port_1, list_to_port_1, BIF_KIND_REGULAR},
  {am_erlang, am_list_to_ref, 1, list_to_ref_1, list_to_ref_1, BIF_KIND_REGULAR},
  {am_erlang, am_list_to_tuple, 1, list_to_tuple_1, list_to_tuple_1, BIF_KIND_REGULAR},
  {am_erlang, am_loaded, 0, loaded_0, loaded_0, BIF_KIND_REGULAR},
  {am_erlang, am_localtime, 0, localtime_0, localtime_0, BIF_KIND_REGULAR},
  {am_erlang, am_localtime_to_universaltime, 2, localtime_to_universaltime_2, localtime_to_universaltime_2, BIF_KIND_REGULAR},
  {am_erlang, am_make_ref, 0, make_ref_0, make_ref_0, BIF_KIND_REGULAR},
  {am_erlang, am_unique_integer, 0, unique_integer_0, unique_integer_0, BIF_KIND_REGULAR},
  {am_erlang, am_unique_integer, 1, unique_integer_1, unique_integer_1, BIF_KIND_REGULAR},
  {am_erlang, am_md5, 1, md5_1, md5_1, BIF_KIND_REGULAR},
  {am_erlang, am_md5_init, 0, md5_init_0, md5_init_0, BIF_KIND_REGULAR},
  {am_erlang, am_md5_update, 2, md5_update_2, md5_update_2, BIF_KIND_REGULAR},
  {am_erlang, am_md5_final, 1, md5_final_1, md5_final_1, BIF_KIND_REGULAR},
  {am_erlang, am_module_loaded, 1, module_loaded_1, module_loaded_1, BIF_KIND_REGULAR},
  {am_erlang, am_function_exported, 3, function_exported_3, function_exported_3, BIF_KIND_REGULAR},
  {am_erlang, am_monitor_node, 2, monitor_node_2, monitor_node_2, BIF_KIND_REGULAR},
  {am_erlang, am_monitor_node, 3, monitor_node_3, monitor_node_3, BIF_KIND_REGULAR},
  {am_erlang, am_node, 1, node_1, node_1, BIF_KIND_GUARD},
  {am_erlang, am_node, 0, node_0, node_0, BIF_KIND_GUARD},
  {am_erlang, am_nodes, 0, nodes_0, nodes_0, BIF_KIND_REGULAR},
  {am_erlang, am_nodes, 1, nodes_1, nodes_1, BIF_KIND_REGULAR},
  {am_erlang, am_nodes, 2, nodes_2, nodes_2, BIF_KIND_REGULAR},
  {am_erlang, am_now, 0, now_0, now_0, BIF_KIND_REGULAR},
  {am_erlang, am_monotonic_time, 0, monotonic_time_0, monotonic_time_0, BIF_KIND_REGULAR},
  {am_erlang, am_monotonic_time, 1, monotonic_time_1, monotonic_time_1, BIF_KIND_REGULAR},
  {am_erlang, am_system_time, 0, system_time_0, system_time_0, BIF_KIND_REGULAR},
  {am_erlang, am_system_time, 1, system_time_1, system_time_1, BIF_KIND_REGULAR},
  {am_erlang, am_time_offset, 0, time_offset_0, time_offset_0, BIF_KIND_REGULAR},
  {am_erlang, am_time_offset, 1, time_offset_1, time_offset_1, BIF_KIND_REGULAR},
  {am_erlang, am_timestamp, 0, timestamp_0, timestamp_0, BIF_KIND_REGULAR},
  {am_erts_internal, am_open_port, 2, erts_internal_open_port_2, erts_internal_open_port_2, BIF_KIND_REGULAR},
  {am_erlang, am_pid_to_list, 1, pid_to_list_1, pid_to_list_1, BIF_KIND_REGULAR},
  {am_erlang, am_ports, 0, ports_0, ports_0, BIF_KIND_REGULAR},
  {am_erlang, am_pre_loaded, 0, pre_loaded_0, pre_loaded_0, BIF_KIND_REGULAR},
  {am_erlang, am_process_flag, 2, process_flag_2, process_flag_2, BIF_KIND_REGULAR},
  {am_erts_internal, am_process_flag, 3, erts_internal_process_flag_3, erts_internal_process_flag_3, BIF_KIND_REGULAR},
  {am_erlang, am_process_info, 1, process_info_1, process_info_1, BIF_KIND_HEAVY},
  {am_erlang, am_process_info, 2, process_info_2, process_info_2, BIF_KIND_HEAVY},
  {am_erlang, am_processes, 0, processes_0, processes_0, BIF_KIND_REGULAR},
  {am_erlang, am_put, 2, put_2, put_2, BIF_KIND_HEAVY},
  {am_erlang, am_register, 2, register_2, register_2, BIF_KIND_REGULAR},
  {am_erlang, am_registered, 0, registered_0, registered_0, BIF_KIND_REGULAR},
  {am_erlang, am_round, 1, round_1, round_1, BIF_KIND_GUARD},
  {am_erlang, am_self, 0, self_0, self_0, BIF_KIND_GUARD},
  {am_erlang, am_setelement, 3, setelement_3, setelement_3, BIF_KIND_REGULAR},
  {am_erlang, am_size, 1, size_1, size_1, BIF_KIND_GUARD},
  {am_erlang, am_spawn, 3, spawn_3, spawn_3, BIF_KIND_REGULAR},
  {am_erlang, am_spawn_link, 3, spawn_link_3, spawn_link_3, BIF_KIND_REGULAR},
  {am_erlang, am_split_binary, 2, split_binary_2, split_binary_2, BIF_KIND_REGULAR},
  {am_erlang, am_statistics, 1, statistics_1, statistics_1, BIF_KIND_REGULAR},
  {am_erlang, am_term_to_binary, 1, term_to_binary_1, term_to_binary_1, BIF_KIND_REGULAR},
  {am_erlang, am_term_to_binary, 2, term_to_binary_2, term_to_binary_2, BIF_KIND_REGULAR},
  {am_erlang, am_term_to_iovec, 1, term_to_iovec_1, term_to_iovec_1, BIF_KIND_REGULAR},
  {am_erlang, am_term_to_iovec, 2, term_to_iovec_2, term_to_iovec_2, BIF_KIND_REGULAR},
  {am_erlang, am_throw, 1, throw_1, throw_1, BIF_KIND_REGULAR},
  {am_erlang, am_time, 0, time_0, time_0, BIF_KIND_REGULAR},
  {am_erlang, am_tl, 1, tl_1, tl_1, BIF_KIND_GUARD},
  {am_erlang, am_trunc, 1, trunc_1, trunc_1, BIF_KIND_GUARD},
  {am_erlang, am_tuple_to_list, 1, tuple_to_list_1, tuple_to_list_1, BIF_KIND_REGULAR},
  {am_erlang, am_universaltime, 0, universaltime_0, universaltime_0, BIF_KIND_REGULAR},
  {am_erlang, am_universaltime_to_localtime, 1, universaltime_to_localtime_1, universaltime_to_localtime_1, BIF_KIND_REGULAR},
  {am_erlang, am_unlink, 1, unlink_1, unlink_1, BIF_KIND_REGULAR},
  {am_erlang, am_unregister, 1, unregister_1, unregister_1, BIF_KIND_REGULAR},
  {am_erlang, am_whereis, 1, whereis_1, whereis_1, BIF_KIND_REGULAR},
  {am_erlang, am_spawn_opt, 4, spawn_opt_4, spawn_opt_4, BIF_KIND_REGULAR},
  {am_erlang, am_setnode, 2, setnode_2, setnode_2, BIF_KIND_REGULAR},
  {am_erlang, am_dist_get_stat, 1, dist_get_stat_1, dist_get_stat_1, BIF_KIND_REGULAR},
  {am_erlang, am_dist_ctrl_input_handler, 2, dist_ctrl_input_handler_2, dist_ctrl_input_handler_2, BIF_KIND_REGULAR},
  {am_erlang, am_dist_ctrl_put_data, 2, dist_ctrl_put_data_2, dist_ctrl_put_data_2, BIF_KIND_REGULAR},
  {am_erlang, am_dist_ctrl_get_data, 1, dist_ctrl_get_data_1, dist_ctrl_get_data_1, BIF_KIND_REGULAR},
  {am_erlang, am_dist_ctrl_get_data_notification, 1, dist_ctrl_get_data_notification_1, dist_ctrl_get_data_notification_1, BIF_KIND_REGULAR},
  {am_erlang, am_dist_ctrl_get_opt, 2, dist_ctrl_get_opt_2, dist_ctrl_get_opt_2, BIF_KIND_REGULAR},
  {am_erlang, am_dist_ctrl_set_opt, 3, dist_ctrl_set_opt_3, dist_ctrl_set_opt_3, BIF_KIND_REGULAR},
  {am_erts_internal, am_port_info, 1, erts_internal_port_info_1, erts_internal_port_info_1, BIF_KIND_REGULAR},
  {am_erts_internal, am_port_info, 2, erts_internal_port_info_2, erts_internal_port_info_2, BIF_KIND_REGULAR},
  {am_erts_internal, am_port_call, 3, erts_internal_port_call_3, erts_internal_port_call_3, BIF_KIND_REGULAR},
  {am_erts_internal, am_port_command, 3, erts_internal_port_command_3, erts_internal_port_command_3, BIF_KIND_REGULAR},
  {am_erts_internal, am_port_control, 3, erts_internal_port_control_3, erts_internal_port_control_3, BIF_KIND_REGULAR},
  {am_erts_internal, am_port_close, 1, erts_internal_port_close_1, erts_internal_port_close_1, BIF_KIND_REGULAR},
  {am_erts_internal, am_port_connect, 2, erts_internal_port_connect_2, erts_internal_port_connect_2, BIF_KIND_REGULAR},
  {am_erts_internal, am_request_system_task, 3, erts_internal_request_system_task_3, erts_internal_request_system_task_3, BIF_KIND_REGULAR},
  {am_erts_internal, am_request_system_task, 4, erts_internal_request_system_task_4, erts_internal_request_system_task_4, BIF_KIND_REGULAR},
  {am_erts_internal, am_check_process_code, 1, erts_internal_check_process_code_1, erts_internal_check_process_code_1, BIF_KIND_HEAVY},
  {am_erts_internal, am_map_to_tuple_keys, 1, erts_internal_map_to_tuple_keys_1, erts_internal_map_to_tuple_keys_1, BIF_KIND_REGULAR},
  {am_erts_internal, am_term_type, 1, erts_internal_term_type_1, erts_internal_term_type_1, BIF_KIND_REGULAR},
  {am_erts_internal, am_map_hashmap_children, 1, erts_internal_map_hashmap_children_1, erts_internal_map_hashmap_children_1, BIF_KIND_REGULAR},
  {am_erts_internal, am_time_unit, 0, erts_internal_time_unit_0, erts_internal_time_unit_0, BIF_KIND_REGULAR},
  {am_erts_internal, am_perf_counter_unit, 0, erts_internal_perf_counter_unit_0, erts_internal_perf_counter_unit_0, BIF_KIND_REGULAR},
  {am_erts_internal, am_is_system_process, 1, erts_internal_is_system_process_1, erts_internal_is_system_process_1, BIF_KIND_REGULAR},
  {am_erts_internal, am_system_check, 1, erts_internal_system_check_1, erts_internal_system_check_1, BIF_KIND_REGULAR},
  {am_erts_internal, am_scheduler_wall_time, 1, erts_internal_scheduler_wall_time_1, erts_internal_scheduler_wall_time_1, BIF_KIND_REGULAR},
  {am_erts_internal, am_dirty_process_handle_signals, 1, erts_internal_dirty_process_handle_signals_1, erts_internal_dirty_process_handle_signals_1, BIF_KIND_REGULAR},
  {am_erts_internal, am_create_dist_channel, 3, erts_internal_create_dist_channel_3, erts_internal_create_dist_channel_3, BIF_KIND_REGULAR},
  {am_erts_internal, am_ets_super_user, 1, erts_internal_ets_super_user_1, erts_internal_ets_super_user_1, BIF_KIND_REGULAR},
  {am_erts_internal, am_spawn_request, 4, erts_internal_spawn_request_4, erts_internal_spawn_request_4, BIF_KIND_REGULAR},
  {am_erts_internal, am_dist_spawn_request, 4, erts_internal_dist_spawn_request_4, erts_internal_dist_spawn_request_4, BIF_KIND_REGULAR},
  {am_erts_internal, am_no_aux_work_threads, 0, erts_internal_no_aux_work_threads_0, erts_internal_no_aux_work_threads_0, BIF_KIND_REGULAR},
  {am_erlang, am_spawn_request_abandon, 1, spawn_request_abandon_1, spawn_request_abandon_1, BIF_KIND_REGULAR},
  {am_erts_literal_area_collector, am_release_area_switch, 0, erts_literal_area_collector_release_area_switch_0, erts_literal_area_collector_release_area_switch_0, BIF_KIND_REGULAR},
  {am_erts_literal_area_collector, am_send_copy_request, 3, erts_literal_area_collector_send_copy_request_3, erts_literal_area_collector_send_copy_request_3, BIF_KIND_REGULAR},
  {am_erlang, am_port_set_data, 2, port_set_data_2, port_set_data_2, BIF_KIND_REGULAR},
  {am_erlang, am_port_get_data, 1, port_get_data_1, port_get_data_1, BIF_KIND_REGULAR},
  {am_erts_internal, am_trace_pattern, 3, erts_internal_trace_pattern_3, erts_internal_trace_pattern_3, BIF_KIND_REGULAR},
  {am_erts_internal, am_trace, 3, erts_internal_trace_3, erts_internal_trace_3, BIF_KIND_REGULAR},
  {am_erlang, am_trace_info, 2, trace_info_2, trace_info_2, BIF_KIND_REGULAR},
  {am_erlang, am_trace_delivered, 1, trace_delivered_1, trace_delivered_1, BIF_KIND_REGULAR},
  {am_erlang, am_seq_trace, 2, seq_trace_2, seq_trace_2, BIF_KIND_REGULAR},
  {am_erlang, am_seq_trace_info, 1, seq_trace_info_1, seq_trace_info_1, BIF_KIND_REGULAR},
  {am_erlang, am_seq_trace_print, 1, seq_trace_print_1, seq_trace_print_1, BIF_KIND_REGULAR},
  {am_erlang, am_seq_trace_print, 2, seq_trace_print_2, seq_trace_print_2, BIF_KIND_REGULAR},
  {am_erts_internal, am_suspend_process, 2, erts_internal_suspend_process_2, erts_internal_suspend_process_2, BIF_KIND_REGULAR},
  {am_erlang, am_resume_process, 1, resume_process_1, resume_process_1, BIF_KIND_REGULAR},
  {am_erts_internal, am_process_display, 2, erts_internal_process_display_2, erts_internal_process_display_2, BIF_KIND_REGULAR},
  {am_erlang, am_bump_reductions, 1, bump_reductions_1, bump_reductions_1, BIF_KIND_REGULAR},
  {am_math, am_cos, 1, math_cos_1, math_cos_1, BIF_KIND_REGULAR},
  {am_math, am_cosh, 1, math_cosh_1, math_cosh_1, BIF_KIND_REGULAR},
  {am_math, am_sin, 1, math_sin_1, math_sin_1, BIF_KIND_REGULAR},
  {am_math, am_sinh, 1, math_sinh_1, math_sinh_1, BIF_KIND_REGULAR},
  {am_math, am_tan, 1, math_tan_1, math_tan_1, BIF_KIND_REGULAR},
  {am_math, am_tanh, 1, math_tanh_1, math_tanh_1, BIF_KIND_REGULAR},
  {am_math, am_acos, 1, math_acos_1, math_acos_1, BIF_KIND_REGULAR},
  {am_math, am_acosh, 1, math_acosh_1, math_acosh_1, BIF_KIND_REGULAR},
  {am_math, am_asin, 1, math_asin_1, math_asin_1, BIF_KIND_REGULAR},
  {am_math, am_asinh, 1, math_asinh_1, math_asinh_1, BIF_KIND_REGULAR},
  {am_math, am_atan, 1, math_atan_1, math_atan_1, BIF_KIND_REGULAR},
  {am_math, am_atanh, 1, math_atanh_1, math_atanh_1, BIF_KIND_REGULAR},
  {am_math, am_erf, 1, math_erf_1, math_erf_1, BIF_KIND_REGULAR},
  {am_math, am_erfc, 1, math_erfc_1, math_erfc_1, BIF_KIND_REGULAR},
  {am_math, am_exp, 1, math_exp_1, math_exp_1, BIF_KIND_REGULAR},
  {am_math, am_log, 1, math_log_1, math_log_1, BIF_KIND_REGULAR},
  {am_math, am_log2, 1, math_log2_1, math_log2_1, BIF_KIND_REGULAR},
  {am_math, am_log10, 1, math_log10_1, math_log10_1, BIF_KIND_REGULAR},
  {am_math, am_sqrt, 1, math_sqrt_1, math_sqrt_1, BIF_KIND_REGULAR},
  {am_math, am_atan2, 2, math_atan2_2, math_atan2_2, BIF_KIND_REGULAR},
  {am_math, am_pow, 2, math_pow_2, math_pow_2, BIF_KIND_REGULAR},
  {am_erlang, am_start_timer, 3, start_timer_3, start_timer_3, BIF_KIND_REGULAR},
  {am_erlang, am_start_timer, 4, start_timer_4, start_timer_4, BIF_KIND_REGULAR},
  {am_erlang, am_send_after, 3, send_after_3, send_after_3, BIF_KIND_REGULAR},
  {am_erlang, am_send_after, 4, send_after_4, send_after_4, BIF_KIND_REGULAR},
  {am_erlang, am_cancel_timer, 1, cancel_timer_1, cancel_timer_1, BIF_KIND_REGULAR},
  {am_erlang, am_cancel_timer, 2, cancel_timer_2, cancel_timer_2, BIF_KIND_REGULAR},
  {am_erlang, am_read_timer, 1, read_timer_1, read_timer_1, BIF_KIND_REGULAR},
  {am_erlang, am_read_timer, 2, read_timer_2, read_timer_2, BIF_KIND_REGULAR},
  {am_erlang, am_make_tuple, 2, make_tuple_2, make_tuple_2, BIF_KIND_REGULAR},
  {am_erlang, am_append_element, 2, append_element_2, append_element_2, BIF_KIND_REGULAR},
  {am_erlang, am_make_tuple, 3, make_tuple_3, make_tuple_3, BIF_KIND_REGULAR},
  {am_erlang, am_system_flag, 2, system_flag_2, system_flag_2, BIF_KIND_REGULAR},
  {am_erlang, am_system_info, 1, system_info_1, system_info_1, BIF_KIND_REGULAR},
  {am_erlang, am_system_profile, 2, system_profile_2, system_profile_2, BIF_KIND_REGULAR},
  {am_erlang, am_system_profile, 0, system_profile_0, system_profile_0, BIF_KIND_REGULAR},
  {am_erlang, am_ref_to_list, 1, ref_to_list_1, ref_to_list_1, BIF_KIND_REGULAR},
  {am_erlang, am_port_to_list, 1, port_to_list_1, port_to_list_1, BIF_KIND_REGULAR},
  {am_erlang, am_fun_to_list, 1, fun_to_list_1, fun_to_list_1, BIF_KIND_REGULAR},
  {am_erlang, am_monitor, 2, monitor_2, monitor_2, BIF_KIND_REGULAR},
  {am_erlang, am_demonitor, 1, demonitor_1, demonitor_1, BIF_KIND_REGULAR},
  {am_erlang, am_demonitor, 2, demonitor_2, demonitor_2, BIF_KIND_REGULAR},
  {am_erlang, am_is_process_alive, 1, is_process_alive_1, is_process_alive_1, BIF_KIND_REGULAR},
  {am_erts_internal, am_is_process_alive, 2, erts_internal_is_process_alive_2, erts_internal_is_process_alive_2, BIF_KIND_REGULAR},
  {am_erlang, am_error, 1, error_1, error_1, BIF_KIND_REGULAR},
  {am_erlang, am_error, 2, error_2, error_2, BIF_KIND_REGULAR},
  {am_erlang, am_raise, 3, raise_3, raise_3, BIF_KIND_REGULAR},
  {am_erlang, am_is_builtin, 3, is_builtin_3, is_builtin_3, BIF_KIND_REGULAR},
  {am_erlang, am_and, 2, and_2, and_2, BIF_KIND_GUARD},
  {am_erlang, am_or, 2, or_2, or_2, BIF_KIND_GUARD},
  {am_erlang, am_xor, 2, xor_2, xor_2, BIF_KIND_GUARD},
  {am_erlang, am_not, 1, not_1, not_1, BIF_KIND_GUARD},
  {am_erlang, am_Gt, 2, sgt_2, sgt_2, BIF_KIND_GUARD},
  {am_erlang, am_Ge, 2, sge_2, sge_2, BIF_KIND_GUARD},
  {am_erlang, am_Lt, 2, slt_2, slt_2, BIF_KIND_GUARD},
  {am_erlang, am_Le, 2, sle_2, sle_2, BIF_KIND_GUARD},
  {am_erlang, am_Eq, 2, seq_2, seq_2, BIF_KIND_GUARD},
  {am_erlang, am_Eqeq, 2, seqeq_2, seqeq_2, BIF_KIND_GUARD},
  {am_erlang, am_Neq, 2, sneq_2, sneq_2, BIF_KIND_GUARD},
  {am_erlang, am_Neqeq, 2, sneqeq_2, sneqeq_2, BIF_KIND_GUARD},
  {am_erlang, am_Plus, 2, splus_2, splus_2, BIF_KIND_GUARD},
  {am_erlang, am_Minus, 2, sminus_2, sminus_2, BIF_KIND_GUARD},
  {am_erlang, am_Times, 2, stimes_2, stimes_2, BIF_KIND_GUARD},
  {am_erlang, am_Div, 2, div_2, div_2, BIF_KIND_GUARD},
  {am_erlang, am_div, 2, intdiv_2, intdiv_2, BIF_KIND_GUARD},
  {am_erlang, am_rem, 2, rem_2, rem_2, BIF_KIND_GUARD},
  {am_erlang, am_bor, 2, bor_2, bor_2, BIF_KIND_GUARD},
  {am_erlang, am_band, 2, band_2, band_2, BIF_KIND_GUARD},
  {am_erlang, am_bxor, 2, bxor_2, bxor_2, BIF_KIND_GUARD},
  {am_erlang, am_bsl, 2, bsl_2, bsl_2, BIF_KIND_GUARD},
  {am_erlang, am_bsr, 2, bsr_2, bsr_2, BIF_KIND_GUARD},
  {am_erlang, am_bnot, 1, bnot_1, bnot_1, BIF_KIND_GUARD},
  {am_erlang, am_Minus, 1, sminus_1, sminus_1, BIF_KIND_GUARD},
  {am_erlang, am_Plus, 1, splus_1, splus_1, BIF_KIND_GUARD},
  {am_erlang, am__AtomAlias26, 2, ebif_bang_2, ebif_bang_2, BIF_KIND_REGULAR},
  {am_erlang, am_send, 2, send_2, send_2, BIF_KIND_REGULAR},
  {am_erlang, am_send, 3, send_3, send_3, BIF_KIND_REGULAR},
  {am_erlang, am_PlusPlus, 2, ebif_plusplus_2, ebif_plusplus_2, BIF_KIND_REGULAR},
  {am_erlang, am_append, 2, append_2, append_2, BIF_KIND_REGULAR},
  {am_erlang, am_MinusMinus, 2, ebif_minusminus_2, ebif_minusminus_2, BIF_KIND_REGULAR},
  {am_erlang, am_subtract, 2, subtract_2, subtract_2, BIF_KIND_REGULAR},
  {am_erlang, am_is_atom, 1, is_atom_1, is_atom_1, BIF_KIND_GUARD},
  {am_erlang, am_is_list, 1, is_list_1, is_list_1, BIF_KIND_GUARD},
  {am_erlang, am_is_tuple, 1, is_tuple_1, is_tuple_1, BIF_KIND_GUARD},
  {am_erlang, am_is_float, 1, is_float_1, is_float_1, BIF_KIND_GUARD},
  {am_erlang, am_is_integer, 1, is_integer_1, is_integer_1, BIF_KIND_GUARD},
  {am_erlang, am_is_number, 1, is_number_1, is_number_1, BIF_KIND_GUARD},
  {am_erlang, am_is_pid, 1, is_pid_1, is_pid_1, BIF_KIND_GUARD},
  {am_erlang, am_is_port, 1, is_port_1, is_port_1, BIF_KIND_GUARD},
  {am_erlang, am_is_reference, 1, is_reference_1, is_reference_1, BIF_KIND_GUARD},
  {am_erlang, am_is_binary, 1, is_binary_1, is_binary_1, BIF_KIND_GUARD},
  {am_erlang, am_is_function, 1, is_function_1, is_function_1, BIF_KIND_GUARD},
  {am_erlang, am_is_function, 2, is_function_2, is_function_2, BIF_KIND_GUARD},
  {am_erlang, am_is_record, 2, is_record_2, is_record_2, BIF_KIND_GUARD},
  {am_erlang, am_is_record, 3, is_record_3, is_record_3, BIF_KIND_GUARD},
  {am_erlang, am_match_spec_test, 3, match_spec_test_3, match_spec_test_3, BIF_KIND_REGULAR},
  {am_ets, am_internal_request_all, 0, ets_internal_request_all_0, ets_internal_request_all_0, BIF_KIND_REGULAR},
  {am_ets, am_new, 2, ets_new_2, ets_new_2, BIF_KIND_REGULAR},
  {am_ets, am_delete, 1, ets_delete_1, ets_delete_1, BIF_KIND_REGULAR},
  {am_ets, am_delete, 2, ets_delete_2, ets_delete_2, BIF_KIND_REGULAR},
  {am_ets, am_delete_object, 2, ets_delete_object_2, ets_delete_object_2, BIF_KIND_REGULAR},
  {am_ets, am_first, 1, ets_first_1, ets_first_1, BIF_KIND_REGULAR},
  {am_ets, am_first_lookup, 1, ets_first_lookup_1, ets_first_lookup_1, BIF_KIND_REGULAR},
  {am_ets, am_is_compiled_ms, 1, ets_is_compiled_ms_1, ets_is_compiled_ms_1, BIF_KIND_REGULAR},
  {am_ets, am_lookup, 2, ets_lookup_2, ets_lookup_2, BIF_KIND_REGULAR},
  {am_ets, am_lookup_element, 3, ets_lookup_element_3, ets_lookup_element_3, BIF_KIND_REGULAR},
  {am_ets, am_lookup_element, 4, ets_lookup_element_4, ets_lookup_element_4, BIF_KIND_REGULAR},
  {am_ets, am_info, 1, ets_info_1, ets_info_1, BIF_KIND_REGULAR},
  {am_ets, am_info, 2, ets_info_2, ets_info_2, BIF_KIND_REGULAR},
  {am_ets, am_last, 1, ets_last_1, ets_last_1, BIF_KIND_REGULAR},
  {am_ets, am_last_lookup, 1, ets_last_lookup_1, ets_last_lookup_1, BIF_KIND_REGULAR},
  {am_ets, am_match, 1, ets_match_1, ets_match_1, BIF_KIND_REGULAR},
  {am_ets, am_match, 2, ets_match_2, ets_match_2, BIF_KIND_REGULAR},
  {am_ets, am_match, 3, ets_match_3, ets_match_3, BIF_KIND_REGULAR},
  {am_ets, am_match_object, 1, ets_match_object_1, ets_match_object_1, BIF_KIND_REGULAR},
  {am_ets, am_match_object, 2, ets_match_object_2, ets_match_object_2, BIF_KIND_REGULAR},
  {am_ets, am_match_object, 3, ets_match_object_3, ets_match_object_3, BIF_KIND_REGULAR},
  {am_ets, am_member, 2, ets_member_2, ets_member_2, BIF_KIND_REGULAR},
  {am_ets, am_next, 2, ets_next_2, ets_next_2, BIF_KIND_REGULAR},
  {am_ets, am_next_lookup, 2, ets_next_lookup_2, ets_next_lookup_2, BIF_KIND_REGULAR},
  {am_ets, am_prev, 2, ets_prev_2, ets_prev_2, BIF_KIND_REGULAR},
  {am_ets, am_prev_lookup, 2, ets_prev_lookup_2, ets_prev_lookup_2, BIF_KIND_REGULAR},
  {am_ets, am_insert, 2, ets_insert_2, ets_insert_2, BIF_KIND_REGULAR},
  {am_ets, am_insert_new, 2, ets_insert_new_2, ets_insert_new_2, BIF_KIND_REGULAR},
  {am_ets, am_rename, 2, ets_rename_2, ets_rename_2, BIF_KIND_REGULAR},
  {am_ets, am_safe_fixtable, 2, ets_safe_fixtable_2, ets_safe_fixtable_2, BIF_KIND_REGULAR},
  {am_ets, am_slot, 2, ets_slot_2, ets_slot_2, BIF_KIND_REGULAR},
  {am_ets, am_update_counter, 3, ets_update_counter_3, ets_update_counter_3, BIF_KIND_REGULAR},
  {am_ets, am_select, 1, ets_select_1, ets_select_1, BIF_KIND_REGULAR},
  {am_ets, am_select, 2, ets_select_2, ets_select_2, BIF_KIND_REGULAR},
  {am_ets, am_select, 3, ets_select_3, ets_select_3, BIF_KIND_REGULAR},
  {am_ets, am_select_count, 2, ets_select_count_2, ets_select_count_2, BIF_KIND_REGULAR},
  {am_ets, am_select_reverse, 1, ets_select_reverse_1, ets_select_reverse_1, BIF_KIND_REGULAR},
  {am_ets, am_select_reverse, 2, ets_select_reverse_2, ets_select_reverse_2, BIF_KIND_REGULAR},
  {am_ets, am_select_reverse, 3, ets_select_reverse_3, ets_select_reverse_3, BIF_KIND_REGULAR},
  {am_ets, am_select_replace, 2, ets_select_replace_2, ets_select_replace_2, BIF_KIND_REGULAR},
  {am_ets, am_match_spec_compile, 1, ets_match_spec_compile_1, ets_match_spec_compile_1, BIF_KIND_REGULAR},
  {am_ets, am_match_spec_run_r, 3, ets_match_spec_run_r_3, ets_match_spec_run_r_3, BIF_KIND_REGULAR},
  {am_os, am_getenv, 1, os_getenv_1, os_getenv_1, BIF_KIND_REGULAR},
  {am_os, am_putenv, 2, os_putenv_2, os_putenv_2, BIF_KIND_REGULAR},
  {am_os, am_unsetenv, 1, os_unsetenv_1, os_unsetenv_1, BIF_KIND_REGULAR},
  {am_os, am_getpid, 0, os_getpid_0, os_getpid_0, BIF_KIND_REGULAR},
  {am_os, am_timestamp, 0, os_timestamp_0, os_timestamp_0, BIF_KIND_REGULAR},
  {am_os, am_system_time, 0, os_system_time_0, os_system_time_0, BIF_KIND_REGULAR},
  {am_os, am_system_time, 1, os_system_time_1, os_system_time_1, BIF_KIND_REGULAR},
  {am_os, am_perf_counter, 0, os_perf_counter_0, os_perf_counter_0, BIF_KIND_REGULAR},
  {am_erl_ddll, am_try_load, 3, erl_ddll_try_load_3, erl_ddll_try_load_3, BIF_KIND_REGULAR},
  {am_erl_ddll, am_try_unload, 2, erl_ddll_try_unload_2, erl_ddll_try_unload_2, BIF_KIND_REGULAR},
  {am_erl_ddll, am_loaded_drivers, 0, erl_ddll_loaded_drivers_0, erl_ddll_loaded_drivers_0, BIF_KIND_REGULAR},
  {am_erl_ddll, am_info, 2, erl_ddll_info_2, erl_ddll_info_2, BIF_KIND_REGULAR},
  {am_erl_ddll, am_format_error_int, 1, erl_ddll_format_error_int_1, erl_ddll_format_error_int_1, BIF_KIND_REGULAR},
  {am_erl_ddll, am_monitor, 2, erl_ddll_monitor_2, erl_ddll_monitor_2, BIF_KIND_REGULAR},
  {am_erl_ddll, am_demonitor, 1, erl_ddll_demonitor_1, erl_ddll_demonitor_1, BIF_KIND_REGULAR},
  {am_re, am_version, 0, re_version_0, re_version_0, BIF_KIND_REGULAR},
  {am_re, am_compile, 1, re_compile_1, re_compile_1, BIF_KIND_REGULAR},
  {am_re, am_compile, 2, re_compile_2, re_compile_2, BIF_KIND_REGULAR},
  {am_re, am_run, 2, re_run_2, re_run_2, BIF_KIND_REGULAR},
  {am_re, am_run, 3, re_run_3, re_run_3, BIF_KIND_REGULAR},
  {am_re, am_internal_run, 4, re_internal_run_4, re_internal_run_4, BIF_KIND_REGULAR},
  {am_lists, am_member, 2, lists_member_2, lists_member_2, BIF_KIND_REGULAR},
  {am_lists, am_reverse, 2, lists_reverse_2, lists_reverse_2, BIF_KIND_REGULAR},
  {am_lists, am_keymember, 3, lists_keymember_3, lists_keymember_3, BIF_KIND_REGULAR},
  {am_lists, am_keysearch, 3, lists_keysearch_3, lists_keysearch_3, BIF_KIND_REGULAR},
  {am_lists, am_keyfind, 3, lists_keyfind_3, lists_keyfind_3, BIF_KIND_REGULAR},
  {am_erts_debug, am_disassemble, 1, erts_debug_disassemble_1, erts_debug_disassemble_1, BIF_KIND_REGULAR},
  {am_erts_debug, am_breakpoint, 2, erts_debug_breakpoint_2, erts_debug_breakpoint_2, BIF_KIND_REGULAR},
  {am_erts_debug, am_same, 2, erts_debug_same_2, erts_debug_same_2, BIF_KIND_REGULAR},
  {am_erts_debug, am_flat_size, 1, erts_debug_flat_size_1, erts_debug_flat_size_1, BIF_KIND_REGULAR},
  {am_erts_debug, am_get_internal_state, 1, erts_debug_get_internal_state_1, erts_debug_get_internal_state_1, BIF_KIND_REGULAR},
  {am_erts_debug, am_set_internal_state, 2, erts_debug_set_internal_state_2, erts_debug_set_internal_state_2, BIF_KIND_REGULAR},
  {am_erts_debug, am_dist_ext_to_term, 2, erts_debug_dist_ext_to_term_2, erts_debug_dist_ext_to_term_2, BIF_KIND_REGULAR},
  {am_erts_debug, am_instructions, 0, erts_debug_instructions_0, erts_debug_instructions_0, BIF_KIND_REGULAR},
  {am_erts_debug, am_interpreter_size, 0, erts_debug_interpreter_size_0, erts_debug_interpreter_size_0, BIF_KIND_REGULAR},
  {am_erts_debug, am_dirty_cpu, 2, schedule_dirty_cpu_erts_debug_dirty_cpu_2, erts_debug_dirty_cpu_2, BIF_KIND_REGULAR},
  {am_erts_debug, am_dirty_io, 2, schedule_dirty_io_erts_debug_dirty_io_2, erts_debug_dirty_io_2, BIF_KIND_REGULAR},
  {am_erts_debug, am_dirty, 3, erts_debug_dirty_3, erts_debug_dirty_3, BIF_KIND_REGULAR},
  {am_erts_debug, am_lcnt_control, 2, schedule_dirty_cpu_erts_debug_lcnt_control_2, erts_debug_lcnt_control_2, BIF_KIND_REGULAR},
  {am_erts_debug, am_lcnt_control, 1, erts_debug_lcnt_control_1, erts_debug_lcnt_control_1, BIF_KIND_REGULAR},
  {am_erts_debug, am_lcnt_collect, 0, schedule_dirty_cpu_erts_debug_lcnt_collect_0, erts_debug_lcnt_collect_0, BIF_KIND_REGULAR},
  {am_erts_debug, am_lcnt_clear, 0, schedule_dirty_cpu_erts_debug_lcnt_clear_0, erts_debug_lcnt_clear_0, BIF_KIND_REGULAR},
  {am_erlang, am_hibernate, 3, hibernate_3, hibernate_3, BIF_KIND_HEAVY},
  {am_error_logger, am_warning_map, 0, error_logger_warning_map_0, error_logger_warning_map_0, BIF_KIND_REGULAR},
  {am_erlang, am_get_module_info, 1, get_module_info_1, get_module_info_1, BIF_KIND_REGULAR},
  {am_erlang, am_get_module_info, 2, get_module_info_2, get_module_info_2, BIF_KIND_REGULAR},
  {am_erlang, am_is_boolean, 1, is_boolean_1, is_boolean_1, BIF_KIND_GUARD},
  {am_string, am_list_to_float, 1, string_list_to_float_1, string_list_to_float_1, BIF_KIND_REGULAR},
  {am_erlang, am_make_fun, 3, make_fun_3, make_fun_3, BIF_KIND_REGULAR},
  {am_erlang, am_iolist_size, 1, iolist_size_1, iolist_size_1, BIF_KIND_REGULAR},
  {am_erlang, am_iolist_to_binary, 1, iolist_to_binary_1, iolist_to_binary_1, BIF_KIND_REGULAR},
  {am_erlang, am_list_to_existing_atom, 1, list_to_existing_atom_1, list_to_existing_atom_1, BIF_KIND_GUARD},
  {am_erlang, am_is_bitstring, 1, is_bitstring_1, is_bitstring_1, BIF_KIND_GUARD},
  {am_erlang, am_tuple_size, 1, tuple_size_1, tuple_size_1, BIF_KIND_GUARD},
  {am_erlang, am_byte_size, 1, byte_size_1, byte_size_1, BIF_KIND_GUARD},
  {am_erlang, am_bit_size, 1, bit_size_1, bit_size_1, BIF_KIND_GUARD},
  {am_erlang, am_list_to_bitstring, 1, list_to_bitstring_1, list_to_bitstring_1, BIF_KIND_REGULAR},
  {am_erlang, am_bitstring_to_list, 1, bitstring_to_list_1, bitstring_to_list_1, BIF_KIND_REGULAR},
  {am_ets, am_update_element, 3, ets_update_element_3, ets_update_element_3, BIF_KIND_REGULAR},
  {am_erlang, am_decode_packet, 3, decode_packet_3, decode_packet_3, BIF_KIND_REGULAR},
  {am_unicode, am_characters_to_binary, 2, unicode_characters_to_binary_2, unicode_characters_to_binary_2, BIF_KIND_REGULAR},
  {am_unicode, am_characters_to_list, 2, unicode_characters_to_list_2, unicode_characters_to_list_2, BIF_KIND_REGULAR},
  {am_unicode, am_bin_is_7bit, 1, unicode_bin_is_7bit_1, unicode_bin_is_7bit_1, BIF_KIND_REGULAR},
  {am_erlang, am_atom_to_binary, 2, atom_to_binary_2, atom_to_binary_2, BIF_KIND_REGULAR},
  {am_erlang, am_binary_to_atom, 2, binary_to_atom_2, binary_to_atom_2, BIF_KIND_GUARD},
  {am_erlang, am_binary_to_existing_atom, 2, binary_to_existing_atom_2, binary_to_existing_atom_2, BIF_KIND_GUARD},
  {am_net_kernel, am_dflag_unicode_io, 1, net_kernel_dflag_unicode_io_1, net_kernel_dflag_unicode_io_1, BIF_KIND_REGULAR},
  {am_ets, am_give_away, 3, ets_give_away_3, ets_give_away_3, BIF_KIND_REGULAR},
  {am_ets, am_setopts, 2, ets_setopts_2, ets_setopts_2, BIF_KIND_REGULAR},
  {am_erlang, am_load_nif, 2, load_nif_2, load_nif_2, BIF_KIND_REGULAR},
  {am_erlang, am_call_on_load_function, 1, call_on_load_function_1, call_on_load_function_1, BIF_KIND_REGULAR},
  {am_erlang, am_finish_after_on_load, 2, finish_after_on_load_2, finish_after_on_load_2, BIF_KIND_REGULAR},
  {am_erlang, am_binary_to_term, 2, binary_to_term_2, binary_to_term_2, BIF_KIND_REGULAR},
  {am_erlang, am_binary_part, 2, binary_part_2, binary_part_2, BIF_KIND_GUARD},
  {am_erlang, am_binary_part, 3, binary_part_3, binary_part_3, BIF_KIND_GUARD},
  {am_binary, am_compile_pattern, 1, binary_compile_pattern_1, binary_compile_pattern_1, BIF_KIND_REGULAR},
  {am_binary, am_match, 2, binary_match_2, binary_match_2, BIF_KIND_REGULAR},
  {am_binary, am_match, 3, binary_match_3, binary_match_3, BIF_KIND_REGULAR},
  {am_binary, am_matches, 2, binary_matches_2, binary_matches_2, BIF_KIND_REGULAR},
  {am_binary, am_matches, 3, binary_matches_3, binary_matches_3, BIF_KIND_REGULAR},
  {am_binary, am_longest_common_prefix, 1, binary_longest_common_prefix_1, binary_longest_common_prefix_1, BIF_KIND_REGULAR},
  {am_binary, am_longest_common_suffix, 1, binary_longest_common_suffix_1, binary_longest_common_suffix_1, BIF_KIND_REGULAR},
  {am_binary, am_first, 1, binary_first_1, binary_first_1, BIF_KIND_REGULAR},
  {am_binary, am_last, 1, binary_last_1, binary_last_1, BIF_KIND_REGULAR},
  {am_binary, am_at, 2, binary_at_2, binary_at_2, BIF_KIND_REGULAR},
  {am_binary, am_part, 2, binary_binary_part_2, binary_binary_part_2, BIF_KIND_REGULAR},
  {am_binary, am_part, 3, binary_binary_part_3, binary_binary_part_3, BIF_KIND_REGULAR},
  {am_binary, am_list_to_bin, 1, binary_list_to_bin_1, binary_list_to_bin_1, BIF_KIND_REGULAR},
  {am_binary, am_copy, 1, binary_copy_1, binary_copy_1, BIF_KIND_REGULAR},
  {am_binary, am_copy, 2, binary_copy_2, binary_copy_2, BIF_KIND_REGULAR},
  {am_binary, am_referenced_byte_size, 1, binary_referenced_byte_size_1, binary_referenced_byte_size_1, BIF_KIND_REGULAR},
  {am_binary, am_encode_unsigned, 1, binary_encode_unsigned_1, binary_encode_unsigned_1, BIF_KIND_REGULAR},
  {am_binary, am_encode_unsigned, 2, binary_encode_unsigned_2, binary_encode_unsigned_2, BIF_KIND_REGULAR},
  {am_binary, am_decode_unsigned, 1, binary_decode_unsigned_1, binary_decode_unsigned_1, BIF_KIND_REGULAR},
  {am_binary, am_decode_unsigned, 2, binary_decode_unsigned_2, binary_decode_unsigned_2, BIF_KIND_REGULAR},
  {am_erlang, am_nif_error, 1, nif_error_1, nif_error_1, BIF_KIND_REGULAR},
  {am_erlang, am_nif_error, 2, nif_error_2, nif_error_2, BIF_KIND_REGULAR},
  {am_prim_file, am_internal_name2native, 1, prim_file_internal_name2native_1, prim_file_internal_name2native_1, BIF_KIND_REGULAR},
  {am_prim_file, am_internal_native2name, 1, prim_file_internal_native2name_1, prim_file_internal_native2name_1, BIF_KIND_REGULAR},
  {am_prim_file, am_internal_normalize_utf8, 1, prim_file_internal_normalize_utf8_1, prim_file_internal_normalize_utf8_1, BIF_KIND_REGULAR},
  {am_prim_file, am_is_translatable, 1, prim_file_is_translatable_1, prim_file_is_translatable_1, BIF_KIND_REGULAR},
  {am_file, am_native_name_encoding, 0, file_native_name_encoding_0, file_native_name_encoding_0, BIF_KIND_REGULAR},
  {am_erlang, am_check_old_code, 1, check_old_code_1, check_old_code_1, BIF_KIND_REGULAR},
  {am_erlang, am_universaltime_to_posixtime, 1, universaltime_to_posixtime_1, universaltime_to_posixtime_1, BIF_KIND_REGULAR},
  {am_erlang, am_posixtime_to_universaltime, 1, posixtime_to_universaltime_1, posixtime_to_universaltime_1, BIF_KIND_REGULAR},
  {am_erlang, am_dt_put_tag, 1, dt_put_tag_1, dt_put_tag_1, BIF_KIND_REGULAR},
  {am_erlang, am_dt_get_tag, 0, dt_get_tag_0, dt_get_tag_0, BIF_KIND_REGULAR},
  {am_erlang, am_dt_get_tag_data, 0, dt_get_tag_data_0, dt_get_tag_data_0, BIF_KIND_REGULAR},
  {am_erlang, am_dt_spread_tag, 1, dt_spread_tag_1, dt_spread_tag_1, BIF_KIND_REGULAR},
  {am_erlang, am_dt_restore_tag, 1, dt_restore_tag_1, dt_restore_tag_1, BIF_KIND_REGULAR},
  {am_erlang, am_dt_prepend_vm_tag_data, 1, dt_prepend_vm_tag_data_1, dt_prepend_vm_tag_data_1, BIF_KIND_REGULAR},
  {am_erlang, am_dt_append_vm_tag_data, 1, dt_append_vm_tag_data_1, dt_append_vm_tag_data_1, BIF_KIND_REGULAR},
  {am_erlang, am_finish_loading, 1, finish_loading_1, finish_loading_1, BIF_KIND_REGULAR},
  {am_erlang, am_insert_element, 3, insert_element_3, insert_element_3, BIF_KIND_REGULAR},
  {am_erlang, am_delete_element, 2, delete_element_2, delete_element_2, BIF_KIND_REGULAR},
  {am_erlang, am_integer_to_binary, 1, integer_to_binary_1, integer_to_binary_1, BIF_KIND_REGULAR},
  {am_erlang, am_float_to_binary, 1, float_to_binary_1, float_to_binary_1, BIF_KIND_REGULAR},
  {am_erlang, am_float_to_binary, 2, float_to_binary_2, float_to_binary_2, BIF_KIND_REGULAR},
  {am_erlang, am_binary_to_float, 1, binary_to_float_1, binary_to_float_1, BIF_KIND_REGULAR},
  {am_io, am_printable_range, 0, io_printable_range_0, io_printable_range_0, BIF_KIND_REGULAR},
  {am_re, am_inspect, 2, re_inspect_2, re_inspect_2, BIF_KIND_REGULAR},
  {am_erlang, am_is_map, 1, is_map_1, is_map_1, BIF_KIND_GUARD},
  {am_erlang, am_map_size, 1, map_size_1, map_size_1, BIF_KIND_GUARD},
  {am_maps, am_find, 2, maps_find_2, maps_find_2, BIF_KIND_REGULAR},
  {am_maps, am_get, 2, maps_get_2, maps_get_2, BIF_KIND_REGULAR},
  {am_maps, am_from_list, 1, maps_from_list_1, maps_from_list_1, BIF_KIND_REGULAR},
  {am_maps, am_is_key, 2, maps_is_key_2, maps_is_key_2, BIF_KIND_REGULAR},
  {am_maps, am_keys, 1, maps_keys_1, maps_keys_1, BIF_KIND_REGULAR},
  {am_maps, am_merge, 2, maps_merge_2, maps_merge_2, BIF_KIND_REGULAR},
  {am_maps, am_put, 3, maps_put_3, maps_put_3, BIF_KIND_REGULAR},
  {am_maps, am_remove, 2, maps_remove_2, maps_remove_2, BIF_KIND_REGULAR},
  {am_maps, am_update, 3, maps_update_3, maps_update_3, BIF_KIND_REGULAR},
  {am_maps, am_values, 1, maps_values_1, maps_values_1, BIF_KIND_REGULAR},
  {am_erts_internal, am_cmp_term, 2, erts_internal_cmp_term_2, erts_internal_cmp_term_2, BIF_KIND_REGULAR},
  {am_ets, am_take, 2, ets_take_2, ets_take_2, BIF_KIND_REGULAR},
  {am_erlang, am_fun_info_mfa, 1, fun_info_mfa_1, fun_info_mfa_1, BIF_KIND_REGULAR},
  {am_erlang, am_get_keys, 0, get_keys_0, get_keys_0, BIF_KIND_REGULAR},
  {am_ets, am_update_counter, 4, ets_update_counter_4, ets_update_counter_4, BIF_KIND_REGULAR},
  {am_erts_debug, am_map_info, 1, erts_debug_map_info_1, erts_debug_map_info_1, BIF_KIND_REGULAR},
  {am_erts_internal, am_is_process_executing_dirty, 1, erts_internal_is_process_executing_dirty_1, erts_internal_is_process_executing_dirty_1, BIF_KIND_REGULAR},
  {am_erts_internal, am_check_dirty_process_code, 2, erts_internal_check_dirty_process_code_2, erts_internal_check_dirty_process_code_2, BIF_KIND_REGULAR},
  {am_erts_internal, am_purge_module, 2, erts_internal_purge_module_2, erts_internal_purge_module_2, BIF_KIND_REGULAR},
  {am_binary, am_split, 2, binary_split_2, binary_split_2, BIF_KIND_REGULAR},
  {am_binary, am_split, 3, binary_split_3, binary_split_3, BIF_KIND_REGULAR},
  {am_erts_debug, am_size_shared, 1, erts_debug_size_shared_1, erts_debug_size_shared_1, BIF_KIND_REGULAR},
  {am_erts_debug, am_copy_shared, 2, erts_debug_copy_shared_2, erts_debug_copy_shared_2, BIF_KIND_REGULAR},
  {am_erlang, am_has_prepared_code_on_load, 1, has_prepared_code_on_load_1, has_prepared_code_on_load_1, BIF_KIND_REGULAR},
  {am_maps, am_take, 2, maps_take_2, maps_take_2, BIF_KIND_REGULAR},
  {am_erlang, am_floor, 1, floor_1, floor_1, BIF_KIND_GUARD},
  {am_erlang, am_ceil, 1, ceil_1, ceil_1, BIF_KIND_GUARD},
  {am_math, am_floor, 1, math_floor_1, math_floor_1, BIF_KIND_REGULAR},
  {am_math, am_ceil, 1, math_ceil_1, math_ceil_1, BIF_KIND_REGULAR},
  {am_math, am_fmod, 2, math_fmod_2, math_fmod_2, BIF_KIND_REGULAR},
  {am_os, am_set_signal, 2, os_set_signal_2, os_set_signal_2, BIF_KIND_REGULAR},
  {am_erlang, am_iolist_to_iovec, 1, iolist_to_iovec_1, iolist_to_iovec_1, BIF_KIND_REGULAR},
  {am_erts_internal, am_get_dflags, 0, erts_internal_get_dflags_0, erts_internal_get_dflags_0, BIF_KIND_REGULAR},
  {am_erts_internal, am_new_connection, 1, erts_internal_new_connection_1, erts_internal_new_connection_1, BIF_KIND_REGULAR},
  {am_erts_internal, am_map_next, 3, erts_internal_map_next_3, erts_internal_map_next_3, BIF_KIND_REGULAR},
  {am_ets, am_whereis, 1, ets_whereis_1, ets_whereis_1, BIF_KIND_REGULAR},
  {am_erts_internal, am_gather_alloc_histograms, 1, erts_internal_gather_alloc_histograms_1, erts_internal_gather_alloc_histograms_1, BIF_KIND_REGULAR},
  {am_erts_internal, am_gather_carrier_info, 1, erts_internal_gather_carrier_info_1, erts_internal_gather_carrier_info_1, BIF_KIND_REGULAR},
  {am_erlang, am_map_get, 2, map_get_2, map_get_2, BIF_KIND_GUARD},
  {am_erlang, am_is_map_key, 2, is_map_key_2, is_map_key_2, BIF_KIND_GUARD},
  {am_ets, am_internal_delete_all, 2, ets_internal_delete_all_2, ets_internal_delete_all_2, BIF_KIND_REGULAR},
  {am_ets, am_internal_select_delete, 2, ets_internal_select_delete_2, ets_internal_select_delete_2, BIF_KIND_REGULAR},
  {am_persistent_term, am_put, 2, persistent_term_put_2, persistent_term_put_2, BIF_KIND_REGULAR},
  {am_persistent_term, am_get, 1, persistent_term_get_1, persistent_term_get_1, BIF_KIND_REGULAR},
  {am_persistent_term, am_get, 0, persistent_term_get_0, persistent_term_get_0, BIF_KIND_REGULAR},
  {am_persistent_term, am_erase, 1, persistent_term_erase_1, persistent_term_erase_1, BIF_KIND_REGULAR},
  {am_persistent_term, am_info, 0, persistent_term_info_0, persistent_term_info_0, BIF_KIND_REGULAR},
  {am_erts_internal, am_erase_persistent_terms, 0, erts_internal_erase_persistent_terms_0, erts_internal_erase_persistent_terms_0, BIF_KIND_REGULAR},
  {am_erts_internal, am_atomics_new, 2, erts_internal_atomics_new_2, erts_internal_atomics_new_2, BIF_KIND_REGULAR},
  {am_atomics, am_get, 2, atomics_get_2, atomics_get_2, BIF_KIND_REGULAR},
  {am_atomics, am_put, 3, atomics_put_3, atomics_put_3, BIF_KIND_REGULAR},
  {am_atomics, am_add, 3, atomics_add_3, atomics_add_3, BIF_KIND_REGULAR},
  {am_atomics, am_add_get, 3, atomics_add_get_3, atomics_add_get_3, BIF_KIND_REGULAR},
  {am_atomics, am_exchange, 3, atomics_exchange_3, atomics_exchange_3, BIF_KIND_REGULAR},
  {am_atomics, am_compare_exchange, 4, atomics_compare_exchange_4, atomics_compare_exchange_4, BIF_KIND_REGULAR},
  {am_atomics, am_info, 1, atomics_info_1, atomics_info_1, BIF_KIND_REGULAR},
  {am_erts_internal, am_counters_new, 1, erts_internal_counters_new_1, erts_internal_counters_new_1, BIF_KIND_REGULAR},
  {am_erts_internal, am_counters_get, 2, erts_internal_counters_get_2, erts_internal_counters_get_2, BIF_KIND_REGULAR},
  {am_erts_internal, am_counters_add, 3, erts_internal_counters_add_3, erts_internal_counters_add_3, BIF_KIND_REGULAR},
  {am_erts_internal, am_counters_put, 3, erts_internal_counters_put_3, erts_internal_counters_put_3, BIF_KIND_REGULAR},
  {am_erts_internal, am_counters_info, 1, erts_internal_counters_info_1, erts_internal_counters_info_1, BIF_KIND_REGULAR},
  {am_erts_internal, am_spawn_system_process, 3, erts_internal_spawn_system_process_3, erts_internal_spawn_system_process_3, BIF_KIND_REGULAR},
  {am_erlang, am_integer_to_list, 2, integer_to_list_2, integer_to_list_2, BIF_KIND_REGULAR},
  {am_erlang, am_integer_to_binary, 2, integer_to_binary_2, integer_to_binary_2, BIF_KIND_REGULAR},
  {am_persistent_term, am_get, 2, persistent_term_get_2, persistent_term_get_2, BIF_KIND_REGULAR},
  {am_erts_internal, am_ets_lookup_binary_info, 2, erts_internal_ets_lookup_binary_info_2, erts_internal_ets_lookup_binary_info_2, BIF_KIND_REGULAR},
  {am_erts_internal, am_ets_raw_first, 1, erts_internal_ets_raw_first_1, erts_internal_ets_raw_first_1, BIF_KIND_REGULAR},
  {am_erts_internal, am_ets_raw_next, 2, erts_internal_ets_raw_next_2, erts_internal_ets_raw_next_2, BIF_KIND_REGULAR},
  {am_erts_internal, am_abort_pending_connection, 2, erts_internal_abort_pending_connection_2, erts_internal_abort_pending_connection_2, BIF_KIND_REGULAR},
  {am_erts_internal, am_get_creation, 0, erts_internal_get_creation_0, erts_internal_get_creation_0, BIF_KIND_REGULAR},
  {am_erts_internal, am_prepare_loading, 2, erts_internal_prepare_loading_2, erts_internal_prepare_loading_2, BIF_KIND_REGULAR},
  {am_erts_internal, am_beamfile_chunk, 2, erts_internal_beamfile_chunk_2, erts_internal_beamfile_chunk_2, BIF_KIND_REGULAR},
  {am_erts_internal, am_beamfile_module_md5, 1, erts_internal_beamfile_module_md5_1, erts_internal_beamfile_module_md5_1, BIF_KIND_REGULAR},
  {am_os, am_env, 0, os_env_0, os_env_0, BIF_KIND_REGULAR},
  {am_erlang, am_alias, 1, alias_1, alias_1, BIF_KIND_REGULAR},
  {am_erlang, am_unalias, 1, unalias_1, unalias_1, BIF_KIND_REGULAR},
  {am_erlang, am_monitor, 3, monitor_3, monitor_3, BIF_KIND_REGULAR},
  {am_erlang, am_error, 3, error_3, error_3, BIF_KIND_REGULAR},
  {am_maps, am_from_keys, 2, maps_from_keys_2, maps_from_keys_2, BIF_KIND_REGULAR},
  {am_erts_internal, am_binary_to_integer, 2, erts_internal_binary_to_integer_2, erts_internal_binary_to_integer_2, BIF_KIND_REGULAR},
  {am_erts_internal, am_list_to_integer, 2, erts_internal_list_to_integer_2, erts_internal_list_to_integer_2, BIF_KIND_REGULAR},
  {am_erlang, am_min, 2, min_2, min_2, BIF_KIND_GUARD},
  {am_erlang, am_max, 2, max_2, max_2, BIF_KIND_GUARD},
  {am_erts_internal, am_term_to_string, 2, erts_internal_term_to_string_2, erts_internal_term_to_string_2, BIF_KIND_REGULAR},
  {am_code, am_coverage_support, 0, code_coverage_support_0, code_coverage_support_0, BIF_KIND_REGULAR},
  {am_code, am_get_coverage_mode, 0, code_get_coverage_mode_0, code_get_coverage_mode_0, BIF_KIND_REGULAR},
  {am_code, am_get_coverage_mode, 1, code_get_coverage_mode_1, code_get_coverage_mode_1, BIF_KIND_REGULAR},
  {am_code, am_get_coverage, 2, code_get_coverage_2, code_get_coverage_2, BIF_KIND_REGULAR},
  {am_code, am_reset_coverage, 1, code_reset_coverage_1, code_reset_coverage_1, BIF_KIND_REGULAR},
  {am_code, am_set_coverage_mode, 1, code_set_coverage_mode_1, code_set_coverage_mode_1, BIF_KIND_REGULAR},
  {am_ets, am_update_element, 4, ets_update_element_4, ets_update_element_4, BIF_KIND_REGULAR},
  {am_erts_internal, am_trace_session_create, 3, erts_internal_trace_session_create_3, erts_internal_trace_session_create_3, BIF_KIND_REGULAR},
  {am_erts_internal, am_trace_session_destroy, 1, erts_internal_trace_session_destroy_1, erts_internal_trace_session_destroy_1, BIF_KIND_REGULAR},
  {am_erts_internal, am_trace, 4, erts_internal_trace_4, erts_internal_trace_4, BIF_KIND_REGULAR},
  {am_erts_internal, am_trace_pattern, 4, erts_internal_trace_pattern_4, erts_internal_trace_pattern_4, BIF_KIND_REGULAR},
  {am_erts_internal, am_trace_info, 3, erts_internal_trace_info_3, erts_internal_trace_info_3, BIF_KIND_REGULAR},
  {am_erts_trace_cleaner, am_check, 0, erts_trace_cleaner_check_0, erts_trace_cleaner_check_0, BIF_KIND_REGULAR},
  {am_erts_trace_cleaner, am_send_trace_clean_signal, 1, erts_trace_cleaner_send_trace_clean_signal_1, erts_trace_cleaner_send_trace_clean_signal_1, BIF_KIND_REGULAR},
  {am_erts_internal, am_system_monitor, 1, erts_internal_system_monitor_1, erts_internal_system_monitor_1, BIF_KIND_REGULAR},
  {am_erts_internal, am_system_monitor, 3, erts_internal_system_monitor_3, erts_internal_system_monitor_3, BIF_KIND_REGULAR},
  {am_erts_internal, am_processes_next, 1, erts_internal_processes_next_1, erts_internal_processes_next_1, BIF_KIND_REGULAR},
  {am_code, am_get_debug_info, 1, code_get_debug_info_1, code_get_debug_info_1, BIF_KIND_REGULAR},
  {am_erlang, am_exit, 3, exit_3, exit_3, BIF_KIND_REGULAR},
  {am_erlang, am_link, 2, link_2, link_2, BIF_KIND_REGULAR},
  {am_erl_debugger, am_supported, 0, erl_debugger_supported_0, erl_debugger_supported_0, BIF_KIND_REGULAR},
  {am_erl_debugger, am_instrumentations, 0, erl_debugger_instrumentations_0, erl_debugger_instrumentations_0, BIF_KIND_REGULAR},
  {am_erl_debugger, am_toggle_instrumentations, 1, erl_debugger_toggle_instrumentations_1, erl_debugger_toggle_instrumentations_1, BIF_KIND_REGULAR},
  {am_erl_debugger, am_register, 1, erl_debugger_register_1, erl_debugger_register_1, BIF_KIND_REGULAR},
  {am_erl_debugger, am_unregister, 2, erl_debugger_unregister_2, erl_debugger_unregister_2, BIF_KIND_REGULAR},
  {am_erl_debugger, am_whereis, 0, erl_debugger_whereis_0, erl_debugger_whereis_0, BIF_KIND_REGULAR},
  {am_erl_debugger, am_breakpoint, 3, erl_debugger_breakpoint_3, erl_debugger_breakpoint_3, BIF_KIND_REGULAR},
  {am_erl_debugger, am_breakpoints, 1, erl_debugger_breakpoints_1, erl_debugger_breakpoints_1, BIF_KIND_REGULAR},
  {am_erl_debugger, am_breakpoints, 3, erl_debugger_breakpoints_3, erl_debugger_breakpoints_3, BIF_KIND_REGULAR},
  {am_erts_internal, am_notify_breakpoint_hit, 3, erts_internal_notify_breakpoint_hit_3, erts_internal_notify_breakpoint_hit_3, BIF_KIND_REGULAR},
  {am_erl_debugger, am_stack_frames, 2, erl_debugger_stack_frames_2, erl_debugger_stack_frames_2, BIF_KIND_REGULAR},
  {am_erl_debugger, am_peek_stack_frame_slot, 4, erl_debugger_peek_stack_frame_slot_4, erl_debugger_peek_stack_frame_slot_4, BIF_KIND_REGULAR},
  {am_erl_debugger, am_xregs_count, 1, erl_debugger_xregs_count_1, erl_debugger_xregs_count_1, BIF_KIND_REGULAR},
  {am_erl_debugger, am_peek_xreg, 3, erl_debugger_peek_xreg_3, erl_debugger_peek_xreg_3, BIF_KIND_REGULAR},
  {am_erts_debug, am_unaligned_bitstring, 2, erts_debug_unaligned_bitstring_2, erts_debug_unaligned_bitstring_2, BIF_KIND_REGULAR},
  {am_re, am_import, 1, re_import_1, re_import_1, BIF_KIND_REGULAR},
  {am_persistent_term, am_put_new, 2, persistent_term_put_new_2, persistent_term_put_new_2, BIF_KIND_REGULAR},
};

