/*
 *  Warning: Do *not* edit this file. It was automatically
 *  generated by 'make_tables'.
 */

#ifndef __ERL_BIF_TABLE_H__
#define __ERL_BIF_TABLE_H__

#include "sys.h"

typedef void *BifFunction;

typedef enum {
    BIF_KIND_REGULAR,
    BIF_KIND_HEAVY,
    BIF_KIND_GUARD
} BifKind;

typedef struct bif_entry {
    Eterm module;
    Eterm name;
    int arity;
    BifFunction f;
    BifFunction impl;
    BifKind kind;
} BifEntry;

typedef struct erts_gc_bif {
    BifFunction bif;
    BifFunction gc_bif;
    int exp_ix;
} ErtsGcBif;

typedef struct erts_u_bif {
    BifFunction bif;
    int exp_ix;
} ErtsUBif;

extern BifEntry bif_table[];
extern const ErtsUBif erts_u_bifs[];

#define BIF_SIZE 542

#define BIF_TRAP_EXPORT(BIF_NAME) (&bif_trap_exports__[BIF_NAME])
extern Export bif_trap_exports__[];

#define BIF_abs_1 0
#define BIF_adler32_1 1
#define BIF_adler32_2 2
#define BIF_adler32_combine_3 3
#define BIF_atom_to_list_1 4
#define BIF_binary_to_list_1 5
#define BIF_binary_to_list_3 6
#define BIF_binary_to_term_1 7
#define BIF_crc32_1 8
#define BIF_crc32_2 9
#define BIF_crc32_combine_3 10
#define BIF_date_0 11
#define BIF_delete_module_1 12
#define BIF_display_1 13
#define BIF_display_string_2 14
#define BIF_element_2 15
#define BIF_erase_0 16
#define BIF_erase_1 17
#define BIF_exit_1 18
#define BIF_exit_2 19
#define BIF_exit_signal_2 20
#define BIF_external_size_1 21
#define BIF_external_size_2 22
#define BIF_float_1 23
#define BIF_float_to_list_1 24
#define BIF_float_to_list_2 25
#define BIF_fun_info_2 26
#define BIF_erts_internal_garbage_collect_1 27
#define BIF_get_0 28
#define BIF_get_1 29
#define BIF_get_keys_1 30
#define BIF_group_leader_0 31
#define BIF_erts_internal_group_leader_2 32
#define BIF_erts_internal_group_leader_3 33
#define BIF_halt_2 34
#define BIF_phash_2 35
#define BIF_phash2_1 36
#define BIF_phash2_2 37
#define BIF_hd_1 38
#define BIF_integer_to_list_1 39
#define BIF_length_1 40
#define BIF_link_1 41
#define BIF_list_to_atom_1 42
#define BIF_list_to_binary_1 43
#define BIF_list_to_float_1 44
#define BIF_list_to_pid_1 45
#define BIF_list_to_port_1 46
#define BIF_list_to_ref_1 47
#define BIF_list_to_tuple_1 48
#define BIF_loaded_0 49
#define BIF_localtime_0 50
#define BIF_localtime_to_universaltime_2 51
#define BIF_make_ref_0 52
#define BIF_unique_integer_0 53
#define BIF_unique_integer_1 54
#define BIF_md5_1 55
#define BIF_md5_init_0 56
#define BIF_md5_update_2 57
#define BIF_md5_final_1 58
#define BIF_module_loaded_1 59
#define BIF_function_exported_3 60
#define BIF_monitor_node_2 61
#define BIF_monitor_node_3 62
#define BIF_node_1 63
#define BIF_node_0 64
#define BIF_nodes_0 65
#define BIF_nodes_1 66
#define BIF_nodes_2 67
#define BIF_now_0 68
#define BIF_monotonic_time_0 69
#define BIF_monotonic_time_1 70
#define BIF_system_time_0 71
#define BIF_system_time_1 72
#define BIF_time_offset_0 73
#define BIF_time_offset_1 74
#define BIF_timestamp_0 75
#define BIF_erts_internal_open_port_2 76
#define BIF_pid_to_list_1 77
#define BIF_ports_0 78
#define BIF_pre_loaded_0 79
#define BIF_process_flag_2 80
#define BIF_erts_internal_process_flag_3 81
#define BIF_process_info_1 82
#define BIF_process_info_2 83
#define BIF_processes_0 84
#define BIF_put_2 85
#define BIF_register_2 86
#define BIF_registered_0 87
#define BIF_round_1 88
#define BIF_self_0 89
#define BIF_setelement_3 90
#define BIF_size_1 91
#define BIF_spawn_3 92
#define BIF_spawn_link_3 93
#define BIF_split_binary_2 94
#define BIF_statistics_1 95
#define BIF_term_to_binary_1 96
#define BIF_term_to_binary_2 97
#define BIF_term_to_iovec_1 98
#define BIF_term_to_iovec_2 99
#define BIF_throw_1 100
#define BIF_time_0 101
#define BIF_tl_1 102
#define BIF_trunc_1 103
#define BIF_tuple_to_list_1 104
#define BIF_universaltime_0 105
#define BIF_universaltime_to_localtime_1 106
#define BIF_unlink_1 107
#define BIF_unregister_1 108
#define BIF_whereis_1 109
#define BIF_spawn_opt_4 110
#define BIF_setnode_2 111
#define BIF_dist_get_stat_1 112
#define BIF_dist_ctrl_input_handler_2 113
#define BIF_dist_ctrl_put_data_2 114
#define BIF_dist_ctrl_get_data_1 115
#define BIF_dist_ctrl_get_data_notification_1 116
#define BIF_dist_ctrl_get_opt_2 117
#define BIF_dist_ctrl_set_opt_3 118
#define BIF_erts_internal_port_info_1 119
#define BIF_erts_internal_port_info_2 120
#define BIF_erts_internal_port_call_3 121
#define BIF_erts_internal_port_command_3 122
#define BIF_erts_internal_port_control_3 123
#define BIF_erts_internal_port_close_1 124
#define BIF_erts_internal_port_connect_2 125
#define BIF_erts_internal_request_system_task_3 126
#define BIF_erts_internal_request_system_task_4 127
#define BIF_erts_internal_check_process_code_1 128
#define BIF_erts_internal_map_to_tuple_keys_1 129
#define BIF_erts_internal_term_type_1 130
#define BIF_erts_internal_map_hashmap_children_1 131
#define BIF_erts_internal_time_unit_0 132
#define BIF_erts_internal_perf_counter_unit_0 133
#define BIF_erts_internal_is_system_process_1 134
#define BIF_erts_internal_system_check_1 135
#define BIF_erts_internal_scheduler_wall_time_1 136
#define BIF_erts_internal_dirty_process_handle_signals_1 137
#define BIF_erts_internal_create_dist_channel_3 138
#define BIF_erts_internal_ets_super_user_1 139
#define BIF_erts_internal_spawn_request_4 140
#define BIF_erts_internal_dist_spawn_request_4 141
#define BIF_erts_internal_no_aux_work_threads_0 142
#define BIF_spawn_request_abandon_1 143
#define BIF_erts_literal_area_collector_release_area_switch_0 144
#define BIF_erts_literal_area_collector_send_copy_request_3 145
#define BIF_port_set_data_2 146
#define BIF_port_get_data_1 147
#define BIF_erts_internal_trace_pattern_3 148
#define BIF_erts_internal_trace_3 149
#define BIF_trace_info_2 150
#define BIF_trace_delivered_1 151
#define BIF_seq_trace_2 152
#define BIF_seq_trace_info_1 153
#define BIF_seq_trace_print_1 154
#define BIF_seq_trace_print_2 155
#define BIF_erts_internal_suspend_process_2 156
#define BIF_resume_process_1 157
#define BIF_erts_internal_process_display_2 158
#define BIF_bump_reductions_1 159
#define BIF_math_cos_1 160
#define BIF_math_cosh_1 161
#define BIF_math_sin_1 162
#define BIF_math_sinh_1 163
#define BIF_math_tan_1 164
#define BIF_math_tanh_1 165
#define BIF_math_acos_1 166
#define BIF_math_acosh_1 167
#define BIF_math_asin_1 168
#define BIF_math_asinh_1 169
#define BIF_math_atan_1 170
#define BIF_math_atanh_1 171
#define BIF_math_erf_1 172
#define BIF_math_erfc_1 173
#define BIF_math_exp_1 174
#define BIF_math_log_1 175
#define BIF_math_log2_1 176
#define BIF_math_log10_1 177
#define BIF_math_sqrt_1 178
#define BIF_math_atan2_2 179
#define BIF_math_pow_2 180
#define BIF_start_timer_3 181
#define BIF_start_timer_4 182
#define BIF_send_after_3 183
#define BIF_send_after_4 184
#define BIF_cancel_timer_1 185
#define BIF_cancel_timer_2 186
#define BIF_read_timer_1 187
#define BIF_read_timer_2 188
#define BIF_make_tuple_2 189
#define BIF_append_element_2 190
#define BIF_make_tuple_3 191
#define BIF_system_flag_2 192
#define BIF_system_info_1 193
#define BIF_system_profile_2 194
#define BIF_system_profile_0 195
#define BIF_ref_to_list_1 196
#define BIF_port_to_list_1 197
#define BIF_fun_to_list_1 198
#define BIF_monitor_2 199
#define BIF_demonitor_1 200
#define BIF_demonitor_2 201
#define BIF_is_process_alive_1 202
#define BIF_erts_internal_is_process_alive_2 203
#define BIF_error_1 204
#define BIF_error_2 205
#define BIF_raise_3 206
#define BIF_is_builtin_3 207
#define BIF_and_2 208
#define BIF_or_2 209
#define BIF_xor_2 210
#define BIF_not_1 211
#define BIF_sgt_2 212
#define BIF_sge_2 213
#define BIF_slt_2 214
#define BIF_sle_2 215
#define BIF_seq_2 216
#define BIF_seqeq_2 217
#define BIF_sneq_2 218
#define BIF_sneqeq_2 219
#define BIF_splus_2 220
#define BIF_sminus_2 221
#define BIF_stimes_2 222
#define BIF_div_2 223
#define BIF_intdiv_2 224
#define BIF_rem_2 225
#define BIF_bor_2 226
#define BIF_band_2 227
#define BIF_bxor_2 228
#define BIF_bsl_2 229
#define BIF_bsr_2 230
#define BIF_bnot_1 231
#define BIF_sminus_1 232
#define BIF_splus_1 233
#define BIF_ebif_bang_2 234
#define BIF_send_2 235
#define BIF_send_3 236
#define BIF_ebif_plusplus_2 237
#define BIF_append_2 238
#define BIF_ebif_minusminus_2 239
#define BIF_subtract_2 240
#define BIF_is_atom_1 241
#define BIF_is_list_1 242
#define BIF_is_tuple_1 243
#define BIF_is_float_1 244
#define BIF_is_integer_1 245
#define BIF_is_number_1 246
#define BIF_is_pid_1 247
#define BIF_is_port_1 248
#define BIF_is_reference_1 249
#define BIF_is_binary_1 250
#define BIF_is_function_1 251
#define BIF_is_function_2 252
#define BIF_is_record_2 253
#define BIF_is_record_3 254
#define BIF_match_spec_test_3 255
#define BIF_ets_internal_request_all_0 256
#define BIF_ets_new_2 257
#define BIF_ets_delete_1 258
#define BIF_ets_delete_2 259
#define BIF_ets_delete_object_2 260
#define BIF_ets_first_1 261
#define BIF_ets_first_lookup_1 262
#define BIF_ets_is_compiled_ms_1 263
#define BIF_ets_lookup_2 264
#define BIF_ets_lookup_element_3 265
#define BIF_ets_lookup_element_4 266
#define BIF_ets_info_1 267
#define BIF_ets_info_2 268
#define BIF_ets_last_1 269
#define BIF_ets_last_lookup_1 270
#define BIF_ets_match_1 271
#define BIF_ets_match_2 272
#define BIF_ets_match_3 273
#define BIF_ets_match_object_1 274
#define BIF_ets_match_object_2 275
#define BIF_ets_match_object_3 276
#define BIF_ets_member_2 277
#define BIF_ets_next_2 278
#define BIF_ets_next_lookup_2 279
#define BIF_ets_prev_2 280
#define BIF_ets_prev_lookup_2 281
#define BIF_ets_insert_2 282
#define BIF_ets_insert_new_2 283
#define BIF_ets_rename_2 284
#define BIF_ets_safe_fixtable_2 285
#define BIF_ets_slot_2 286
#define BIF_ets_update_counter_3 287
#define BIF_ets_select_1 288
#define BIF_ets_select_2 289
#define BIF_ets_select_3 290
#define BIF_ets_select_count_2 291
#define BIF_ets_select_reverse_1 292
#define BIF_ets_select_reverse_2 293
#define BIF_ets_select_reverse_3 294
#define BIF_ets_select_replace_2 295
#define BIF_ets_match_spec_compile_1 296
#define BIF_ets_match_spec_run_r_3 297
#define BIF_os_getenv_1 298
#define BIF_os_putenv_2 299
#define BIF_os_unsetenv_1 300
#define BIF_os_getpid_0 301
#define BIF_os_timestamp_0 302
#define BIF_os_system_time_0 303
#define BIF_os_system_time_1 304
#define BIF_os_perf_counter_0 305
#define BIF_erl_ddll_try_load_3 306
#define BIF_erl_ddll_try_unload_2 307
#define BIF_erl_ddll_loaded_drivers_0 308
#define BIF_erl_ddll_info_2 309
#define BIF_erl_ddll_format_error_int_1 310
#define BIF_erl_ddll_monitor_2 311
#define BIF_erl_ddll_demonitor_1 312
#define BIF_re_version_0 313
#define BIF_re_compile_1 314
#define BIF_re_compile_2 315
#define BIF_re_run_2 316
#define BIF_re_run_3 317
#define BIF_re_internal_run_4 318
#define BIF_lists_member_2 319
#define BIF_lists_reverse_2 320
#define BIF_lists_keymember_3 321
#define BIF_lists_keysearch_3 322
#define BIF_lists_keyfind_3 323
#define BIF_erts_debug_disassemble_1 324
#define BIF_erts_debug_breakpoint_2 325
#define BIF_erts_debug_same_2 326
#define BIF_erts_debug_flat_size_1 327
#define BIF_erts_debug_get_internal_state_1 328
#define BIF_erts_debug_set_internal_state_2 329
#define BIF_erts_debug_dist_ext_to_term_2 330
#define BIF_erts_debug_instructions_0 331
#define BIF_erts_debug_interpreter_size_0 332
#define BIF_erts_debug_dirty_cpu_2 333
#define BIF_erts_debug_dirty_io_2 334
#define BIF_erts_debug_dirty_3 335
#define BIF_erts_debug_lcnt_control_2 336
#define BIF_erts_debug_lcnt_control_1 337
#define BIF_erts_debug_lcnt_collect_0 338
#define BIF_erts_debug_lcnt_clear_0 339
#define BIF_hibernate_3 340
#define BIF_error_logger_warning_map_0 341
#define BIF_get_module_info_1 342
#define BIF_get_module_info_2 343
#define BIF_is_boolean_1 344
#define BIF_string_list_to_float_1 345
#define BIF_make_fun_3 346
#define BIF_iolist_size_1 347
#define BIF_iolist_to_binary_1 348
#define BIF_list_to_existing_atom_1 349
#define BIF_is_bitstring_1 350
#define BIF_tuple_size_1 351
#define BIF_byte_size_1 352
#define BIF_bit_size_1 353
#define BIF_list_to_bitstring_1 354
#define BIF_bitstring_to_list_1 355
#define BIF_ets_update_element_3 356
#define BIF_decode_packet_3 357
#define BIF_unicode_characters_to_binary_2 358
#define BIF_unicode_characters_to_list_2 359
#define BIF_unicode_bin_is_7bit_1 360
#define BIF_atom_to_binary_2 361
#define BIF_binary_to_atom_2 362
#define BIF_binary_to_existing_atom_2 363
#define BIF_net_kernel_dflag_unicode_io_1 364
#define BIF_ets_give_away_3 365
#define BIF_ets_setopts_2 366
#define BIF_load_nif_2 367
#define BIF_call_on_load_function_1 368
#define BIF_finish_after_on_load_2 369
#define BIF_binary_to_term_2 370
#define BIF_binary_part_2 371
#define BIF_binary_part_3 372
#define BIF_binary_compile_pattern_1 373
#define BIF_binary_match_2 374
#define BIF_binary_match_3 375
#define BIF_binary_matches_2 376
#define BIF_binary_matches_3 377
#define BIF_binary_longest_common_prefix_1 378
#define BIF_binary_longest_common_suffix_1 379
#define BIF_binary_first_1 380
#define BIF_binary_last_1 381
#define BIF_binary_at_2 382
#define BIF_binary_binary_part_2 383
#define BIF_binary_binary_part_3 384
#define BIF_binary_list_to_bin_1 385
#define BIF_binary_copy_1 386
#define BIF_binary_copy_2 387
#define BIF_binary_referenced_byte_size_1 388
#define BIF_binary_encode_unsigned_1 389
#define BIF_binary_encode_unsigned_2 390
#define BIF_binary_decode_unsigned_1 391
#define BIF_binary_decode_unsigned_2 392
#define BIF_nif_error_1 393
#define BIF_nif_error_2 394
#define BIF_prim_file_internal_name2native_1 395
#define BIF_prim_file_internal_native2name_1 396
#define BIF_prim_file_internal_normalize_utf8_1 397
#define BIF_prim_file_is_translatable_1 398
#define BIF_file_native_name_encoding_0 399
#define BIF_check_old_code_1 400
#define BIF_universaltime_to_posixtime_1 401
#define BIF_posixtime_to_universaltime_1 402
#define BIF_dt_put_tag_1 403
#define BIF_dt_get_tag_0 404
#define BIF_dt_get_tag_data_0 405
#define BIF_dt_spread_tag_1 406
#define BIF_dt_restore_tag_1 407
#define BIF_dt_prepend_vm_tag_data_1 408
#define BIF_dt_append_vm_tag_data_1 409
#define BIF_finish_loading_1 410
#define BIF_insert_element_3 411
#define BIF_delete_element_2 412
#define BIF_integer_to_binary_1 413
#define BIF_float_to_binary_1 414
#define BIF_float_to_binary_2 415
#define BIF_binary_to_float_1 416
#define BIF_io_printable_range_0 417
#define BIF_re_inspect_2 418
#define BIF_is_map_1 419
#define BIF_map_size_1 420
#define BIF_maps_find_2 421
#define BIF_maps_get_2 422
#define BIF_maps_from_list_1 423
#define BIF_maps_is_key_2 424
#define BIF_maps_keys_1 425
#define BIF_maps_merge_2 426
#define BIF_maps_put_3 427
#define BIF_maps_remove_2 428
#define BIF_maps_update_3 429
#define BIF_maps_values_1 430
#define BIF_erts_internal_cmp_term_2 431
#define BIF_ets_take_2 432
#define BIF_fun_info_mfa_1 433
#define BIF_get_keys_0 434
#define BIF_ets_update_counter_4 435
#define BIF_erts_debug_map_info_1 436
#define BIF_erts_internal_is_process_executing_dirty_1 437
#define BIF_erts_internal_check_dirty_process_code_2 438
#define BIF_erts_internal_purge_module_2 439
#define BIF_binary_split_2 440
#define BIF_binary_split_3 441
#define BIF_erts_debug_size_shared_1 442
#define BIF_erts_debug_copy_shared_2 443
#define BIF_has_prepared_code_on_load_1 444
#define BIF_maps_take_2 445
#define BIF_floor_1 446
#define BIF_ceil_1 447
#define BIF_math_floor_1 448
#define BIF_math_ceil_1 449
#define BIF_math_fmod_2 450
#define BIF_os_set_signal_2 451
#define BIF_iolist_to_iovec_1 452
#define BIF_erts_internal_get_dflags_0 453
#define BIF_erts_internal_new_connection_1 454
#define BIF_erts_internal_map_next_3 455
#define BIF_ets_whereis_1 456
#define BIF_erts_internal_gather_alloc_histograms_1 457
#define BIF_erts_internal_gather_carrier_info_1 458
#define BIF_map_get_2 459
#define BIF_is_map_key_2 460
#define BIF_ets_internal_delete_all_2 461
#define BIF_ets_internal_select_delete_2 462
#define BIF_persistent_term_put_2 463
#define BIF_persistent_term_get_1 464
#define BIF_persistent_term_get_0 465
#define BIF_persistent_term_erase_1 466
#define BIF_persistent_term_info_0 467
#define BIF_erts_internal_erase_persistent_terms_0 468
#define BIF_erts_internal_atomics_new_2 469
#define BIF_atomics_get_2 470
#define BIF_atomics_put_3 471
#define BIF_atomics_add_3 472
#define BIF_atomics_add_get_3 473
#define BIF_atomics_exchange_3 474
#define BIF_atomics_compare_exchange_4 475
#define BIF_atomics_info_1 476
#define BIF_erts_internal_counters_new_1 477
#define BIF_erts_internal_counters_get_2 478
#define BIF_erts_internal_counters_add_3 479
#define BIF_erts_internal_counters_put_3 480
#define BIF_erts_internal_counters_info_1 481
#define BIF_erts_internal_spawn_system_process_3 482
#define BIF_integer_to_list_2 483
#define BIF_integer_to_binary_2 484
#define BIF_persistent_term_get_2 485
#define BIF_erts_internal_ets_lookup_binary_info_2 486
#define BIF_erts_internal_ets_raw_first_1 487
#define BIF_erts_internal_ets_raw_next_2 488
#define BIF_erts_internal_abort_pending_connection_2 489
#define BIF_erts_internal_get_creation_0 490
#define BIF_erts_internal_prepare_loading_2 491
#define BIF_erts_internal_beamfile_chunk_2 492
#define BIF_erts_internal_beamfile_module_md5_1 493
#define BIF_os_env_0 494
#define BIF_alias_1 495
#define BIF_unalias_1 496
#define BIF_monitor_3 497
#define BIF_error_3 498
#define BIF_maps_from_keys_2 499
#define BIF_erts_internal_binary_to_integer_2 500
#define BIF_erts_internal_list_to_integer_2 501
#define BIF_min_2 502
#define BIF_max_2 503
#define BIF_erts_internal_term_to_string_2 504
#define BIF_code_coverage_support_0 505
#define BIF_code_get_coverage_mode_0 506
#define BIF_code_get_coverage_mode_1 507
#define BIF_code_get_coverage_2 508
#define BIF_code_reset_coverage_1 509
#define BIF_code_set_coverage_mode_1 510
#define BIF_ets_update_element_4 511
#define BIF_erts_internal_trace_session_create_3 512
#define BIF_erts_internal_trace_session_destroy_1 513
#define BIF_erts_internal_trace_4 514
#define BIF_erts_internal_trace_pattern_4 515
#define BIF_erts_internal_trace_info_3 516
#define BIF_erts_trace_cleaner_check_0 517
#define BIF_erts_trace_cleaner_send_trace_clean_signal_1 518
#define BIF_erts_internal_system_monitor_1 519
#define BIF_erts_internal_system_monitor_3 520
#define BIF_erts_internal_processes_next_1 521
#define BIF_code_get_debug_info_1 522
#define BIF_exit_3 523
#define BIF_link_2 524
#define BIF_erl_debugger_supported_0 525
#define BIF_erl_debugger_instrumentations_0 526
#define BIF_erl_debugger_toggle_instrumentations_1 527
#define BIF_erl_debugger_register_1 528
#define BIF_erl_debugger_unregister_2 529
#define BIF_erl_debugger_whereis_0 530
#define BIF_erl_debugger_breakpoint_3 531
#define BIF_erl_debugger_breakpoints_1 532
#define BIF_erl_debugger_breakpoints_3 533
#define BIF_erts_internal_notify_breakpoint_hit_3 534
#define BIF_erl_debugger_stack_frames_2 535
#define BIF_erl_debugger_peek_stack_frame_slot_4 536
#define BIF_erl_debugger_xregs_count_1 537
#define BIF_erl_debugger_peek_xreg_3 538
#define BIF_erts_debug_unaligned_bitstring_2 539
#define BIF_re_import_1 540
#define BIF_persistent_term_put_new_2 541

Eterm abs_1(Process*, Eterm*, ErtsCodePtr);

Eterm adler32_1(Process*, Eterm*, ErtsCodePtr);

Eterm adler32_2(Process*, Eterm*, ErtsCodePtr);

Eterm adler32_combine_3(Process*, Eterm*, ErtsCodePtr);

Eterm atom_to_list_1(Process*, Eterm*, ErtsCodePtr);

Eterm binary_to_list_1(Process*, Eterm*, ErtsCodePtr);

Eterm binary_to_list_3(Process*, Eterm*, ErtsCodePtr);

Eterm binary_to_term_1(Process*, Eterm*, ErtsCodePtr);

Eterm crc32_1(Process*, Eterm*, ErtsCodePtr);

Eterm crc32_2(Process*, Eterm*, ErtsCodePtr);

Eterm crc32_combine_3(Process*, Eterm*, ErtsCodePtr);

Eterm date_0(Process*, Eterm*, ErtsCodePtr);

Eterm delete_module_1(Process*, Eterm*, ErtsCodePtr);

Eterm display_1(Process*, Eterm*, ErtsCodePtr);

Eterm display_string_2(Process*, Eterm*, ErtsCodePtr);
Eterm schedule_dirty_cpu_display_string_2(Process*, Eterm*, ErtsCodePtr);

Eterm element_2(Process*, Eterm*, ErtsCodePtr);

Eterm erase_0(Process*, Eterm*, ErtsCodePtr);

Eterm erase_1(Process*, Eterm*, ErtsCodePtr);

Eterm exit_1(Process*, Eterm*, ErtsCodePtr);

Eterm exit_2(Process*, Eterm*, ErtsCodePtr);

Eterm exit_signal_2(Process*, Eterm*, ErtsCodePtr);

Eterm external_size_1(Process*, Eterm*, ErtsCodePtr);

Eterm external_size_2(Process*, Eterm*, ErtsCodePtr);

Eterm float_1(Process*, Eterm*, ErtsCodePtr);

Eterm float_to_list_1(Process*, Eterm*, ErtsCodePtr);

Eterm float_to_list_2(Process*, Eterm*, ErtsCodePtr);

Eterm fun_info_2(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_garbage_collect_1(Process*, Eterm*, ErtsCodePtr);

Eterm get_0(Process*, Eterm*, ErtsCodePtr);

Eterm get_1(Process*, Eterm*, ErtsCodePtr);

Eterm get_keys_1(Process*, Eterm*, ErtsCodePtr);

Eterm group_leader_0(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_group_leader_2(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_group_leader_3(Process*, Eterm*, ErtsCodePtr);

Eterm halt_2(Process*, Eterm*, ErtsCodePtr);

Eterm phash_2(Process*, Eterm*, ErtsCodePtr);

Eterm phash2_1(Process*, Eterm*, ErtsCodePtr);

Eterm phash2_2(Process*, Eterm*, ErtsCodePtr);

Eterm hd_1(Process*, Eterm*, ErtsCodePtr);

Eterm integer_to_list_1(Process*, Eterm*, ErtsCodePtr);

Eterm length_1(Process*, Eterm*, ErtsCodePtr);

Eterm link_1(Process*, Eterm*, ErtsCodePtr);

Eterm list_to_atom_1(Process*, Eterm*, ErtsCodePtr);

Eterm list_to_binary_1(Process*, Eterm*, ErtsCodePtr);

Eterm list_to_float_1(Process*, Eterm*, ErtsCodePtr);

Eterm list_to_pid_1(Process*, Eterm*, ErtsCodePtr);

Eterm list_to_port_1(Process*, Eterm*, ErtsCodePtr);

Eterm list_to_ref_1(Process*, Eterm*, ErtsCodePtr);

Eterm list_to_tuple_1(Process*, Eterm*, ErtsCodePtr);

Eterm loaded_0(Process*, Eterm*, ErtsCodePtr);

Eterm localtime_0(Process*, Eterm*, ErtsCodePtr);

Eterm localtime_to_universaltime_2(Process*, Eterm*, ErtsCodePtr);

Eterm make_ref_0(Process*, Eterm*, ErtsCodePtr);

Eterm unique_integer_0(Process*, Eterm*, ErtsCodePtr);

Eterm unique_integer_1(Process*, Eterm*, ErtsCodePtr);

Eterm md5_1(Process*, Eterm*, ErtsCodePtr);

Eterm md5_init_0(Process*, Eterm*, ErtsCodePtr);

Eterm md5_update_2(Process*, Eterm*, ErtsCodePtr);

Eterm md5_final_1(Process*, Eterm*, ErtsCodePtr);

Eterm module_loaded_1(Process*, Eterm*, ErtsCodePtr);

Eterm function_exported_3(Process*, Eterm*, ErtsCodePtr);

Eterm monitor_node_2(Process*, Eterm*, ErtsCodePtr);

Eterm monitor_node_3(Process*, Eterm*, ErtsCodePtr);

Eterm node_1(Process*, Eterm*, ErtsCodePtr);

Eterm node_0(Process*, Eterm*, ErtsCodePtr);

Eterm nodes_0(Process*, Eterm*, ErtsCodePtr);

Eterm nodes_1(Process*, Eterm*, ErtsCodePtr);

Eterm nodes_2(Process*, Eterm*, ErtsCodePtr);

Eterm now_0(Process*, Eterm*, ErtsCodePtr);

Eterm monotonic_time_0(Process*, Eterm*, ErtsCodePtr);

Eterm monotonic_time_1(Process*, Eterm*, ErtsCodePtr);

Eterm system_time_0(Process*, Eterm*, ErtsCodePtr);

Eterm system_time_1(Process*, Eterm*, ErtsCodePtr);

Eterm time_offset_0(Process*, Eterm*, ErtsCodePtr);

Eterm time_offset_1(Process*, Eterm*, ErtsCodePtr);

Eterm timestamp_0(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_open_port_2(Process*, Eterm*, ErtsCodePtr);

Eterm pid_to_list_1(Process*, Eterm*, ErtsCodePtr);

Eterm ports_0(Process*, Eterm*, ErtsCodePtr);

Eterm pre_loaded_0(Process*, Eterm*, ErtsCodePtr);

Eterm process_flag_2(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_process_flag_3(Process*, Eterm*, ErtsCodePtr);

Eterm process_info_1(Process*, Eterm*, ErtsCodePtr);

Eterm process_info_2(Process*, Eterm*, ErtsCodePtr);

Eterm processes_0(Process*, Eterm*, ErtsCodePtr);

Eterm put_2(Process*, Eterm*, ErtsCodePtr);

Eterm register_2(Process*, Eterm*, ErtsCodePtr);

Eterm registered_0(Process*, Eterm*, ErtsCodePtr);

Eterm round_1(Process*, Eterm*, ErtsCodePtr);

Eterm self_0(Process*, Eterm*, ErtsCodePtr);

Eterm setelement_3(Process*, Eterm*, ErtsCodePtr);

Eterm size_1(Process*, Eterm*, ErtsCodePtr);

Eterm spawn_3(Process*, Eterm*, ErtsCodePtr);

Eterm spawn_link_3(Process*, Eterm*, ErtsCodePtr);

Eterm split_binary_2(Process*, Eterm*, ErtsCodePtr);

Eterm statistics_1(Process*, Eterm*, ErtsCodePtr);

Eterm term_to_binary_1(Process*, Eterm*, ErtsCodePtr);

Eterm term_to_binary_2(Process*, Eterm*, ErtsCodePtr);

Eterm term_to_iovec_1(Process*, Eterm*, ErtsCodePtr);

Eterm term_to_iovec_2(Process*, Eterm*, ErtsCodePtr);

Eterm throw_1(Process*, Eterm*, ErtsCodePtr);

Eterm time_0(Process*, Eterm*, ErtsCodePtr);

Eterm tl_1(Process*, Eterm*, ErtsCodePtr);

Eterm trunc_1(Process*, Eterm*, ErtsCodePtr);

Eterm tuple_to_list_1(Process*, Eterm*, ErtsCodePtr);

Eterm universaltime_0(Process*, Eterm*, ErtsCodePtr);

Eterm universaltime_to_localtime_1(Process*, Eterm*, ErtsCodePtr);

Eterm unlink_1(Process*, Eterm*, ErtsCodePtr);

Eterm unregister_1(Process*, Eterm*, ErtsCodePtr);

Eterm whereis_1(Process*, Eterm*, ErtsCodePtr);

Eterm spawn_opt_4(Process*, Eterm*, ErtsCodePtr);

Eterm setnode_2(Process*, Eterm*, ErtsCodePtr);

Eterm dist_get_stat_1(Process*, Eterm*, ErtsCodePtr);

Eterm dist_ctrl_input_handler_2(Process*, Eterm*, ErtsCodePtr);

Eterm dist_ctrl_put_data_2(Process*, Eterm*, ErtsCodePtr);

Eterm dist_ctrl_get_data_1(Process*, Eterm*, ErtsCodePtr);

Eterm dist_ctrl_get_data_notification_1(Process*, Eterm*, ErtsCodePtr);

Eterm dist_ctrl_get_opt_2(Process*, Eterm*, ErtsCodePtr);

Eterm dist_ctrl_set_opt_3(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_port_info_1(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_port_info_2(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_port_call_3(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_port_command_3(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_port_control_3(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_port_close_1(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_port_connect_2(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_request_system_task_3(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_request_system_task_4(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_check_process_code_1(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_map_to_tuple_keys_1(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_term_type_1(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_map_hashmap_children_1(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_time_unit_0(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_perf_counter_unit_0(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_is_system_process_1(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_system_check_1(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_scheduler_wall_time_1(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_dirty_process_handle_signals_1(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_create_dist_channel_3(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_ets_super_user_1(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_spawn_request_4(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_dist_spawn_request_4(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_no_aux_work_threads_0(Process*, Eterm*, ErtsCodePtr);

Eterm spawn_request_abandon_1(Process*, Eterm*, ErtsCodePtr);

Eterm erts_literal_area_collector_release_area_switch_0(Process*, Eterm*, ErtsCodePtr);

Eterm erts_literal_area_collector_send_copy_request_3(Process*, Eterm*, ErtsCodePtr);

Eterm port_set_data_2(Process*, Eterm*, ErtsCodePtr);

Eterm port_get_data_1(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_trace_pattern_3(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_trace_3(Process*, Eterm*, ErtsCodePtr);

Eterm trace_info_2(Process*, Eterm*, ErtsCodePtr);

Eterm trace_delivered_1(Process*, Eterm*, ErtsCodePtr);

Eterm seq_trace_2(Process*, Eterm*, ErtsCodePtr);

Eterm seq_trace_info_1(Process*, Eterm*, ErtsCodePtr);

Eterm seq_trace_print_1(Process*, Eterm*, ErtsCodePtr);

Eterm seq_trace_print_2(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_suspend_process_2(Process*, Eterm*, ErtsCodePtr);

Eterm resume_process_1(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_process_display_2(Process*, Eterm*, ErtsCodePtr);

Eterm bump_reductions_1(Process*, Eterm*, ErtsCodePtr);

Eterm math_cos_1(Process*, Eterm*, ErtsCodePtr);

Eterm math_cosh_1(Process*, Eterm*, ErtsCodePtr);

Eterm math_sin_1(Process*, Eterm*, ErtsCodePtr);

Eterm math_sinh_1(Process*, Eterm*, ErtsCodePtr);

Eterm math_tan_1(Process*, Eterm*, ErtsCodePtr);

Eterm math_tanh_1(Process*, Eterm*, ErtsCodePtr);

Eterm math_acos_1(Process*, Eterm*, ErtsCodePtr);

Eterm math_acosh_1(Process*, Eterm*, ErtsCodePtr);

Eterm math_asin_1(Process*, Eterm*, ErtsCodePtr);

Eterm math_asinh_1(Process*, Eterm*, ErtsCodePtr);

Eterm math_atan_1(Process*, Eterm*, ErtsCodePtr);

Eterm math_atanh_1(Process*, Eterm*, ErtsCodePtr);

Eterm math_erf_1(Process*, Eterm*, ErtsCodePtr);

Eterm math_erfc_1(Process*, Eterm*, ErtsCodePtr);

Eterm math_exp_1(Process*, Eterm*, ErtsCodePtr);

Eterm math_log_1(Process*, Eterm*, ErtsCodePtr);

Eterm math_log2_1(Process*, Eterm*, ErtsCodePtr);

Eterm math_log10_1(Process*, Eterm*, ErtsCodePtr);

Eterm math_sqrt_1(Process*, Eterm*, ErtsCodePtr);

Eterm math_atan2_2(Process*, Eterm*, ErtsCodePtr);

Eterm math_pow_2(Process*, Eterm*, ErtsCodePtr);

Eterm start_timer_3(Process*, Eterm*, ErtsCodePtr);

Eterm start_timer_4(Process*, Eterm*, ErtsCodePtr);

Eterm send_after_3(Process*, Eterm*, ErtsCodePtr);

Eterm send_after_4(Process*, Eterm*, ErtsCodePtr);

Eterm cancel_timer_1(Process*, Eterm*, ErtsCodePtr);

Eterm cancel_timer_2(Process*, Eterm*, ErtsCodePtr);

Eterm read_timer_1(Process*, Eterm*, ErtsCodePtr);

Eterm read_timer_2(Process*, Eterm*, ErtsCodePtr);

Eterm make_tuple_2(Process*, Eterm*, ErtsCodePtr);

Eterm append_element_2(Process*, Eterm*, ErtsCodePtr);

Eterm make_tuple_3(Process*, Eterm*, ErtsCodePtr);

Eterm system_flag_2(Process*, Eterm*, ErtsCodePtr);

Eterm system_info_1(Process*, Eterm*, ErtsCodePtr);

Eterm system_profile_2(Process*, Eterm*, ErtsCodePtr);

Eterm system_profile_0(Process*, Eterm*, ErtsCodePtr);

Eterm ref_to_list_1(Process*, Eterm*, ErtsCodePtr);

Eterm port_to_list_1(Process*, Eterm*, ErtsCodePtr);

Eterm fun_to_list_1(Process*, Eterm*, ErtsCodePtr);

Eterm monitor_2(Process*, Eterm*, ErtsCodePtr);

Eterm demonitor_1(Process*, Eterm*, ErtsCodePtr);

Eterm demonitor_2(Process*, Eterm*, ErtsCodePtr);

Eterm is_process_alive_1(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_is_process_alive_2(Process*, Eterm*, ErtsCodePtr);

Eterm error_1(Process*, Eterm*, ErtsCodePtr);

Eterm error_2(Process*, Eterm*, ErtsCodePtr);

Eterm raise_3(Process*, Eterm*, ErtsCodePtr);

Eterm is_builtin_3(Process*, Eterm*, ErtsCodePtr);

Eterm and_2(Process*, Eterm*, ErtsCodePtr);

Eterm or_2(Process*, Eterm*, ErtsCodePtr);

Eterm xor_2(Process*, Eterm*, ErtsCodePtr);

Eterm not_1(Process*, Eterm*, ErtsCodePtr);

Eterm sgt_2(Process*, Eterm*, ErtsCodePtr);

Eterm sge_2(Process*, Eterm*, ErtsCodePtr);

Eterm slt_2(Process*, Eterm*, ErtsCodePtr);

Eterm sle_2(Process*, Eterm*, ErtsCodePtr);

Eterm seq_2(Process*, Eterm*, ErtsCodePtr);

Eterm seqeq_2(Process*, Eterm*, ErtsCodePtr);

Eterm sneq_2(Process*, Eterm*, ErtsCodePtr);

Eterm sneqeq_2(Process*, Eterm*, ErtsCodePtr);

Eterm splus_2(Process*, Eterm*, ErtsCodePtr);

Eterm sminus_2(Process*, Eterm*, ErtsCodePtr);

Eterm stimes_2(Process*, Eterm*, ErtsCodePtr);

Eterm div_2(Process*, Eterm*, ErtsCodePtr);

Eterm intdiv_2(Process*, Eterm*, ErtsCodePtr);

Eterm rem_2(Process*, Eterm*, ErtsCodePtr);

Eterm bor_2(Process*, Eterm*, ErtsCodePtr);

Eterm band_2(Process*, Eterm*, ErtsCodePtr);

Eterm bxor_2(Process*, Eterm*, ErtsCodePtr);

Eterm bsl_2(Process*, Eterm*, ErtsCodePtr);

Eterm bsr_2(Process*, Eterm*, ErtsCodePtr);

Eterm bnot_1(Process*, Eterm*, ErtsCodePtr);

Eterm sminus_1(Process*, Eterm*, ErtsCodePtr);

Eterm splus_1(Process*, Eterm*, ErtsCodePtr);

Eterm ebif_bang_2(Process*, Eterm*, ErtsCodePtr);

Eterm send_2(Process*, Eterm*, ErtsCodePtr);

Eterm send_3(Process*, Eterm*, ErtsCodePtr);

Eterm ebif_plusplus_2(Process*, Eterm*, ErtsCodePtr);

Eterm append_2(Process*, Eterm*, ErtsCodePtr);

Eterm ebif_minusminus_2(Process*, Eterm*, ErtsCodePtr);

Eterm subtract_2(Process*, Eterm*, ErtsCodePtr);

Eterm is_atom_1(Process*, Eterm*, ErtsCodePtr);

Eterm is_list_1(Process*, Eterm*, ErtsCodePtr);

Eterm is_tuple_1(Process*, Eterm*, ErtsCodePtr);

Eterm is_float_1(Process*, Eterm*, ErtsCodePtr);

Eterm is_integer_1(Process*, Eterm*, ErtsCodePtr);

Eterm is_number_1(Process*, Eterm*, ErtsCodePtr);

Eterm is_pid_1(Process*, Eterm*, ErtsCodePtr);

Eterm is_port_1(Process*, Eterm*, ErtsCodePtr);

Eterm is_reference_1(Process*, Eterm*, ErtsCodePtr);

Eterm is_binary_1(Process*, Eterm*, ErtsCodePtr);

Eterm is_function_1(Process*, Eterm*, ErtsCodePtr);

Eterm is_function_2(Process*, Eterm*, ErtsCodePtr);

Eterm is_record_2(Process*, Eterm*, ErtsCodePtr);

Eterm is_record_3(Process*, Eterm*, ErtsCodePtr);

Eterm match_spec_test_3(Process*, Eterm*, ErtsCodePtr);

Eterm ets_internal_request_all_0(Process*, Eterm*, ErtsCodePtr);

Eterm ets_new_2(Process*, Eterm*, ErtsCodePtr);

Eterm ets_delete_1(Process*, Eterm*, ErtsCodePtr);

Eterm ets_delete_2(Process*, Eterm*, ErtsCodePtr);

Eterm ets_delete_object_2(Process*, Eterm*, ErtsCodePtr);

Eterm ets_first_1(Process*, Eterm*, ErtsCodePtr);

Eterm ets_first_lookup_1(Process*, Eterm*, ErtsCodePtr);

Eterm ets_is_compiled_ms_1(Process*, Eterm*, ErtsCodePtr);

Eterm ets_lookup_2(Process*, Eterm*, ErtsCodePtr);

Eterm ets_lookup_element_3(Process*, Eterm*, ErtsCodePtr);

Eterm ets_lookup_element_4(Process*, Eterm*, ErtsCodePtr);

Eterm ets_info_1(Process*, Eterm*, ErtsCodePtr);

Eterm ets_info_2(Process*, Eterm*, ErtsCodePtr);

Eterm ets_last_1(Process*, Eterm*, ErtsCodePtr);

Eterm ets_last_lookup_1(Process*, Eterm*, ErtsCodePtr);

Eterm ets_match_1(Process*, Eterm*, ErtsCodePtr);

Eterm ets_match_2(Process*, Eterm*, ErtsCodePtr);

Eterm ets_match_3(Process*, Eterm*, ErtsCodePtr);

Eterm ets_match_object_1(Process*, Eterm*, ErtsCodePtr);

Eterm ets_match_object_2(Process*, Eterm*, ErtsCodePtr);

Eterm ets_match_object_3(Process*, Eterm*, ErtsCodePtr);

Eterm ets_member_2(Process*, Eterm*, ErtsCodePtr);

Eterm ets_next_2(Process*, Eterm*, ErtsCodePtr);

Eterm ets_next_lookup_2(Process*, Eterm*, ErtsCodePtr);

Eterm ets_prev_2(Process*, Eterm*, ErtsCodePtr);

Eterm ets_prev_lookup_2(Process*, Eterm*, ErtsCodePtr);

Eterm ets_insert_2(Process*, Eterm*, ErtsCodePtr);

Eterm ets_insert_new_2(Process*, Eterm*, ErtsCodePtr);

Eterm ets_rename_2(Process*, Eterm*, ErtsCodePtr);

Eterm ets_safe_fixtable_2(Process*, Eterm*, ErtsCodePtr);

Eterm ets_slot_2(Process*, Eterm*, ErtsCodePtr);

Eterm ets_update_counter_3(Process*, Eterm*, ErtsCodePtr);

Eterm ets_select_1(Process*, Eterm*, ErtsCodePtr);

Eterm ets_select_2(Process*, Eterm*, ErtsCodePtr);

Eterm ets_select_3(Process*, Eterm*, ErtsCodePtr);

Eterm ets_select_count_2(Process*, Eterm*, ErtsCodePtr);

Eterm ets_select_reverse_1(Process*, Eterm*, ErtsCodePtr);

Eterm ets_select_reverse_2(Process*, Eterm*, ErtsCodePtr);

Eterm ets_select_reverse_3(Process*, Eterm*, ErtsCodePtr);

Eterm ets_select_replace_2(Process*, Eterm*, ErtsCodePtr);

Eterm ets_match_spec_compile_1(Process*, Eterm*, ErtsCodePtr);

Eterm ets_match_spec_run_r_3(Process*, Eterm*, ErtsCodePtr);

Eterm os_getenv_1(Process*, Eterm*, ErtsCodePtr);

Eterm os_putenv_2(Process*, Eterm*, ErtsCodePtr);

Eterm os_unsetenv_1(Process*, Eterm*, ErtsCodePtr);

Eterm os_getpid_0(Process*, Eterm*, ErtsCodePtr);

Eterm os_timestamp_0(Process*, Eterm*, ErtsCodePtr);

Eterm os_system_time_0(Process*, Eterm*, ErtsCodePtr);

Eterm os_system_time_1(Process*, Eterm*, ErtsCodePtr);

Eterm os_perf_counter_0(Process*, Eterm*, ErtsCodePtr);

Eterm erl_ddll_try_load_3(Process*, Eterm*, ErtsCodePtr);

Eterm erl_ddll_try_unload_2(Process*, Eterm*, ErtsCodePtr);

Eterm erl_ddll_loaded_drivers_0(Process*, Eterm*, ErtsCodePtr);

Eterm erl_ddll_info_2(Process*, Eterm*, ErtsCodePtr);

Eterm erl_ddll_format_error_int_1(Process*, Eterm*, ErtsCodePtr);

Eterm erl_ddll_monitor_2(Process*, Eterm*, ErtsCodePtr);

Eterm erl_ddll_demonitor_1(Process*, Eterm*, ErtsCodePtr);

Eterm re_version_0(Process*, Eterm*, ErtsCodePtr);

Eterm re_compile_1(Process*, Eterm*, ErtsCodePtr);

Eterm re_compile_2(Process*, Eterm*, ErtsCodePtr);

Eterm re_run_2(Process*, Eterm*, ErtsCodePtr);

Eterm re_run_3(Process*, Eterm*, ErtsCodePtr);

Eterm re_internal_run_4(Process*, Eterm*, ErtsCodePtr);

Eterm lists_member_2(Process*, Eterm*, ErtsCodePtr);

Eterm lists_reverse_2(Process*, Eterm*, ErtsCodePtr);

Eterm lists_keymember_3(Process*, Eterm*, ErtsCodePtr);

Eterm lists_keysearch_3(Process*, Eterm*, ErtsCodePtr);

Eterm lists_keyfind_3(Process*, Eterm*, ErtsCodePtr);

Eterm erts_debug_disassemble_1(Process*, Eterm*, ErtsCodePtr);

Eterm erts_debug_breakpoint_2(Process*, Eterm*, ErtsCodePtr);

Eterm erts_debug_same_2(Process*, Eterm*, ErtsCodePtr);

Eterm erts_debug_flat_size_1(Process*, Eterm*, ErtsCodePtr);

Eterm erts_debug_get_internal_state_1(Process*, Eterm*, ErtsCodePtr);

Eterm erts_debug_set_internal_state_2(Process*, Eterm*, ErtsCodePtr);

Eterm erts_debug_dist_ext_to_term_2(Process*, Eterm*, ErtsCodePtr);

Eterm erts_debug_instructions_0(Process*, Eterm*, ErtsCodePtr);

Eterm erts_debug_interpreter_size_0(Process*, Eterm*, ErtsCodePtr);

Eterm erts_debug_dirty_cpu_2(Process*, Eterm*, ErtsCodePtr);
Eterm schedule_dirty_cpu_erts_debug_dirty_cpu_2(Process*, Eterm*, ErtsCodePtr);

Eterm erts_debug_dirty_io_2(Process*, Eterm*, ErtsCodePtr);
Eterm schedule_dirty_io_erts_debug_dirty_io_2(Process*, Eterm*, ErtsCodePtr);

Eterm erts_debug_dirty_3(Process*, Eterm*, ErtsCodePtr);

Eterm erts_debug_lcnt_control_2(Process*, Eterm*, ErtsCodePtr);
Eterm schedule_dirty_cpu_erts_debug_lcnt_control_2(Process*, Eterm*, ErtsCodePtr);

Eterm erts_debug_lcnt_control_1(Process*, Eterm*, ErtsCodePtr);

Eterm erts_debug_lcnt_collect_0(Process*, Eterm*, ErtsCodePtr);
Eterm schedule_dirty_cpu_erts_debug_lcnt_collect_0(Process*, Eterm*, ErtsCodePtr);

Eterm erts_debug_lcnt_clear_0(Process*, Eterm*, ErtsCodePtr);
Eterm schedule_dirty_cpu_erts_debug_lcnt_clear_0(Process*, Eterm*, ErtsCodePtr);

Eterm hibernate_3(Process*, Eterm*, ErtsCodePtr);

Eterm error_logger_warning_map_0(Process*, Eterm*, ErtsCodePtr);

Eterm get_module_info_1(Process*, Eterm*, ErtsCodePtr);

Eterm get_module_info_2(Process*, Eterm*, ErtsCodePtr);

Eterm is_boolean_1(Process*, Eterm*, ErtsCodePtr);

Eterm string_list_to_float_1(Process*, Eterm*, ErtsCodePtr);

Eterm make_fun_3(Process*, Eterm*, ErtsCodePtr);

Eterm iolist_size_1(Process*, Eterm*, ErtsCodePtr);

Eterm iolist_to_binary_1(Process*, Eterm*, ErtsCodePtr);

Eterm list_to_existing_atom_1(Process*, Eterm*, ErtsCodePtr);

Eterm is_bitstring_1(Process*, Eterm*, ErtsCodePtr);

Eterm tuple_size_1(Process*, Eterm*, ErtsCodePtr);

Eterm byte_size_1(Process*, Eterm*, ErtsCodePtr);

Eterm bit_size_1(Process*, Eterm*, ErtsCodePtr);

Eterm list_to_bitstring_1(Process*, Eterm*, ErtsCodePtr);

Eterm bitstring_to_list_1(Process*, Eterm*, ErtsCodePtr);

Eterm ets_update_element_3(Process*, Eterm*, ErtsCodePtr);

Eterm decode_packet_3(Process*, Eterm*, ErtsCodePtr);

Eterm unicode_characters_to_binary_2(Process*, Eterm*, ErtsCodePtr);

Eterm unicode_characters_to_list_2(Process*, Eterm*, ErtsCodePtr);

Eterm unicode_bin_is_7bit_1(Process*, Eterm*, ErtsCodePtr);

Eterm atom_to_binary_2(Process*, Eterm*, ErtsCodePtr);

Eterm binary_to_atom_2(Process*, Eterm*, ErtsCodePtr);

Eterm binary_to_existing_atom_2(Process*, Eterm*, ErtsCodePtr);

Eterm net_kernel_dflag_unicode_io_1(Process*, Eterm*, ErtsCodePtr);

Eterm ets_give_away_3(Process*, Eterm*, ErtsCodePtr);

Eterm ets_setopts_2(Process*, Eterm*, ErtsCodePtr);

Eterm load_nif_2(Process*, Eterm*, ErtsCodePtr);

Eterm call_on_load_function_1(Process*, Eterm*, ErtsCodePtr);

Eterm finish_after_on_load_2(Process*, Eterm*, ErtsCodePtr);

Eterm binary_to_term_2(Process*, Eterm*, ErtsCodePtr);

Eterm binary_part_2(Process*, Eterm*, ErtsCodePtr);

Eterm binary_part_3(Process*, Eterm*, ErtsCodePtr);

Eterm binary_compile_pattern_1(Process*, Eterm*, ErtsCodePtr);

Eterm binary_match_2(Process*, Eterm*, ErtsCodePtr);

Eterm binary_match_3(Process*, Eterm*, ErtsCodePtr);

Eterm binary_matches_2(Process*, Eterm*, ErtsCodePtr);

Eterm binary_matches_3(Process*, Eterm*, ErtsCodePtr);

Eterm binary_longest_common_prefix_1(Process*, Eterm*, ErtsCodePtr);

Eterm binary_longest_common_suffix_1(Process*, Eterm*, ErtsCodePtr);

Eterm binary_first_1(Process*, Eterm*, ErtsCodePtr);

Eterm binary_last_1(Process*, Eterm*, ErtsCodePtr);

Eterm binary_at_2(Process*, Eterm*, ErtsCodePtr);

Eterm binary_binary_part_2(Process*, Eterm*, ErtsCodePtr);

Eterm binary_binary_part_3(Process*, Eterm*, ErtsCodePtr);

Eterm binary_list_to_bin_1(Process*, Eterm*, ErtsCodePtr);

Eterm binary_copy_1(Process*, Eterm*, ErtsCodePtr);

Eterm binary_copy_2(Process*, Eterm*, ErtsCodePtr);

Eterm binary_referenced_byte_size_1(Process*, Eterm*, ErtsCodePtr);

Eterm binary_encode_unsigned_1(Process*, Eterm*, ErtsCodePtr);

Eterm binary_encode_unsigned_2(Process*, Eterm*, ErtsCodePtr);

Eterm binary_decode_unsigned_1(Process*, Eterm*, ErtsCodePtr);

Eterm binary_decode_unsigned_2(Process*, Eterm*, ErtsCodePtr);

Eterm nif_error_1(Process*, Eterm*, ErtsCodePtr);

Eterm nif_error_2(Process*, Eterm*, ErtsCodePtr);

Eterm prim_file_internal_name2native_1(Process*, Eterm*, ErtsCodePtr);

Eterm prim_file_internal_native2name_1(Process*, Eterm*, ErtsCodePtr);

Eterm prim_file_internal_normalize_utf8_1(Process*, Eterm*, ErtsCodePtr);

Eterm prim_file_is_translatable_1(Process*, Eterm*, ErtsCodePtr);

Eterm file_native_name_encoding_0(Process*, Eterm*, ErtsCodePtr);

Eterm check_old_code_1(Process*, Eterm*, ErtsCodePtr);

Eterm universaltime_to_posixtime_1(Process*, Eterm*, ErtsCodePtr);

Eterm posixtime_to_universaltime_1(Process*, Eterm*, ErtsCodePtr);

Eterm dt_put_tag_1(Process*, Eterm*, ErtsCodePtr);

Eterm dt_get_tag_0(Process*, Eterm*, ErtsCodePtr);

Eterm dt_get_tag_data_0(Process*, Eterm*, ErtsCodePtr);

Eterm dt_spread_tag_1(Process*, Eterm*, ErtsCodePtr);

Eterm dt_restore_tag_1(Process*, Eterm*, ErtsCodePtr);

Eterm dt_prepend_vm_tag_data_1(Process*, Eterm*, ErtsCodePtr);

Eterm dt_append_vm_tag_data_1(Process*, Eterm*, ErtsCodePtr);

Eterm finish_loading_1(Process*, Eterm*, ErtsCodePtr);

Eterm insert_element_3(Process*, Eterm*, ErtsCodePtr);

Eterm delete_element_2(Process*, Eterm*, ErtsCodePtr);

Eterm integer_to_binary_1(Process*, Eterm*, ErtsCodePtr);

Eterm float_to_binary_1(Process*, Eterm*, ErtsCodePtr);

Eterm float_to_binary_2(Process*, Eterm*, ErtsCodePtr);

Eterm binary_to_float_1(Process*, Eterm*, ErtsCodePtr);

Eterm io_printable_range_0(Process*, Eterm*, ErtsCodePtr);

Eterm re_inspect_2(Process*, Eterm*, ErtsCodePtr);

Eterm is_map_1(Process*, Eterm*, ErtsCodePtr);

Eterm map_size_1(Process*, Eterm*, ErtsCodePtr);

Eterm maps_find_2(Process*, Eterm*, ErtsCodePtr);

Eterm maps_get_2(Process*, Eterm*, ErtsCodePtr);

Eterm maps_from_list_1(Process*, Eterm*, ErtsCodePtr);

Eterm maps_is_key_2(Process*, Eterm*, ErtsCodePtr);

Eterm maps_keys_1(Process*, Eterm*, ErtsCodePtr);

Eterm maps_merge_2(Process*, Eterm*, ErtsCodePtr);

Eterm maps_put_3(Process*, Eterm*, ErtsCodePtr);

Eterm maps_remove_2(Process*, Eterm*, ErtsCodePtr);

Eterm maps_update_3(Process*, Eterm*, ErtsCodePtr);

Eterm maps_values_1(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_cmp_term_2(Process*, Eterm*, ErtsCodePtr);

Eterm ets_take_2(Process*, Eterm*, ErtsCodePtr);

Eterm fun_info_mfa_1(Process*, Eterm*, ErtsCodePtr);

Eterm get_keys_0(Process*, Eterm*, ErtsCodePtr);

Eterm ets_update_counter_4(Process*, Eterm*, ErtsCodePtr);

Eterm erts_debug_map_info_1(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_is_process_executing_dirty_1(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_check_dirty_process_code_2(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_purge_module_2(Process*, Eterm*, ErtsCodePtr);

Eterm binary_split_2(Process*, Eterm*, ErtsCodePtr);

Eterm binary_split_3(Process*, Eterm*, ErtsCodePtr);

Eterm erts_debug_size_shared_1(Process*, Eterm*, ErtsCodePtr);

Eterm erts_debug_copy_shared_2(Process*, Eterm*, ErtsCodePtr);

Eterm has_prepared_code_on_load_1(Process*, Eterm*, ErtsCodePtr);

Eterm maps_take_2(Process*, Eterm*, ErtsCodePtr);

Eterm floor_1(Process*, Eterm*, ErtsCodePtr);

Eterm ceil_1(Process*, Eterm*, ErtsCodePtr);

Eterm math_floor_1(Process*, Eterm*, ErtsCodePtr);

Eterm math_ceil_1(Process*, Eterm*, ErtsCodePtr);

Eterm math_fmod_2(Process*, Eterm*, ErtsCodePtr);

Eterm os_set_signal_2(Process*, Eterm*, ErtsCodePtr);

Eterm iolist_to_iovec_1(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_get_dflags_0(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_new_connection_1(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_map_next_3(Process*, Eterm*, ErtsCodePtr);

Eterm ets_whereis_1(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_gather_alloc_histograms_1(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_gather_carrier_info_1(Process*, Eterm*, ErtsCodePtr);

Eterm map_get_2(Process*, Eterm*, ErtsCodePtr);

Eterm is_map_key_2(Process*, Eterm*, ErtsCodePtr);

Eterm ets_internal_delete_all_2(Process*, Eterm*, ErtsCodePtr);

Eterm ets_internal_select_delete_2(Process*, Eterm*, ErtsCodePtr);

Eterm persistent_term_put_2(Process*, Eterm*, ErtsCodePtr);

Eterm persistent_term_get_1(Process*, Eterm*, ErtsCodePtr);

Eterm persistent_term_get_0(Process*, Eterm*, ErtsCodePtr);

Eterm persistent_term_erase_1(Process*, Eterm*, ErtsCodePtr);

Eterm persistent_term_info_0(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_erase_persistent_terms_0(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_atomics_new_2(Process*, Eterm*, ErtsCodePtr);

Eterm atomics_get_2(Process*, Eterm*, ErtsCodePtr);

Eterm atomics_put_3(Process*, Eterm*, ErtsCodePtr);

Eterm atomics_add_3(Process*, Eterm*, ErtsCodePtr);

Eterm atomics_add_get_3(Process*, Eterm*, ErtsCodePtr);

Eterm atomics_exchange_3(Process*, Eterm*, ErtsCodePtr);

Eterm atomics_compare_exchange_4(Process*, Eterm*, ErtsCodePtr);

Eterm atomics_info_1(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_counters_new_1(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_counters_get_2(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_counters_add_3(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_counters_put_3(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_counters_info_1(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_spawn_system_process_3(Process*, Eterm*, ErtsCodePtr);

Eterm integer_to_list_2(Process*, Eterm*, ErtsCodePtr);

Eterm integer_to_binary_2(Process*, Eterm*, ErtsCodePtr);

Eterm persistent_term_get_2(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_ets_lookup_binary_info_2(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_ets_raw_first_1(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_ets_raw_next_2(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_abort_pending_connection_2(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_get_creation_0(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_prepare_loading_2(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_beamfile_chunk_2(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_beamfile_module_md5_1(Process*, Eterm*, ErtsCodePtr);

Eterm os_env_0(Process*, Eterm*, ErtsCodePtr);

Eterm alias_1(Process*, Eterm*, ErtsCodePtr);

Eterm unalias_1(Process*, Eterm*, ErtsCodePtr);

Eterm monitor_3(Process*, Eterm*, ErtsCodePtr);

Eterm error_3(Process*, Eterm*, ErtsCodePtr);

Eterm maps_from_keys_2(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_binary_to_integer_2(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_list_to_integer_2(Process*, Eterm*, ErtsCodePtr);

Eterm min_2(Process*, Eterm*, ErtsCodePtr);

Eterm max_2(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_term_to_string_2(Process*, Eterm*, ErtsCodePtr);

Eterm code_coverage_support_0(Process*, Eterm*, ErtsCodePtr);

Eterm code_get_coverage_mode_0(Process*, Eterm*, ErtsCodePtr);

Eterm code_get_coverage_mode_1(Process*, Eterm*, ErtsCodePtr);

Eterm code_get_coverage_2(Process*, Eterm*, ErtsCodePtr);

Eterm code_reset_coverage_1(Process*, Eterm*, ErtsCodePtr);

Eterm code_set_coverage_mode_1(Process*, Eterm*, ErtsCodePtr);

Eterm ets_update_element_4(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_trace_session_create_3(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_trace_session_destroy_1(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_trace_4(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_trace_pattern_4(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_trace_info_3(Process*, Eterm*, ErtsCodePtr);

Eterm erts_trace_cleaner_check_0(Process*, Eterm*, ErtsCodePtr);

Eterm erts_trace_cleaner_send_trace_clean_signal_1(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_system_monitor_1(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_system_monitor_3(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_processes_next_1(Process*, Eterm*, ErtsCodePtr);

Eterm code_get_debug_info_1(Process*, Eterm*, ErtsCodePtr);

Eterm exit_3(Process*, Eterm*, ErtsCodePtr);

Eterm link_2(Process*, Eterm*, ErtsCodePtr);

Eterm erl_debugger_supported_0(Process*, Eterm*, ErtsCodePtr);

Eterm erl_debugger_instrumentations_0(Process*, Eterm*, ErtsCodePtr);

Eterm erl_debugger_toggle_instrumentations_1(Process*, Eterm*, ErtsCodePtr);

Eterm erl_debugger_register_1(Process*, Eterm*, ErtsCodePtr);

Eterm erl_debugger_unregister_2(Process*, Eterm*, ErtsCodePtr);

Eterm erl_debugger_whereis_0(Process*, Eterm*, ErtsCodePtr);

Eterm erl_debugger_breakpoint_3(Process*, Eterm*, ErtsCodePtr);

Eterm erl_debugger_breakpoints_1(Process*, Eterm*, ErtsCodePtr);

Eterm erl_debugger_breakpoints_3(Process*, Eterm*, ErtsCodePtr);

Eterm erts_internal_notify_breakpoint_hit_3(Process*, Eterm*, ErtsCodePtr);

Eterm erl_debugger_stack_frames_2(Process*, Eterm*, ErtsCodePtr);

Eterm erl_debugger_peek_stack_frame_slot_4(Process*, Eterm*, ErtsCodePtr);

Eterm erl_debugger_xregs_count_1(Process*, Eterm*, ErtsCodePtr);

Eterm erl_debugger_peek_xreg_3(Process*, Eterm*, ErtsCodePtr);

Eterm erts_debug_unaligned_bitstring_2(Process*, Eterm*, ErtsCodePtr);

Eterm re_import_1(Process*, Eterm*, ErtsCodePtr);

Eterm persistent_term_put_new_2(Process*, Eterm*, ErtsCodePtr);


#endif
