/*
 * %CopyrightBegin%
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Copyright Ericsson AB 2008-2025. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * %CopyrightEnd%
*/
/***** This file is generated do not edit ****/

gl_fns_t gl_fns[] =
{
  {5000, "init_opengl", NULL, NULL, &ecb_init_opengl},
  {5001, "unused", NULL, NULL, NULL},
  {5002, "unused", NULL, NULL, NULL},
  {5003, "unused", NULL, NULL, NULL},
  {5004, "unused", NULL, NULL, NULL},
  {5005, "unused", NULL, NULL, NULL},
  {5006, "unused", NULL, NULL, NULL},
  {5007, "unused", NULL, NULL, NULL},
  {5008, "unused", NULL, NULL, NULL},
  {5009, "gluTesselate", NULL, NULL, &erl_tess_impl},
  {5010, "gluBuild1DMipmapLevels", NULL, &wegluBuild1DMipmapLevels, &ecb_gluBuild1DMipmapLevels},
  {5011, "gluBuild1DMipmaps", NULL, &wegluBuild1DMipmaps, &ecb_gluBuild1DMipmaps},
  {5012, "gluBuild2DMipmapLevels", NULL, &wegluBuild2DMipmapLevels, &ecb_gluBuild2DMipmapLevels},
  {5013, "gluBuild2DMipmaps", NULL, &wegluBuild2DMipmaps, &ecb_gluBuild2DMipmaps},
  {5014, "gluBuild3DMipmapLevels", NULL, &wegluBuild3DMipmapLevels, &ecb_gluBuild3DMipmapLevels},
  {5015, "gluBuild3DMipmaps", NULL, &wegluBuild3DMipmaps, &ecb_gluBuild3DMipmaps},
  {5016, "gluCheckExtension", NULL, &wegluCheckExtension, &ecb_gluCheckExtension},
  {5017, "gluCylinder", NULL, &wegluCylinder, &ecb_gluCylinder},
  {5018, "gluDeleteQuadric", NULL, &wegluDeleteQuadric, &ecb_gluDeleteQuadric},
  {5019, "gluDisk", NULL, &wegluDisk, &ecb_gluDisk},
  {5020, "gluErrorString", NULL, (void *) &wegluErrorString, &ecb_gluErrorString},
  {5021, "gluGetString", NULL, (void *) &wegluGetString, &ecb_gluGetString},
  {5022, "gluLookAt", NULL, &wegluLookAt, &ecb_gluLookAt},
  {5023, "gluNewQuadric", NULL, &wegluNewQuadric, &ecb_gluNewQuadric},
  {5024, "gluOrtho2D", NULL, &wegluOrtho2D, &ecb_gluOrtho2D},
  {5025, "gluPartialDisk", NULL, &wegluPartialDisk, &ecb_gluPartialDisk},
  {5026, "gluPerspective", NULL, &wegluPerspective, &ecb_gluPerspective},
  {5027, "gluPickMatrix", NULL, &wegluPickMatrix, &ecb_gluPickMatrix},
  {5028, "gluProject", NULL, &wegluProject, &ecb_gluProject},
  {5029, "gluQuadricDrawStyle", NULL, &wegluQuadricDrawStyle, &ecb_gluQuadricDrawStyle},
  {5030, "gluQuadricNormals", NULL, &wegluQuadricNormals, &ecb_gluQuadricNormals},
  {5031, "gluQuadricOrientation", NULL, &wegluQuadricOrientation, &ecb_gluQuadricOrientation},
  {5032, "gluQuadricTexture", NULL, &wegluQuadricTexture, &ecb_gluQuadricTexture},
  {5033, "gluScaleImage", NULL, &wegluScaleImage, &ecb_gluScaleImage},
  {5034, "gluSphere", NULL, &wegluSphere, &ecb_gluSphere},
  {5035, "gluUnProject", NULL, &wegluUnProject, &ecb_gluUnProject},
  {5036, "gluUnProject4", NULL, &wegluUnProject4, &ecb_gluUnProject4},
  {5037, "glClearIndex", NULL, &weglClearIndex, &ecb_glClearIndex},
  {5038, "glClearColor", NULL, &weglClearColor, &ecb_glClearColor},
  {5039, "glClear", NULL, &weglClear, &ecb_glClear},
  {5040, "glIndexMask", NULL, &weglIndexMask, &ecb_glIndexMask},
  {5041, "glColorMask", NULL, &weglColorMask, &ecb_glColorMask},
  {5042, "glAlphaFunc", NULL, &weglAlphaFunc, &ecb_glAlphaFunc},
  {5043, "glBlendFunc", NULL, &weglBlendFunc, &ecb_glBlendFunc},
  {5044, "glLogicOp", NULL, &weglLogicOp, &ecb_glLogicOp},
  {5045, "glCullFace", NULL, &weglCullFace, &ecb_glCullFace},
  {5046, "glFrontFace", NULL, &weglFrontFace, &ecb_glFrontFace},
  {5047, "glPointSize", NULL, &weglPointSize, &ecb_glPointSize},
  {5048, "glLineWidth", NULL, &weglLineWidth, &ecb_glLineWidth},
  {5049, "glLineStipple", NULL, &weglLineStipple, &ecb_glLineStipple},
  {5050, "glPolygonMode", NULL, &weglPolygonMode, &ecb_glPolygonMode},
  {5051, "glPolygonOffset", NULL, &weglPolygonOffset, &ecb_glPolygonOffset},
  {5052, "glPolygonStipple", NULL, &weglPolygonStipple, &ecb_glPolygonStipple},
  {5053, "glGetPolygonStipple", NULL, &weglGetPolygonStipple, &ecb_glGetPolygonStipple},
  {5054, "glEdgeFlag", NULL, &weglEdgeFlag, &ecb_glEdgeFlag},
  {5055, "glScissor", NULL, &weglScissor, &ecb_glScissor},
  {5056, "glClipPlane", NULL, &weglClipPlane, &ecb_glClipPlane},
  {5057, "glGetClipPlane", NULL, &weglGetClipPlane, &ecb_glGetClipPlane},
  {5058, "glDrawBuffer", NULL, &weglDrawBuffer, &ecb_glDrawBuffer},
  {5059, "glReadBuffer", NULL, &weglReadBuffer, &ecb_glReadBuffer},
  {5060, "glEnable", NULL, &weglEnable, &ecb_glEnable},
  {5061, "glDisable", NULL, &weglDisable, &ecb_glDisable},
  {5062, "glIsEnabled", NULL, &weglIsEnabled, &ecb_glIsEnabled},
  {5063, "glEnableClientState", NULL, &weglEnableClientState, &ecb_glEnableClientState},
  {5064, "glDisableClientState", NULL, &weglDisableClientState, &ecb_glDisableClientState},
  {5065, "glGetBooleanv", NULL, &weglGetBooleanv, &ecb_glGetBooleanv},
  {5066, "glGetDoublev", NULL, &weglGetDoublev, &ecb_glGetDoublev},
  {5067, "glGetFloatv", NULL, &weglGetFloatv, &ecb_glGetFloatv},
  {5068, "glGetIntegerv", NULL, &weglGetIntegerv, &ecb_glGetIntegerv},
  {5069, "glPushAttrib", NULL, &weglPushAttrib, &ecb_glPushAttrib},
  {5070, "glPopAttrib", NULL, &weglPopAttrib, &ecb_glPopAttrib},
  {5071, "glPushClientAttrib", NULL, &weglPushClientAttrib, &ecb_glPushClientAttrib},
  {5072, "glPopClientAttrib", NULL, &weglPopClientAttrib, &ecb_glPopClientAttrib},
  {5073, "glRenderMode", NULL, &weglRenderMode, &ecb_glRenderMode},
  {5074, "glGetError", NULL, &weglGetError, &ecb_glGetError},
  {5075, "glGetString", NULL, (void *) &weglGetString, &ecb_glGetString},
  {5076, "glFinish", NULL, &weglFinish, &ecb_glFinish},
  {5077, "glFlush", NULL, &weglFlush, &ecb_glFlush},
  {5078, "glHint", NULL, &weglHint, &ecb_glHint},
  {5079, "glClearDepth", NULL, &weglClearDepth, &ecb_glClearDepth},
  {5080, "glDepthFunc", NULL, &weglDepthFunc, &ecb_glDepthFunc},
  {5081, "glDepthMask", NULL, &weglDepthMask, &ecb_glDepthMask},
  {5082, "glDepthRange", NULL, &weglDepthRange, &ecb_glDepthRange},
  {5083, "glClearAccum", NULL, &weglClearAccum, &ecb_glClearAccum},
  {5084, "glAccum", NULL, &weglAccum, &ecb_glAccum},
  {5085, "glMatrixMode", NULL, &weglMatrixMode, &ecb_glMatrixMode},
  {5086, "glOrtho", NULL, &weglOrtho, &ecb_glOrtho},
  {5087, "glFrustum", NULL, &weglFrustum, &ecb_glFrustum},
  {5088, "glViewport", NULL, &weglViewport, &ecb_glViewport},
  {5089, "glPushMatrix", NULL, &weglPushMatrix, &ecb_glPushMatrix},
  {5090, "glPopMatrix", NULL, &weglPopMatrix, &ecb_glPopMatrix},
  {5091, "glLoadIdentity", NULL, &weglLoadIdentity, &ecb_glLoadIdentity},
  {5092, "glLoadMatrixd", NULL, &weglLoadMatrixd, &ecb_glLoadMatrixd},
  {5093, "glLoadMatrixf", NULL, &weglLoadMatrixf, &ecb_glLoadMatrixf},
  {5094, "glMultMatrixd", NULL, &weglMultMatrixd, &ecb_glMultMatrixd},
  {5095, "glMultMatrixf", NULL, &weglMultMatrixf, &ecb_glMultMatrixf},
  {5096, "glRotated", NULL, &weglRotated, &ecb_glRotated},
  {5097, "glRotatef", NULL, &weglRotatef, &ecb_glRotatef},
  {5098, "glScaled", NULL, &weglScaled, &ecb_glScaled},
  {5099, "glScalef", NULL, &weglScalef, &ecb_glScalef},
  {5100, "glTranslated", NULL, &weglTranslated, &ecb_glTranslated},
  {5101, "glTranslatef", NULL, &weglTranslatef, &ecb_glTranslatef},
  {5102, "glIsList", NULL, &weglIsList, &ecb_glIsList},
  {5103, "glDeleteLists", NULL, &weglDeleteLists, &ecb_glDeleteLists},
  {5104, "glGenLists", NULL, &weglGenLists, &ecb_glGenLists},
  {5105, "glNewList", NULL, &weglNewList, &ecb_glNewList},
  {5106, "glEndList", NULL, &weglEndList, &ecb_glEndList},
  {5107, "glCallList", NULL, &weglCallList, &ecb_glCallList},
  {5108, "glCallLists", NULL, &weglCallLists, &ecb_glCallLists},
  {5109, "glListBase", NULL, &weglListBase, &ecb_glListBase},
  {5110, "glBegin", NULL, &weglBegin, &ecb_glBegin},
  {5111, "glEnd", NULL, &weglEnd, &ecb_glEnd},
  {5112, "glVertex2d", NULL, &weglVertex2d, &ecb_glVertex2d},
  {5113, "glVertex2f", NULL, &weglVertex2f, &ecb_glVertex2f},
  {5114, "glVertex2i", NULL, &weglVertex2i, &ecb_glVertex2i},
  {5115, "glVertex2s", NULL, &weglVertex2s, &ecb_glVertex2s},
  {5116, "glVertex3d", NULL, &weglVertex3d, &ecb_glVertex3d},
  {5117, "glVertex3f", NULL, &weglVertex3f, &ecb_glVertex3f},
  {5118, "glVertex3i", NULL, &weglVertex3i, &ecb_glVertex3i},
  {5119, "glVertex3s", NULL, &weglVertex3s, &ecb_glVertex3s},
  {5120, "glVertex4d", NULL, &weglVertex4d, &ecb_glVertex4d},
  {5121, "glVertex4f", NULL, &weglVertex4f, &ecb_glVertex4f},
  {5122, "glVertex4i", NULL, &weglVertex4i, &ecb_glVertex4i},
  {5123, "glVertex4s", NULL, &weglVertex4s, &ecb_glVertex4s},
  {5124, "glNormal3b", NULL, &weglNormal3b, &ecb_glNormal3b},
  {5125, "glNormal3d", NULL, &weglNormal3d, &ecb_glNormal3d},
  {5126, "glNormal3f", NULL, &weglNormal3f, &ecb_glNormal3f},
  {5127, "glNormal3i", NULL, &weglNormal3i, &ecb_glNormal3i},
  {5128, "glNormal3s", NULL, &weglNormal3s, &ecb_glNormal3s},
  {5129, "glIndexd", NULL, &weglIndexd, &ecb_glIndexd},
  {5130, "glIndexf", NULL, &weglIndexf, &ecb_glIndexf},
  {5131, "glIndexi", NULL, &weglIndexi, &ecb_glIndexi},
  {5132, "glIndexs", NULL, &weglIndexs, &ecb_glIndexs},
  {5133, "glIndexub", NULL, &weglIndexub, &ecb_glIndexub},
  {5134, "glColor3b", NULL, &weglColor3b, &ecb_glColor3b},
  {5135, "glColor3d", NULL, &weglColor3d, &ecb_glColor3d},
  {5136, "glColor3f", NULL, &weglColor3f, &ecb_glColor3f},
  {5137, "glColor3i", NULL, &weglColor3i, &ecb_glColor3i},
  {5138, "glColor3s", NULL, &weglColor3s, &ecb_glColor3s},
  {5139, "glColor3ub", NULL, &weglColor3ub, &ecb_glColor3ub},
  {5140, "glColor3ui", NULL, &weglColor3ui, &ecb_glColor3ui},
  {5141, "glColor3us", NULL, &weglColor3us, &ecb_glColor3us},
  {5142, "glColor4b", NULL, &weglColor4b, &ecb_glColor4b},
  {5143, "glColor4d", NULL, &weglColor4d, &ecb_glColor4d},
  {5144, "glColor4f", NULL, &weglColor4f, &ecb_glColor4f},
  {5145, "glColor4i", NULL, &weglColor4i, &ecb_glColor4i},
  {5146, "glColor4s", NULL, &weglColor4s, &ecb_glColor4s},
  {5147, "glColor4ub", NULL, &weglColor4ub, &ecb_glColor4ub},
  {5148, "glColor4ui", NULL, &weglColor4ui, &ecb_glColor4ui},
  {5149, "glColor4us", NULL, &weglColor4us, &ecb_glColor4us},
  {5150, "glTexCoord1d", NULL, &weglTexCoord1d, &ecb_glTexCoord1d},
  {5151, "glTexCoord1f", NULL, &weglTexCoord1f, &ecb_glTexCoord1f},
  {5152, "glTexCoord1i", NULL, &weglTexCoord1i, &ecb_glTexCoord1i},
  {5153, "glTexCoord1s", NULL, &weglTexCoord1s, &ecb_glTexCoord1s},
  {5154, "glTexCoord2d", NULL, &weglTexCoord2d, &ecb_glTexCoord2d},
  {5155, "glTexCoord2f", NULL, &weglTexCoord2f, &ecb_glTexCoord2f},
  {5156, "glTexCoord2i", NULL, &weglTexCoord2i, &ecb_glTexCoord2i},
  {5157, "glTexCoord2s", NULL, &weglTexCoord2s, &ecb_glTexCoord2s},
  {5158, "glTexCoord3d", NULL, &weglTexCoord3d, &ecb_glTexCoord3d},
  {5159, "glTexCoord3f", NULL, &weglTexCoord3f, &ecb_glTexCoord3f},
  {5160, "glTexCoord3i", NULL, &weglTexCoord3i, &ecb_glTexCoord3i},
  {5161, "glTexCoord3s", NULL, &weglTexCoord3s, &ecb_glTexCoord3s},
  {5162, "glTexCoord4d", NULL, &weglTexCoord4d, &ecb_glTexCoord4d},
  {5163, "glTexCoord4f", NULL, &weglTexCoord4f, &ecb_glTexCoord4f},
  {5164, "glTexCoord4i", NULL, &weglTexCoord4i, &ecb_glTexCoord4i},
  {5165, "glTexCoord4s", NULL, &weglTexCoord4s, &ecb_glTexCoord4s},
  {5166, "glRasterPos2d", NULL, &weglRasterPos2d, &ecb_glRasterPos2d},
  {5167, "glRasterPos2f", NULL, &weglRasterPos2f, &ecb_glRasterPos2f},
  {5168, "glRasterPos2i", NULL, &weglRasterPos2i, &ecb_glRasterPos2i},
  {5169, "glRasterPos2s", NULL, &weglRasterPos2s, &ecb_glRasterPos2s},
  {5170, "glRasterPos3d", NULL, &weglRasterPos3d, &ecb_glRasterPos3d},
  {5171, "glRasterPos3f", NULL, &weglRasterPos3f, &ecb_glRasterPos3f},
  {5172, "glRasterPos3i", NULL, &weglRasterPos3i, &ecb_glRasterPos3i},
  {5173, "glRasterPos3s", NULL, &weglRasterPos3s, &ecb_glRasterPos3s},
  {5174, "glRasterPos4d", NULL, &weglRasterPos4d, &ecb_glRasterPos4d},
  {5175, "glRasterPos4f", NULL, &weglRasterPos4f, &ecb_glRasterPos4f},
  {5176, "glRasterPos4i", NULL, &weglRasterPos4i, &ecb_glRasterPos4i},
  {5177, "glRasterPos4s", NULL, &weglRasterPos4s, &ecb_glRasterPos4s},
  {5178, "glRectd", NULL, &weglRectd, &ecb_glRectd},
  {5179, "glRectf", NULL, &weglRectf, &ecb_glRectf},
  {5180, "glRecti", NULL, &weglRecti, &ecb_glRecti},
  {5181, "glRects", NULL, &weglRects, &ecb_glRects},
  {5182, "glRectdv", NULL, &weglRectdv, &ecb_glRectdv},
  {5183, "glRectfv", NULL, &weglRectfv, &ecb_glRectfv},
  {5184, "glRectiv", NULL, &weglRectiv, &ecb_glRectiv},
  {5185, "glRectsv", NULL, &weglRectsv, &ecb_glRectsv},
  {5186, "glVertexPointer", NULL, &weglVertexPointer, &ecb_glVertexPointer},
  {5187, "unused", NULL, NULL, NULL},
  {5188, "glNormalPointer", NULL, &weglNormalPointer, &ecb_glNormalPointer},
  {5189, "unused", NULL, NULL, NULL},
  {5190, "glColorPointer", NULL, &weglColorPointer, &ecb_glColorPointer},
  {5191, "unused", NULL, NULL, NULL},
  {5192, "glIndexPointer", NULL, &weglIndexPointer, &ecb_glIndexPointer},
  {5193, "unused", NULL, NULL, NULL},
  {5194, "glTexCoordPointer", NULL, &weglTexCoordPointer, &ecb_glTexCoordPointer},
  {5195, "unused", NULL, NULL, NULL},
  {5196, "glEdgeFlagPointer", NULL, &weglEdgeFlagPointer, &ecb_glEdgeFlagPointer},
  {5197, "unused", NULL, NULL, NULL},
  {5198, "glArrayElement", NULL, &weglArrayElement, &ecb_glArrayElement},
  {5199, "glDrawArrays", NULL, &weglDrawArrays, &ecb_glDrawArrays},
  {5200, "glDrawElements", NULL, &weglDrawElements, &ecb_glDrawElements},
  {5201, "unused", NULL, NULL, NULL},
  {5202, "glInterleavedArrays", NULL, &weglInterleavedArrays, &ecb_glInterleavedArrays},
  {5203, "unused", NULL, NULL, NULL},
  {5204, "glShadeModel", NULL, &weglShadeModel, &ecb_glShadeModel},
  {5205, "glLightf", NULL, &weglLightf, &ecb_glLightf},
  {5206, "glLighti", NULL, &weglLighti, &ecb_glLighti},
  {5207, "glLightfv", NULL, &weglLightfv, &ecb_glLightfv},
  {5208, "glLightiv", NULL, &weglLightiv, &ecb_glLightiv},
  {5209, "glGetLightfv", NULL, &weglGetLightfv, &ecb_glGetLightfv},
  {5210, "glGetLightiv", NULL, &weglGetLightiv, &ecb_glGetLightiv},
  {5211, "glLightModelf", NULL, &weglLightModelf, &ecb_glLightModelf},
  {5212, "glLightModeli", NULL, &weglLightModeli, &ecb_glLightModeli},
  {5213, "glLightModelfv", NULL, &weglLightModelfv, &ecb_glLightModelfv},
  {5214, "glLightModeliv", NULL, &weglLightModeliv, &ecb_glLightModeliv},
  {5215, "glMaterialf", NULL, &weglMaterialf, &ecb_glMaterialf},
  {5216, "glMateriali", NULL, &weglMateriali, &ecb_glMateriali},
  {5217, "glMaterialfv", NULL, &weglMaterialfv, &ecb_glMaterialfv},
  {5218, "glMaterialiv", NULL, &weglMaterialiv, &ecb_glMaterialiv},
  {5219, "glGetMaterialfv", NULL, &weglGetMaterialfv, &ecb_glGetMaterialfv},
  {5220, "glGetMaterialiv", NULL, &weglGetMaterialiv, &ecb_glGetMaterialiv},
  {5221, "glColorMaterial", NULL, &weglColorMaterial, &ecb_glColorMaterial},
  {5222, "glPixelZoom", NULL, &weglPixelZoom, &ecb_glPixelZoom},
  {5223, "glPixelStoref", NULL, &weglPixelStoref, &ecb_glPixelStoref},
  {5224, "glPixelStorei", NULL, &weglPixelStorei, &ecb_glPixelStorei},
  {5225, "glPixelTransferf", NULL, &weglPixelTransferf, &ecb_glPixelTransferf},
  {5226, "glPixelTransferi", NULL, &weglPixelTransferi, &ecb_glPixelTransferi},
  {5227, "glPixelMapfv", NULL, &weglPixelMapfv, &ecb_glPixelMapfv},
  {5228, "glPixelMapuiv", NULL, &weglPixelMapuiv, &ecb_glPixelMapuiv},
  {5229, "glPixelMapusv", NULL, &weglPixelMapusv, &ecb_glPixelMapusv},
  {5230, "glGetPixelMapfv", NULL, &weglGetPixelMapfv, &ecb_glGetPixelMapfv},
  {5231, "glGetPixelMapuiv", NULL, &weglGetPixelMapuiv, &ecb_glGetPixelMapuiv},
  {5232, "glGetPixelMapusv", NULL, &weglGetPixelMapusv, &ecb_glGetPixelMapusv},
  {5233, "glBitmap", NULL, &weglBitmap, &ecb_glBitmap},
  {5234, "unused", NULL, NULL, NULL},
  {5235, "glReadPixels", NULL, &weglReadPixels, &ecb_glReadPixels},
  {5236, "glDrawPixels", NULL, &weglDrawPixels, &ecb_glDrawPixels},
  {5237, "unused", NULL, NULL, NULL},
  {5238, "glCopyPixels", NULL, &weglCopyPixels, &ecb_glCopyPixels},
  {5239, "glStencilFunc", NULL, &weglStencilFunc, &ecb_glStencilFunc},
  {5240, "glStencilMask", NULL, &weglStencilMask, &ecb_glStencilMask},
  {5241, "glStencilOp", NULL, &weglStencilOp, &ecb_glStencilOp},
  {5242, "glClearStencil", NULL, &weglClearStencil, &ecb_glClearStencil},
  {5243, "glTexGend", NULL, &weglTexGend, &ecb_glTexGend},
  {5244, "glTexGenf", NULL, &weglTexGenf, &ecb_glTexGenf},
  {5245, "glTexGeni", NULL, &weglTexGeni, &ecb_glTexGeni},
  {5246, "glTexGendv", NULL, &weglTexGendv, &ecb_glTexGendv},
  {5247, "glTexGenfv", NULL, &weglTexGenfv, &ecb_glTexGenfv},
  {5248, "glTexGeniv", NULL, &weglTexGeniv, &ecb_glTexGeniv},
  {5249, "glGetTexGendv", NULL, &weglGetTexGendv, &ecb_glGetTexGendv},
  {5250, "glGetTexGenfv", NULL, &weglGetTexGenfv, &ecb_glGetTexGenfv},
  {5251, "glGetTexGeniv", NULL, &weglGetTexGeniv, &ecb_glGetTexGeniv},
  {5252, "glTexEnvf", NULL, &weglTexEnvf, &ecb_glTexEnvf},
  {5253, "glTexEnvi", NULL, &weglTexEnvi, &ecb_glTexEnvi},
  {5254, "glTexEnvfv", NULL, &weglTexEnvfv, &ecb_glTexEnvfv},
  {5255, "glTexEnviv", NULL, &weglTexEnviv, &ecb_glTexEnviv},
  {5256, "glGetTexEnvfv", NULL, &weglGetTexEnvfv, &ecb_glGetTexEnvfv},
  {5257, "glGetTexEnviv", NULL, &weglGetTexEnviv, &ecb_glGetTexEnviv},
  {5258, "glTexParameterf", NULL, &weglTexParameterf, &ecb_glTexParameterf},
  {5259, "glTexParameteri", NULL, &weglTexParameteri, &ecb_glTexParameteri},
  {5260, "glTexParameterfv", NULL, &weglTexParameterfv, &ecb_glTexParameterfv},
  {5261, "glTexParameteriv", NULL, &weglTexParameteriv, &ecb_glTexParameteriv},
  {5262, "glGetTexParameterfv", NULL, &weglGetTexParameterfv, &ecb_glGetTexParameterfv},
  {5263, "glGetTexParameteriv", NULL, &weglGetTexParameteriv, &ecb_glGetTexParameteriv},
  {5264, "glGetTexLevelParameterfv", NULL, &weglGetTexLevelParameterfv, &ecb_glGetTexLevelParameterfv},
  {5265, "glGetTexLevelParameteriv", NULL, &weglGetTexLevelParameteriv, &ecb_glGetTexLevelParameteriv},
  {5266, "glTexImage1D", NULL, &weglTexImage1D, &ecb_glTexImage1D},
  {5267, "unused", NULL, NULL, NULL},
  {5268, "glTexImage2D", NULL, &weglTexImage2D, &ecb_glTexImage2D},
  {5269, "unused", NULL, NULL, NULL},
  {5270, "glGetTexImage", NULL, &weglGetTexImage, &ecb_glGetTexImage},
  {5271, "glGenTextures", NULL, &weglGenTextures, &ecb_glGenTextures},
  {5272, "glDeleteTextures", NULL, &weglDeleteTextures, &ecb_glDeleteTextures},
  {5273, "glBindTexture", NULL, &weglBindTexture, &ecb_glBindTexture},
  {5274, "glPrioritizeTextures", NULL, &weglPrioritizeTextures, &ecb_glPrioritizeTextures},
  {5275, "glAreTexturesResident", NULL, &weglAreTexturesResident, &ecb_glAreTexturesResident},
  {5276, "glIsTexture", NULL, &weglIsTexture, &ecb_glIsTexture},
  {5277, "glTexSubImage1D", NULL, &weglTexSubImage1D, &ecb_glTexSubImage1D},
  {5278, "unused", NULL, NULL, NULL},
  {5279, "glTexSubImage2D", NULL, &weglTexSubImage2D, &ecb_glTexSubImage2D},
  {5280, "unused", NULL, NULL, NULL},
  {5281, "glCopyTexImage1D", NULL, &weglCopyTexImage1D, &ecb_glCopyTexImage1D},
  {5282, "glCopyTexImage2D", NULL, &weglCopyTexImage2D, &ecb_glCopyTexImage2D},
  {5283, "glCopyTexSubImage1D", NULL, &weglCopyTexSubImage1D, &ecb_glCopyTexSubImage1D},
  {5284, "glCopyTexSubImage2D", NULL, &weglCopyTexSubImage2D, &ecb_glCopyTexSubImage2D},
  {5285, "glMap1d", NULL, &weglMap1d, &ecb_glMap1d},
  {5286, "glMap1f", NULL, &weglMap1f, &ecb_glMap1f},
  {5287, "glMap2d", NULL, &weglMap2d, &ecb_glMap2d},
  {5288, "glMap2f", NULL, &weglMap2f, &ecb_glMap2f},
  {5289, "glGetMapdv", NULL, &weglGetMapdv, &ecb_glGetMapdv},
  {5290, "glGetMapfv", NULL, &weglGetMapfv, &ecb_glGetMapfv},
  {5291, "glGetMapiv", NULL, &weglGetMapiv, &ecb_glGetMapiv},
  {5292, "glEvalCoord1d", NULL, &weglEvalCoord1d, &ecb_glEvalCoord1d},
  {5293, "glEvalCoord1f", NULL, &weglEvalCoord1f, &ecb_glEvalCoord1f},
  {5294, "glEvalCoord2d", NULL, &weglEvalCoord2d, &ecb_glEvalCoord2d},
  {5295, "glEvalCoord2f", NULL, &weglEvalCoord2f, &ecb_glEvalCoord2f},
  {5296, "glMapGrid1d", NULL, &weglMapGrid1d, &ecb_glMapGrid1d},
  {5297, "glMapGrid1f", NULL, &weglMapGrid1f, &ecb_glMapGrid1f},
  {5298, "glMapGrid2d", NULL, &weglMapGrid2d, &ecb_glMapGrid2d},
  {5299, "glMapGrid2f", NULL, &weglMapGrid2f, &ecb_glMapGrid2f},
  {5300, "glEvalPoint1", NULL, &weglEvalPoint1, &ecb_glEvalPoint1},
  {5301, "glEvalPoint2", NULL, &weglEvalPoint2, &ecb_glEvalPoint2},
  {5302, "glEvalMesh1", NULL, &weglEvalMesh1, &ecb_glEvalMesh1},
  {5303, "glEvalMesh2", NULL, &weglEvalMesh2, &ecb_glEvalMesh2},
  {5304, "glFogf", NULL, &weglFogf, &ecb_glFogf},
  {5305, "glFogi", NULL, &weglFogi, &ecb_glFogi},
  {5306, "glFogfv", NULL, &weglFogfv, &ecb_glFogfv},
  {5307, "glFogiv", NULL, &weglFogiv, &ecb_glFogiv},
  {5308, "glFeedbackBuffer", NULL, &weglFeedbackBuffer, &ecb_glFeedbackBuffer},
  {5309, "glPassThrough", NULL, &weglPassThrough, &ecb_glPassThrough},
  {5310, "glSelectBuffer", NULL, &weglSelectBuffer, &ecb_glSelectBuffer},
  {5311, "glInitNames", NULL, &weglInitNames, &ecb_glInitNames},
  {5312, "glLoadName", NULL, &weglLoadName, &ecb_glLoadName},
  {5313, "glPushName", NULL, &weglPushName, &ecb_glPushName},
  {5314, "glPopName", NULL, &weglPopName, &ecb_glPopName},
  {5315, "glDrawRangeElements", NULL, &weglDrawRangeElements, &ecb_glDrawRangeElements},
  {5316, "unused", NULL, NULL, NULL},
  {5317, "glTexImage3D", NULL, &weglTexImage3D, &ecb_glTexImage3D},
  {5318, "unused", NULL, NULL, NULL},
  {5319, "glTexSubImage3D", NULL, &weglTexSubImage3D, &ecb_glTexSubImage3D},
  {5320, "unused", NULL, NULL, NULL},
  {5321, "glCopyTexSubImage3D", NULL, &weglCopyTexSubImage3D, &ecb_glCopyTexSubImage3D},
  {5322, "glActiveTexture", "glActiveTextureARB", &weglActiveTexture, &ecb_glActiveTexture},
  {5323, "glSampleCoverage", NULL, &weglSampleCoverage, &ecb_glSampleCoverage},
  {5324, "glCompressedTexImage3D", "glCompressedTexImage3DARB", &weglCompressedTexImage3D, &ecb_glCompressedTexImage3D},
  {5325, "unused", NULL, NULL, NULL},
  {5326, "glCompressedTexImage2D", "glCompressedTexImage2DARB", &weglCompressedTexImage2D, &ecb_glCompressedTexImage2D},
  {5327, "unused", NULL, NULL, NULL},
  {5328, "glCompressedTexImage1D", "glCompressedTexImage1DARB", &weglCompressedTexImage1D, &ecb_glCompressedTexImage1D},
  {5329, "unused", NULL, NULL, NULL},
  {5330, "glCompressedTexSubImage3D", "glCompressedTexSubImage3DARB", &weglCompressedTexSubImage3D, &ecb_glCompressedTexSubImage3D},
  {5331, "unused", NULL, NULL, NULL},
  {5332, "glCompressedTexSubImage2D", "glCompressedTexSubImage2DARB", &weglCompressedTexSubImage2D, &ecb_glCompressedTexSubImage2D},
  {5333, "unused", NULL, NULL, NULL},
  {5334, "glCompressedTexSubImage1D", "glCompressedTexSubImage1DARB", &weglCompressedTexSubImage1D, &ecb_glCompressedTexSubImage1D},
  {5335, "unused", NULL, NULL, NULL},
  {5336, "glGetCompressedTexImage", NULL, &weglGetCompressedTexImage, &ecb_glGetCompressedTexImage},
  {5337, "glClientActiveTexture", "glClientActiveTextureARB", &weglClientActiveTexture, &ecb_glClientActiveTexture},
  {5338, "glMultiTexCoord1d", "glMultiTexCoord1dARB", &weglMultiTexCoord1d, &ecb_glMultiTexCoord1d},
  {5339, "glMultiTexCoord1f", "glMultiTexCoord1fARB", &weglMultiTexCoord1f, &ecb_glMultiTexCoord1f},
  {5340, "glMultiTexCoord1i", "glMultiTexCoord1iARB", &weglMultiTexCoord1i, &ecb_glMultiTexCoord1i},
  {5341, "glMultiTexCoord1s", "glMultiTexCoord1sARB", &weglMultiTexCoord1s, &ecb_glMultiTexCoord1s},
  {5342, "glMultiTexCoord2d", "glMultiTexCoord2dARB", &weglMultiTexCoord2d, &ecb_glMultiTexCoord2d},
  {5343, "glMultiTexCoord2f", "glMultiTexCoord2fARB", &weglMultiTexCoord2f, &ecb_glMultiTexCoord2f},
  {5344, "glMultiTexCoord2i", "glMultiTexCoord2iARB", &weglMultiTexCoord2i, &ecb_glMultiTexCoord2i},
  {5345, "glMultiTexCoord2s", "glMultiTexCoord2sARB", &weglMultiTexCoord2s, &ecb_glMultiTexCoord2s},
  {5346, "glMultiTexCoord3d", "glMultiTexCoord3dARB", &weglMultiTexCoord3d, &ecb_glMultiTexCoord3d},
  {5347, "glMultiTexCoord3f", "glMultiTexCoord3fARB", &weglMultiTexCoord3f, &ecb_glMultiTexCoord3f},
  {5348, "glMultiTexCoord3i", "glMultiTexCoord3iARB", &weglMultiTexCoord3i, &ecb_glMultiTexCoord3i},
  {5349, "glMultiTexCoord3s", "glMultiTexCoord3sARB", &weglMultiTexCoord3s, &ecb_glMultiTexCoord3s},
  {5350, "glMultiTexCoord4d", "glMultiTexCoord4dARB", &weglMultiTexCoord4d, &ecb_glMultiTexCoord4d},
  {5351, "glMultiTexCoord4f", "glMultiTexCoord4fARB", &weglMultiTexCoord4f, &ecb_glMultiTexCoord4f},
  {5352, "glMultiTexCoord4i", "glMultiTexCoord4iARB", &weglMultiTexCoord4i, &ecb_glMultiTexCoord4i},
  {5353, "glMultiTexCoord4s", "glMultiTexCoord4sARB", &weglMultiTexCoord4s, &ecb_glMultiTexCoord4s},
  {5354, "glLoadTransposeMatrixf", NULL, &weglLoadTransposeMatrixf, &ecb_glLoadTransposeMatrixf},
  {5355, "glLoadTransposeMatrixd", NULL, &weglLoadTransposeMatrixd, &ecb_glLoadTransposeMatrixd},
  {5356, "glMultTransposeMatrixf", NULL, &weglMultTransposeMatrixf, &ecb_glMultTransposeMatrixf},
  {5357, "glMultTransposeMatrixd", NULL, &weglMultTransposeMatrixd, &ecb_glMultTransposeMatrixd},
  {5358, "glBlendFuncSeparate", NULL, &weglBlendFuncSeparate, &ecb_glBlendFuncSeparate},
  {5359, "glMultiDrawArrays", NULL, &weglMultiDrawArrays, &ecb_glMultiDrawArrays},
  {5360, "unused", NULL, NULL, NULL},
  {5361, "glPointParameterf", "glPointParameterfARB", &weglPointParameterf, &ecb_glPointParameterf},
  {5362, "glPointParameterfv", "glPointParameterfvARB", &weglPointParameterfv, &ecb_glPointParameterfv},
  {5363, "glPointParameteri", NULL, &weglPointParameteri, &ecb_glPointParameteri},
  {5364, "glPointParameteriv", NULL, &weglPointParameteriv, &ecb_glPointParameteriv},
  {5365, "glFogCoordf", NULL, &weglFogCoordf, &ecb_glFogCoordf},
  {5366, "glFogCoordd", NULL, &weglFogCoordd, &ecb_glFogCoordd},
  {5367, "glFogCoordPointer", NULL, &weglFogCoordPointer, &ecb_glFogCoordPointer},
  {5368, "unused", NULL, NULL, NULL},
  {5369, "glSecondaryColor3b", NULL, &weglSecondaryColor3b, &ecb_glSecondaryColor3b},
  {5370, "glSecondaryColor3d", NULL, &weglSecondaryColor3d, &ecb_glSecondaryColor3d},
  {5371, "glSecondaryColor3f", NULL, &weglSecondaryColor3f, &ecb_glSecondaryColor3f},
  {5372, "glSecondaryColor3i", NULL, &weglSecondaryColor3i, &ecb_glSecondaryColor3i},
  {5373, "glSecondaryColor3s", NULL, &weglSecondaryColor3s, &ecb_glSecondaryColor3s},
  {5374, "glSecondaryColor3ub", NULL, &weglSecondaryColor3ub, &ecb_glSecondaryColor3ub},
  {5375, "glSecondaryColor3ui", NULL, &weglSecondaryColor3ui, &ecb_glSecondaryColor3ui},
  {5376, "glSecondaryColor3us", NULL, &weglSecondaryColor3us, &ecb_glSecondaryColor3us},
  {5377, "glSecondaryColorPointer", NULL, &weglSecondaryColorPointer, &ecb_glSecondaryColorPointer},
  {5378, "unused", NULL, NULL, NULL},
  {5379, "glWindowPos2d", "glWindowPos2dARB", &weglWindowPos2d, &ecb_glWindowPos2d},
  {5380, "glWindowPos2f", "glWindowPos2fARB", &weglWindowPos2f, &ecb_glWindowPos2f},
  {5381, "glWindowPos2i", "glWindowPos2iARB", &weglWindowPos2i, &ecb_glWindowPos2i},
  {5382, "glWindowPos2s", "glWindowPos2sARB", &weglWindowPos2s, &ecb_glWindowPos2s},
  {5383, "glWindowPos3d", "glWindowPos3dARB", &weglWindowPos3d, &ecb_glWindowPos3d},
  {5384, "glWindowPos3f", "glWindowPos3fARB", &weglWindowPos3f, &ecb_glWindowPos3f},
  {5385, "glWindowPos3i", "glWindowPos3iARB", &weglWindowPos3i, &ecb_glWindowPos3i},
  {5386, "glWindowPos3s", "glWindowPos3sARB", &weglWindowPos3s, &ecb_glWindowPos3s},
  {5387, "glBlendColor", NULL, &weglBlendColor, &ecb_glBlendColor},
  {5388, "glBlendEquation", NULL, &weglBlendEquation, &ecb_glBlendEquation},
  {5389, "glGenQueries", "glGenQueriesARB", &weglGenQueries, &ecb_glGenQueries},
  {5390, "glDeleteQueries", "glDeleteQueriesARB", &weglDeleteQueries, &ecb_glDeleteQueries},
  {5391, "glIsQuery", "glIsQueryARB", &weglIsQuery, &ecb_glIsQuery},
  {5392, "glBeginQuery", "glBeginQueryARB", &weglBeginQuery, &ecb_glBeginQuery},
  {5393, "glEndQuery", "glEndQueryARB", &weglEndQuery, &ecb_glEndQuery},
  {5394, "glGetQueryiv", "glGetQueryivARB", &weglGetQueryiv, &ecb_glGetQueryiv},
  {5395, "glGetQueryObjectiv", "glGetQueryObjectivARB", &weglGetQueryObjectiv, &ecb_glGetQueryObjectiv},
  {5396, "glGetQueryObjectuiv", "glGetQueryObjectuivARB", &weglGetQueryObjectuiv, &ecb_glGetQueryObjectuiv},
  {5397, "glBindBuffer", "glBindBufferARB", &weglBindBuffer, &ecb_glBindBuffer},
  {5398, "glDeleteBuffers", "glDeleteBuffersARB", &weglDeleteBuffers, &ecb_glDeleteBuffers},
  {5399, "glGenBuffers", "glGenBuffersARB", &weglGenBuffers, &ecb_glGenBuffers},
  {5400, "glIsBuffer", "glIsBufferARB", &weglIsBuffer, &ecb_glIsBuffer},
  {5401, "glBufferData", "glBufferDataARB", &weglBufferData, &ecb_glBufferData},
  {5402, "unused", NULL, NULL, NULL},
  {5403, "glBufferSubData", "glBufferSubDataARB", &weglBufferSubData, &ecb_glBufferSubData},
  {5404, "unused", NULL, NULL, NULL},
  {5405, "glGetBufferSubData", "glGetBufferSubDataARB", &weglGetBufferSubData, &ecb_glGetBufferSubData},
  {5406, "glGetBufferParameteriv", NULL, &weglGetBufferParameteriv, &ecb_glGetBufferParameteriv},
  {5407, "glBlendEquationSeparate", "glBlendEquationSeparateEXT", &weglBlendEquationSeparate, &ecb_glBlendEquationSeparate},
  {5408, "glDrawBuffers", "glDrawBuffersARB", &weglDrawBuffers, &ecb_glDrawBuffers},
  {5409, "glStencilOpSeparate", "glStencilOpSeparateATI", &weglStencilOpSeparate, &ecb_glStencilOpSeparate},
  {5410, "glStencilFuncSeparate", "glStencilFuncSeparateATI", &weglStencilFuncSeparate, &ecb_glStencilFuncSeparate},
  {5411, "glStencilMaskSeparate", NULL, &weglStencilMaskSeparate, &ecb_glStencilMaskSeparate},
  {5412, "glAttachShader", NULL, &weglAttachShader, &ecb_glAttachShader},
  {5413, "glBindAttribLocation", NULL, &weglBindAttribLocation, &ecb_glBindAttribLocation},
  {5414, "glCompileShader", NULL, &weglCompileShader, &ecb_glCompileShader},
  {5415, "glCreateProgram", NULL, &weglCreateProgram, &ecb_glCreateProgram},
  {5416, "glCreateShader", NULL, &weglCreateShader, &ecb_glCreateShader},
  {5417, "glDeleteProgram", NULL, &weglDeleteProgram, &ecb_glDeleteProgram},
  {5418, "glDeleteShader", NULL, &weglDeleteShader, &ecb_glDeleteShader},
  {5419, "glDetachShader", NULL, &weglDetachShader, &ecb_glDetachShader},
  {5420, "glDisableVertexAttribArray", "glDisableVertexAttribArrayARB", &weglDisableVertexAttribArray, &ecb_glDisableVertexAttribArray},
  {5421, "glEnableVertexAttribArray", "glEnableVertexAttribArrayARB", &weglEnableVertexAttribArray, &ecb_glEnableVertexAttribArray},
  {5422, "glGetActiveAttrib", NULL, &weglGetActiveAttrib, &ecb_glGetActiveAttrib},
  {5423, "glGetActiveUniform", NULL, &weglGetActiveUniform, &ecb_glGetActiveUniform},
  {5424, "glGetAttachedShaders", NULL, &weglGetAttachedShaders, &ecb_glGetAttachedShaders},
  {5425, "glGetAttribLocation", NULL, &weglGetAttribLocation, &ecb_glGetAttribLocation},
  {5426, "glGetProgramiv", "glGetProgramivARB", &weglGetProgramiv, &ecb_glGetProgramiv},
  {5427, "glGetProgramInfoLog", NULL, &weglGetProgramInfoLog, &ecb_glGetProgramInfoLog},
  {5428, "glGetShaderiv", NULL, &weglGetShaderiv, &ecb_glGetShaderiv},
  {5429, "glGetShaderInfoLog", NULL, &weglGetShaderInfoLog, &ecb_glGetShaderInfoLog},
  {5430, "glGetShaderSource", NULL, &weglGetShaderSource, &ecb_glGetShaderSource},
  {5431, "glGetUniformLocation", NULL, &weglGetUniformLocation, &ecb_glGetUniformLocation},
  {5432, "glGetUniformfv", NULL, &weglGetUniformfv, &ecb_glGetUniformfv},
  {5433, "glGetUniformiv", NULL, &weglGetUniformiv, &ecb_glGetUniformiv},
  {5434, "glGetVertexAttribdv", "glGetVertexAttribdvARB", &weglGetVertexAttribdv, &ecb_glGetVertexAttribdv},
  {5435, "glGetVertexAttribfv", "glGetVertexAttribfvARB", &weglGetVertexAttribfv, &ecb_glGetVertexAttribfv},
  {5436, "glGetVertexAttribiv", "glGetVertexAttribivARB", &weglGetVertexAttribiv, &ecb_glGetVertexAttribiv},
  {5437, "glIsProgram", "glIsProgramARB", &weglIsProgram, &ecb_glIsProgram},
  {5438, "glIsShader", NULL, &weglIsShader, &ecb_glIsShader},
  {5439, "glLinkProgram", NULL, &weglLinkProgram, &ecb_glLinkProgram},
  {5440, "glShaderSource", NULL, &weglShaderSource, &ecb_glShaderSource},
  {5441, "glUseProgram", NULL, &weglUseProgram, &ecb_glUseProgram},
  {5442, "glUniform1f", "glUniform1fARB", &weglUniform1f, &ecb_glUniform1f},
  {5443, "glUniform2f", "glUniform2fARB", &weglUniform2f, &ecb_glUniform2f},
  {5444, "glUniform3f", "glUniform3fARB", &weglUniform3f, &ecb_glUniform3f},
  {5445, "glUniform4f", "glUniform4fARB", &weglUniform4f, &ecb_glUniform4f},
  {5446, "glUniform1i", "glUniform1iARB", &weglUniform1i, &ecb_glUniform1i},
  {5447, "glUniform2i", "glUniform2iARB", &weglUniform2i, &ecb_glUniform2i},
  {5448, "glUniform3i", "glUniform3iARB", &weglUniform3i, &ecb_glUniform3i},
  {5449, "glUniform4i", "glUniform4iARB", &weglUniform4i, &ecb_glUniform4i},
  {5450, "glUniform1fv", "glUniform1fvARB", &weglUniform1fv, &ecb_glUniform1fv},
  {5451, "glUniform2fv", "glUniform2fvARB", &weglUniform2fv, &ecb_glUniform2fv},
  {5452, "glUniform3fv", "glUniform3fvARB", &weglUniform3fv, &ecb_glUniform3fv},
  {5453, "glUniform4fv", "glUniform4fvARB", &weglUniform4fv, &ecb_glUniform4fv},
  {5454, "glUniform1iv", "glUniform1ivARB", &weglUniform1iv, &ecb_glUniform1iv},
  {5455, "glUniform2iv", "glUniform2ivARB", &weglUniform2iv, &ecb_glUniform2iv},
  {5456, "glUniform3iv", "glUniform3ivARB", &weglUniform3iv, &ecb_glUniform3iv},
  {5457, "glUniform4iv", "glUniform4ivARB", &weglUniform4iv, &ecb_glUniform4iv},
  {5458, "glUniformMatrix2fv", "glUniformMatrix2fvARB", &weglUniformMatrix2fv, &ecb_glUniformMatrix2fv},
  {5459, "glUniformMatrix3fv", "glUniformMatrix3fvARB", &weglUniformMatrix3fv, &ecb_glUniformMatrix3fv},
  {5460, "glUniformMatrix4fv", "glUniformMatrix4fvARB", &weglUniformMatrix4fv, &ecb_glUniformMatrix4fv},
  {5461, "glValidateProgram", NULL, &weglValidateProgram, &ecb_glValidateProgram},
  {5462, "glVertexAttrib1d", "glVertexAttrib1dARB", &weglVertexAttrib1d, &ecb_glVertexAttrib1d},
  {5463, "glVertexAttrib1f", "glVertexAttrib1fARB", &weglVertexAttrib1f, &ecb_glVertexAttrib1f},
  {5464, "glVertexAttrib1s", "glVertexAttrib1sARB", &weglVertexAttrib1s, &ecb_glVertexAttrib1s},
  {5465, "glVertexAttrib2d", "glVertexAttrib2dARB", &weglVertexAttrib2d, &ecb_glVertexAttrib2d},
  {5466, "glVertexAttrib2f", "glVertexAttrib2fARB", &weglVertexAttrib2f, &ecb_glVertexAttrib2f},
  {5467, "glVertexAttrib2s", "glVertexAttrib2sARB", &weglVertexAttrib2s, &ecb_glVertexAttrib2s},
  {5468, "glVertexAttrib3d", "glVertexAttrib3dARB", &weglVertexAttrib3d, &ecb_glVertexAttrib3d},
  {5469, "glVertexAttrib3f", "glVertexAttrib3fARB", &weglVertexAttrib3f, &ecb_glVertexAttrib3f},
  {5470, "glVertexAttrib3s", "glVertexAttrib3sARB", &weglVertexAttrib3s, &ecb_glVertexAttrib3s},
  {5471, "glVertexAttrib4Nbv", "glVertexAttrib4NbvARB", &weglVertexAttrib4Nbv, &ecb_glVertexAttrib4Nbv},
  {5472, "glVertexAttrib4Niv", "glVertexAttrib4NivARB", &weglVertexAttrib4Niv, &ecb_glVertexAttrib4Niv},
  {5473, "glVertexAttrib4Nsv", "glVertexAttrib4NsvARB", &weglVertexAttrib4Nsv, &ecb_glVertexAttrib4Nsv},
  {5474, "glVertexAttrib4Nub", "glVertexAttrib4NubARB", &weglVertexAttrib4Nub, &ecb_glVertexAttrib4Nub},
  {5475, "glVertexAttrib4Nuiv", "glVertexAttrib4NuivARB", &weglVertexAttrib4Nuiv, &ecb_glVertexAttrib4Nuiv},
  {5476, "glVertexAttrib4Nusv", "glVertexAttrib4NusvARB", &weglVertexAttrib4Nusv, &ecb_glVertexAttrib4Nusv},
  {5477, "glVertexAttrib4bv", "glVertexAttrib4bvARB", &weglVertexAttrib4bv, &ecb_glVertexAttrib4bv},
  {5478, "glVertexAttrib4d", "glVertexAttrib4dARB", &weglVertexAttrib4d, &ecb_glVertexAttrib4d},
  {5479, "glVertexAttrib4f", "glVertexAttrib4fARB", &weglVertexAttrib4f, &ecb_glVertexAttrib4f},
  {5480, "glVertexAttrib4iv", "glVertexAttrib4ivARB", &weglVertexAttrib4iv, &ecb_glVertexAttrib4iv},
  {5481, "glVertexAttrib4s", "glVertexAttrib4sARB", &weglVertexAttrib4s, &ecb_glVertexAttrib4s},
  {5482, "glVertexAttrib4ubv", "glVertexAttrib4ubvARB", &weglVertexAttrib4ubv, &ecb_glVertexAttrib4ubv},
  {5483, "glVertexAttrib4uiv", "glVertexAttrib4uivARB", &weglVertexAttrib4uiv, &ecb_glVertexAttrib4uiv},
  {5484, "glVertexAttrib4usv", "glVertexAttrib4usvARB", &weglVertexAttrib4usv, &ecb_glVertexAttrib4usv},
  {5485, "glVertexAttribPointer", "glVertexAttribPointerARB", &weglVertexAttribPointer, &ecb_glVertexAttribPointer},
  {5486, "unused", NULL, NULL, NULL},
  {5487, "glUniformMatrix2x3fv", NULL, &weglUniformMatrix2x3fv, &ecb_glUniformMatrix2x3fv},
  {5488, "glUniformMatrix3x2fv", NULL, &weglUniformMatrix3x2fv, &ecb_glUniformMatrix3x2fv},
  {5489, "glUniformMatrix2x4fv", NULL, &weglUniformMatrix2x4fv, &ecb_glUniformMatrix2x4fv},
  {5490, "glUniformMatrix4x2fv", NULL, &weglUniformMatrix4x2fv, &ecb_glUniformMatrix4x2fv},
  {5491, "glUniformMatrix3x4fv", NULL, &weglUniformMatrix3x4fv, &ecb_glUniformMatrix3x4fv},
  {5492, "glUniformMatrix4x3fv", NULL, &weglUniformMatrix4x3fv, &ecb_glUniformMatrix4x3fv},
  {5493, "glColorMaski", NULL, &weglColorMaski, &ecb_glColorMaski},
  {5494, "glGetBooleani_v", NULL, &weglGetBooleani_v, &ecb_glGetBooleani_v},
  {5495, "glGetIntegeri_v", NULL, &weglGetIntegeri_v, &ecb_glGetIntegeri_v},
  {5496, "glEnablei", NULL, &weglEnablei, &ecb_glEnablei},
  {5497, "glDisablei", NULL, &weglDisablei, &ecb_glDisablei},
  {5498, "glIsEnabledi", NULL, &weglIsEnabledi, &ecb_glIsEnabledi},
  {5499, "glBeginTransformFeedback", NULL, &weglBeginTransformFeedback, &ecb_glBeginTransformFeedback},
  {5500, "glEndTransformFeedback", NULL, &weglEndTransformFeedback, &ecb_glEndTransformFeedback},
  {5501, "glBindBufferRange", NULL, &weglBindBufferRange, &ecb_glBindBufferRange},
  {5502, "glBindBufferBase", NULL, &weglBindBufferBase, &ecb_glBindBufferBase},
  {5503, "glTransformFeedbackVaryings", NULL, &weglTransformFeedbackVaryings, &ecb_glTransformFeedbackVaryings},
  {5504, "glGetTransformFeedbackVarying", NULL, &weglGetTransformFeedbackVarying, &ecb_glGetTransformFeedbackVarying},
  {5505, "glClampColor", "glClampColorARB", &weglClampColor, &ecb_glClampColor},
  {5506, "glBeginConditionalRender", NULL, &weglBeginConditionalRender, &ecb_glBeginConditionalRender},
  {5507, "glEndConditionalRender", NULL, &weglEndConditionalRender, &ecb_glEndConditionalRender},
  {5508, "glVertexAttribIPointer", NULL, &weglVertexAttribIPointer, &ecb_glVertexAttribIPointer},
  {5509, "unused", NULL, NULL, NULL},
  {5510, "glGetVertexAttribIiv", NULL, &weglGetVertexAttribIiv, &ecb_glGetVertexAttribIiv},
  {5511, "glGetVertexAttribIuiv", NULL, &weglGetVertexAttribIuiv, &ecb_glGetVertexAttribIuiv},
  {5512, "glVertexAttribI1i", NULL, &weglVertexAttribI1i, &ecb_glVertexAttribI1i},
  {5513, "glVertexAttribI2i", NULL, &weglVertexAttribI2i, &ecb_glVertexAttribI2i},
  {5514, "glVertexAttribI3i", NULL, &weglVertexAttribI3i, &ecb_glVertexAttribI3i},
  {5515, "glVertexAttribI4i", NULL, &weglVertexAttribI4i, &ecb_glVertexAttribI4i},
  {5516, "glVertexAttribI1ui", NULL, &weglVertexAttribI1ui, &ecb_glVertexAttribI1ui},
  {5517, "glVertexAttribI2ui", NULL, &weglVertexAttribI2ui, &ecb_glVertexAttribI2ui},
  {5518, "glVertexAttribI3ui", NULL, &weglVertexAttribI3ui, &ecb_glVertexAttribI3ui},
  {5519, "glVertexAttribI4ui", NULL, &weglVertexAttribI4ui, &ecb_glVertexAttribI4ui},
  {5520, "glVertexAttribI4bv", NULL, &weglVertexAttribI4bv, &ecb_glVertexAttribI4bv},
  {5521, "glVertexAttribI4sv", NULL, &weglVertexAttribI4sv, &ecb_glVertexAttribI4sv},
  {5522, "glVertexAttribI4ubv", NULL, &weglVertexAttribI4ubv, &ecb_glVertexAttribI4ubv},
  {5523, "glVertexAttribI4usv", NULL, &weglVertexAttribI4usv, &ecb_glVertexAttribI4usv},
  {5524, "glGetUniformuiv", NULL, &weglGetUniformuiv, &ecb_glGetUniformuiv},
  {5525, "glBindFragDataLocation", NULL, &weglBindFragDataLocation, &ecb_glBindFragDataLocation},
  {5526, "glGetFragDataLocation", NULL, &weglGetFragDataLocation, &ecb_glGetFragDataLocation},
  {5527, "glUniform1ui", NULL, &weglUniform1ui, &ecb_glUniform1ui},
  {5528, "glUniform2ui", NULL, &weglUniform2ui, &ecb_glUniform2ui},
  {5529, "glUniform3ui", NULL, &weglUniform3ui, &ecb_glUniform3ui},
  {5530, "glUniform4ui", NULL, &weglUniform4ui, &ecb_glUniform4ui},
  {5531, "glUniform1uiv", NULL, &weglUniform1uiv, &ecb_glUniform1uiv},
  {5532, "glUniform2uiv", NULL, &weglUniform2uiv, &ecb_glUniform2uiv},
  {5533, "glUniform3uiv", NULL, &weglUniform3uiv, &ecb_glUniform3uiv},
  {5534, "glUniform4uiv", NULL, &weglUniform4uiv, &ecb_glUniform4uiv},
  {5535, "glTexParameterIiv", NULL, &weglTexParameterIiv, &ecb_glTexParameterIiv},
  {5536, "glTexParameterIuiv", NULL, &weglTexParameterIuiv, &ecb_glTexParameterIuiv},
  {5537, "glGetTexParameterIiv", NULL, &weglGetTexParameterIiv, &ecb_glGetTexParameterIiv},
  {5538, "glGetTexParameterIuiv", NULL, &weglGetTexParameterIuiv, &ecb_glGetTexParameterIuiv},
  {5539, "glClearBufferiv", NULL, &weglClearBufferiv, &ecb_glClearBufferiv},
  {5540, "glClearBufferuiv", NULL, &weglClearBufferuiv, &ecb_glClearBufferuiv},
  {5541, "glClearBufferfv", NULL, &weglClearBufferfv, &ecb_glClearBufferfv},
  {5542, "glClearBufferfi", NULL, &weglClearBufferfi, &ecb_glClearBufferfi},
  {5543, "glGetStringi", NULL, (void *) &weglGetStringi, &ecb_glGetStringi},
  {5544, "glIsRenderbuffer", "glIsRenderbufferEXT", &weglIsRenderbuffer, &ecb_glIsRenderbuffer},
  {5545, "glBindRenderbuffer", "glBindRenderbufferEXT", &weglBindRenderbuffer, &ecb_glBindRenderbuffer},
  {5546, "glDeleteRenderbuffers", "glDeleteRenderbuffersEXT", &weglDeleteRenderbuffers, &ecb_glDeleteRenderbuffers},
  {5547, "glGenRenderbuffers", "glGenRenderbuffersEXT", &weglGenRenderbuffers, &ecb_glGenRenderbuffers},
  {5548, "glRenderbufferStorage", "glRenderbufferStorageEXT", &weglRenderbufferStorage, &ecb_glRenderbufferStorage},
  {5549, "glGetRenderbufferParameteriv", "glGetRenderbufferParameterivEXT", &weglGetRenderbufferParameteriv, &ecb_glGetRenderbufferParameteriv},
  {5550, "glIsFramebuffer", "glIsFramebufferEXT", &weglIsFramebuffer, &ecb_glIsFramebuffer},
  {5551, "glBindFramebuffer", "glBindFramebufferEXT", &weglBindFramebuffer, &ecb_glBindFramebuffer},
  {5552, "glDeleteFramebuffers", "glDeleteFramebuffersEXT", &weglDeleteFramebuffers, &ecb_glDeleteFramebuffers},
  {5553, "glGenFramebuffers", "glGenFramebuffersEXT", &weglGenFramebuffers, &ecb_glGenFramebuffers},
  {5554, "glCheckFramebufferStatus", "glCheckFramebufferStatusEXT", &weglCheckFramebufferStatus, &ecb_glCheckFramebufferStatus},
  {5555, "glFramebufferTexture1D", "glFramebufferTexture1DEXT", &weglFramebufferTexture1D, &ecb_glFramebufferTexture1D},
  {5556, "glFramebufferTexture2D", "glFramebufferTexture2DEXT", &weglFramebufferTexture2D, &ecb_glFramebufferTexture2D},
  {5557, "glFramebufferTexture3D", "glFramebufferTexture3DEXT", &weglFramebufferTexture3D, &ecb_glFramebufferTexture3D},
  {5558, "glFramebufferRenderbuffer", "glFramebufferRenderbufferEXT", &weglFramebufferRenderbuffer, &ecb_glFramebufferRenderbuffer},
  {5559, "glGetFramebufferAttachmentParameteriv", "glGetFramebufferAttachmentParameterivEXT", &weglGetFramebufferAttachmentParameteriv, &ecb_glGetFramebufferAttachmentParameteriv},
  {5560, "glGenerateMipmap", "glGenerateMipmapEXT", &weglGenerateMipmap, &ecb_glGenerateMipmap},
  {5561, "glBlitFramebuffer", "glBlitFramebufferEXT", &weglBlitFramebuffer, &ecb_glBlitFramebuffer},
  {5562, "glRenderbufferStorageMultisample", "glRenderbufferStorageMultisampleEXT", &weglRenderbufferStorageMultisample, &ecb_glRenderbufferStorageMultisample},
  {5563, "glFramebufferTextureLayer", "glFramebufferTextureLayerARB", &weglFramebufferTextureLayer, &ecb_glFramebufferTextureLayer},
  {5564, "glFlushMappedBufferRange", NULL, &weglFlushMappedBufferRange, &ecb_glFlushMappedBufferRange},
  {5565, "glBindVertexArray", NULL, &weglBindVertexArray, &ecb_glBindVertexArray},
  {5566, "glDeleteVertexArrays", NULL, &weglDeleteVertexArrays, &ecb_glDeleteVertexArrays},
  {5567, "glGenVertexArrays", NULL, &weglGenVertexArrays, &ecb_glGenVertexArrays},
  {5568, "glIsVertexArray", NULL, &weglIsVertexArray, &ecb_glIsVertexArray},
  {5569, "glDrawArraysInstanced", "glDrawArraysInstancedARB", &weglDrawArraysInstanced, &ecb_glDrawArraysInstanced},
  {5570, "glDrawElementsInstanced", "glDrawElementsInstancedARB", &weglDrawElementsInstanced, &ecb_glDrawElementsInstanced},
  {5571, "unused", NULL, NULL, NULL},
  {5572, "glTexBuffer", "glTexBufferARB", &weglTexBuffer, &ecb_glTexBuffer},
  {5573, "glPrimitiveRestartIndex", NULL, &weglPrimitiveRestartIndex, &ecb_glPrimitiveRestartIndex},
  {5574, "glCopyBufferSubData", NULL, &weglCopyBufferSubData, &ecb_glCopyBufferSubData},
  {5575, "glGetUniformIndices", NULL, &weglGetUniformIndices, &ecb_glGetUniformIndices},
  {5576, "glGetActiveUniformsiv", NULL, &weglGetActiveUniformsiv, &ecb_glGetActiveUniformsiv},
  {5577, "glGetActiveUniformName", NULL, &weglGetActiveUniformName, &ecb_glGetActiveUniformName},
  {5578, "glGetUniformBlockIndex", NULL, &weglGetUniformBlockIndex, &ecb_glGetUniformBlockIndex},
  {5579, "glGetActiveUniformBlockiv", NULL, &weglGetActiveUniformBlockiv, &ecb_glGetActiveUniformBlockiv},
  {5580, "glGetActiveUniformBlockName", NULL, &weglGetActiveUniformBlockName, &ecb_glGetActiveUniformBlockName},
  {5581, "glUniformBlockBinding", NULL, &weglUniformBlockBinding, &ecb_glUniformBlockBinding},
  {5582, "glDrawElementsBaseVertex", NULL, &weglDrawElementsBaseVertex, &ecb_glDrawElementsBaseVertex},
  {5583, "unused", NULL, NULL, NULL},
  {5584, "glDrawRangeElementsBaseVertex", NULL, &weglDrawRangeElementsBaseVertex, &ecb_glDrawRangeElementsBaseVertex},
  {5585, "unused", NULL, NULL, NULL},
  {5586, "glDrawElementsInstancedBaseVertex", NULL, &weglDrawElementsInstancedBaseVertex, &ecb_glDrawElementsInstancedBaseVertex},
  {5587, "unused", NULL, NULL, NULL},
  {5588, "glProvokingVertex", NULL, &weglProvokingVertex, &ecb_glProvokingVertex},
  {5589, "glFenceSync", NULL, &weglFenceSync, &ecb_glFenceSync},
  {5590, "glIsSync", NULL, &weglIsSync, &ecb_glIsSync},
  {5591, "glDeleteSync", NULL, &weglDeleteSync, &ecb_glDeleteSync},
  {5592, "glClientWaitSync", NULL, &weglClientWaitSync, &ecb_glClientWaitSync},
  {5593, "glWaitSync", NULL, &weglWaitSync, &ecb_glWaitSync},
  {5594, "glGetInteger64v", NULL, &weglGetInteger64v, &ecb_glGetInteger64v},
  {5595, "glGetSynciv", NULL, &weglGetSynciv, &ecb_glGetSynciv},
  {5596, "glGetInteger64i_v", NULL, &weglGetInteger64i_v, &ecb_glGetInteger64i_v},
  {5597, "glGetBufferParameteri64v", NULL, &weglGetBufferParameteri64v, &ecb_glGetBufferParameteri64v},
  {5598, "glFramebufferTexture", "glFramebufferTextureARB", &weglFramebufferTexture, &ecb_glFramebufferTexture},
  {5599, "glTexImage2DMultisample", NULL, &weglTexImage2DMultisample, &ecb_glTexImage2DMultisample},
  {5600, "glTexImage3DMultisample", NULL, &weglTexImage3DMultisample, &ecb_glTexImage3DMultisample},
  {5601, "glGetMultisamplefv", NULL, &weglGetMultisamplefv, &ecb_glGetMultisamplefv},
  {5602, "glSampleMaski", NULL, &weglSampleMaski, &ecb_glSampleMaski},
  {5603, "glBindFragDataLocationIndexed", NULL, &weglBindFragDataLocationIndexed, &ecb_glBindFragDataLocationIndexed},
  {5604, "glGetFragDataIndex", NULL, &weglGetFragDataIndex, &ecb_glGetFragDataIndex},
  {5605, "glGenSamplers", NULL, &weglGenSamplers, &ecb_glGenSamplers},
  {5606, "glDeleteSamplers", NULL, &weglDeleteSamplers, &ecb_glDeleteSamplers},
  {5607, "glIsSampler", NULL, &weglIsSampler, &ecb_glIsSampler},
  {5608, "glBindSampler", NULL, &weglBindSampler, &ecb_glBindSampler},
  {5609, "glSamplerParameteri", NULL, &weglSamplerParameteri, &ecb_glSamplerParameteri},
  {5610, "glSamplerParameteriv", NULL, &weglSamplerParameteriv, &ecb_glSamplerParameteriv},
  {5611, "glSamplerParameterf", NULL, &weglSamplerParameterf, &ecb_glSamplerParameterf},
  {5612, "glSamplerParameterfv", NULL, &weglSamplerParameterfv, &ecb_glSamplerParameterfv},
  {5613, "glSamplerParameterIiv", NULL, &weglSamplerParameterIiv, &ecb_glSamplerParameterIiv},
  {5614, "glSamplerParameterIuiv", NULL, &weglSamplerParameterIuiv, &ecb_glSamplerParameterIuiv},
  {5615, "glGetSamplerParameteriv", NULL, &weglGetSamplerParameteriv, &ecb_glGetSamplerParameteriv},
  {5616, "glGetSamplerParameterIiv", NULL, &weglGetSamplerParameterIiv, &ecb_glGetSamplerParameterIiv},
  {5617, "glGetSamplerParameterfv", NULL, &weglGetSamplerParameterfv, &ecb_glGetSamplerParameterfv},
  {5618, "glGetSamplerParameterIuiv", NULL, &weglGetSamplerParameterIuiv, &ecb_glGetSamplerParameterIuiv},
  {5619, "glQueryCounter", NULL, &weglQueryCounter, &ecb_glQueryCounter},
  {5620, "glGetQueryObjecti64v", NULL, &weglGetQueryObjecti64v, &ecb_glGetQueryObjecti64v},
  {5621, "glGetQueryObjectui64v", NULL, &weglGetQueryObjectui64v, &ecb_glGetQueryObjectui64v},
  {5622, "glVertexAttribDivisor", "glVertexAttribDivisorARB", &weglVertexAttribDivisor, &ecb_glVertexAttribDivisor},
  {5623, "glMinSampleShading", "glMinSampleShadingARB", &weglMinSampleShading, &ecb_glMinSampleShading},
  {5624, "glBlendEquationi", "glBlendEquationiARB", &weglBlendEquationi, &ecb_glBlendEquationi},
  {5625, "glBlendEquationSeparatei", "glBlendEquationSeparateiARB", &weglBlendEquationSeparatei, &ecb_glBlendEquationSeparatei},
  {5626, "glBlendFunci", "glBlendFunciARB", &weglBlendFunci, &ecb_glBlendFunci},
  {5627, "glBlendFuncSeparatei", "glBlendFuncSeparateiARB", &weglBlendFuncSeparatei, &ecb_glBlendFuncSeparatei},
  {5628, "glDrawArraysIndirect", NULL, &weglDrawArraysIndirect, &ecb_glDrawArraysIndirect},
  {5629, "unused", NULL, NULL, NULL},
  {5630, "glDrawElementsIndirect", NULL, &weglDrawElementsIndirect, &ecb_glDrawElementsIndirect},
  {5631, "unused", NULL, NULL, NULL},
  {5632, "glUniform1d", NULL, &weglUniform1d, &ecb_glUniform1d},
  {5633, "glUniform2d", NULL, &weglUniform2d, &ecb_glUniform2d},
  {5634, "glUniform3d", NULL, &weglUniform3d, &ecb_glUniform3d},
  {5635, "glUniform4d", NULL, &weglUniform4d, &ecb_glUniform4d},
  {5636, "glUniform1dv", NULL, &weglUniform1dv, &ecb_glUniform1dv},
  {5637, "glUniform2dv", NULL, &weglUniform2dv, &ecb_glUniform2dv},
  {5638, "glUniform3dv", NULL, &weglUniform3dv, &ecb_glUniform3dv},
  {5639, "glUniform4dv", NULL, &weglUniform4dv, &ecb_glUniform4dv},
  {5640, "glUniformMatrix2dv", NULL, &weglUniformMatrix2dv, &ecb_glUniformMatrix2dv},
  {5641, "glUniformMatrix3dv", NULL, &weglUniformMatrix3dv, &ecb_glUniformMatrix3dv},
  {5642, "glUniformMatrix4dv", NULL, &weglUniformMatrix4dv, &ecb_glUniformMatrix4dv},
  {5643, "glUniformMatrix2x3dv", NULL, &weglUniformMatrix2x3dv, &ecb_glUniformMatrix2x3dv},
  {5644, "glUniformMatrix2x4dv", NULL, &weglUniformMatrix2x4dv, &ecb_glUniformMatrix2x4dv},
  {5645, "glUniformMatrix3x2dv", NULL, &weglUniformMatrix3x2dv, &ecb_glUniformMatrix3x2dv},
  {5646, "glUniformMatrix3x4dv", NULL, &weglUniformMatrix3x4dv, &ecb_glUniformMatrix3x4dv},
  {5647, "glUniformMatrix4x2dv", NULL, &weglUniformMatrix4x2dv, &ecb_glUniformMatrix4x2dv},
  {5648, "glUniformMatrix4x3dv", NULL, &weglUniformMatrix4x3dv, &ecb_glUniformMatrix4x3dv},
  {5649, "glGetUniformdv", NULL, &weglGetUniformdv, &ecb_glGetUniformdv},
  {5650, "glGetSubroutineUniformLocation", NULL, &weglGetSubroutineUniformLocation, &ecb_glGetSubroutineUniformLocation},
  {5651, "glGetSubroutineIndex", NULL, &weglGetSubroutineIndex, &ecb_glGetSubroutineIndex},
  {5652, "glGetActiveSubroutineUniformName", NULL, &weglGetActiveSubroutineUniformName, &ecb_glGetActiveSubroutineUniformName},
  {5653, "glGetActiveSubroutineName", NULL, &weglGetActiveSubroutineName, &ecb_glGetActiveSubroutineName},
  {5654, "glUniformSubroutinesuiv", NULL, &weglUniformSubroutinesuiv, &ecb_glUniformSubroutinesuiv},
  {5655, "glGetUniformSubroutineuiv", NULL, &weglGetUniformSubroutineuiv, &ecb_glGetUniformSubroutineuiv},
  {5656, "glGetProgramStageiv", NULL, &weglGetProgramStageiv, &ecb_glGetProgramStageiv},
  {5657, "glPatchParameteri", NULL, &weglPatchParameteri, &ecb_glPatchParameteri},
  {5658, "glPatchParameterfv", NULL, &weglPatchParameterfv, &ecb_glPatchParameterfv},
  {5659, "glBindTransformFeedback", NULL, &weglBindTransformFeedback, &ecb_glBindTransformFeedback},
  {5660, "glDeleteTransformFeedbacks", NULL, &weglDeleteTransformFeedbacks, &ecb_glDeleteTransformFeedbacks},
  {5661, "glGenTransformFeedbacks", NULL, &weglGenTransformFeedbacks, &ecb_glGenTransformFeedbacks},
  {5662, "glIsTransformFeedback", NULL, &weglIsTransformFeedback, &ecb_glIsTransformFeedback},
  {5663, "glPauseTransformFeedback", NULL, &weglPauseTransformFeedback, &ecb_glPauseTransformFeedback},
  {5664, "glResumeTransformFeedback", NULL, &weglResumeTransformFeedback, &ecb_glResumeTransformFeedback},
  {5665, "glDrawTransformFeedback", NULL, &weglDrawTransformFeedback, &ecb_glDrawTransformFeedback},
  {5666, "glDrawTransformFeedbackStream", NULL, &weglDrawTransformFeedbackStream, &ecb_glDrawTransformFeedbackStream},
  {5667, "glBeginQueryIndexed", NULL, &weglBeginQueryIndexed, &ecb_glBeginQueryIndexed},
  {5668, "glEndQueryIndexed", NULL, &weglEndQueryIndexed, &ecb_glEndQueryIndexed},
  {5669, "glGetQueryIndexediv", NULL, &weglGetQueryIndexediv, &ecb_glGetQueryIndexediv},
  {5670, "glReleaseShaderCompiler", NULL, &weglReleaseShaderCompiler, &ecb_glReleaseShaderCompiler},
  {5671, "glShaderBinary", NULL, &weglShaderBinary, &ecb_glShaderBinary},
  {5672, "glGetShaderPrecisionFormat", NULL, &weglGetShaderPrecisionFormat, &ecb_glGetShaderPrecisionFormat},
  {5673, "glDepthRangef", NULL, &weglDepthRangef, &ecb_glDepthRangef},
  {5674, "glClearDepthf", NULL, &weglClearDepthf, &ecb_glClearDepthf},
  {5675, "glGetProgramBinary", NULL, &weglGetProgramBinary, &ecb_glGetProgramBinary},
  {5676, "glProgramBinary", NULL, &weglProgramBinary, &ecb_glProgramBinary},
  {5677, "glProgramParameteri", "glProgramParameteriARB", &weglProgramParameteri, &ecb_glProgramParameteri},
  {5678, "glUseProgramStages", NULL, &weglUseProgramStages, &ecb_glUseProgramStages},
  {5679, "glActiveShaderProgram", NULL, &weglActiveShaderProgram, &ecb_glActiveShaderProgram},
  {5680, "glCreateShaderProgramv", NULL, &weglCreateShaderProgramv, &ecb_glCreateShaderProgramv},
  {5681, "glBindProgramPipeline", NULL, &weglBindProgramPipeline, &ecb_glBindProgramPipeline},
  {5682, "glDeleteProgramPipelines", NULL, &weglDeleteProgramPipelines, &ecb_glDeleteProgramPipelines},
  {5683, "glGenProgramPipelines", NULL, &weglGenProgramPipelines, &ecb_glGenProgramPipelines},
  {5684, "glIsProgramPipeline", NULL, &weglIsProgramPipeline, &ecb_glIsProgramPipeline},
  {5685, "glGetProgramPipelineiv", NULL, &weglGetProgramPipelineiv, &ecb_glGetProgramPipelineiv},
  {5686, "glProgramUniform1i", NULL, &weglProgramUniform1i, &ecb_glProgramUniform1i},
  {5687, "glProgramUniform1iv", NULL, &weglProgramUniform1iv, &ecb_glProgramUniform1iv},
  {5688, "glProgramUniform1f", NULL, &weglProgramUniform1f, &ecb_glProgramUniform1f},
  {5689, "glProgramUniform1fv", NULL, &weglProgramUniform1fv, &ecb_glProgramUniform1fv},
  {5690, "glProgramUniform1d", NULL, &weglProgramUniform1d, &ecb_glProgramUniform1d},
  {5691, "glProgramUniform1dv", NULL, &weglProgramUniform1dv, &ecb_glProgramUniform1dv},
  {5692, "glProgramUniform1ui", NULL, &weglProgramUniform1ui, &ecb_glProgramUniform1ui},
  {5693, "glProgramUniform1uiv", NULL, &weglProgramUniform1uiv, &ecb_glProgramUniform1uiv},
  {5694, "glProgramUniform2i", NULL, &weglProgramUniform2i, &ecb_glProgramUniform2i},
  {5695, "glProgramUniform2iv", NULL, &weglProgramUniform2iv, &ecb_glProgramUniform2iv},
  {5696, "glProgramUniform2f", NULL, &weglProgramUniform2f, &ecb_glProgramUniform2f},
  {5697, "glProgramUniform2fv", NULL, &weglProgramUniform2fv, &ecb_glProgramUniform2fv},
  {5698, "glProgramUniform2d", NULL, &weglProgramUniform2d, &ecb_glProgramUniform2d},
  {5699, "glProgramUniform2dv", NULL, &weglProgramUniform2dv, &ecb_glProgramUniform2dv},
  {5700, "glProgramUniform2ui", NULL, &weglProgramUniform2ui, &ecb_glProgramUniform2ui},
  {5701, "glProgramUniform2uiv", NULL, &weglProgramUniform2uiv, &ecb_glProgramUniform2uiv},
  {5702, "glProgramUniform3i", NULL, &weglProgramUniform3i, &ecb_glProgramUniform3i},
  {5703, "glProgramUniform3iv", NULL, &weglProgramUniform3iv, &ecb_glProgramUniform3iv},
  {5704, "glProgramUniform3f", NULL, &weglProgramUniform3f, &ecb_glProgramUniform3f},
  {5705, "glProgramUniform3fv", NULL, &weglProgramUniform3fv, &ecb_glProgramUniform3fv},
  {5706, "glProgramUniform3d", NULL, &weglProgramUniform3d, &ecb_glProgramUniform3d},
  {5707, "glProgramUniform3dv", NULL, &weglProgramUniform3dv, &ecb_glProgramUniform3dv},
  {5708, "glProgramUniform3ui", NULL, &weglProgramUniform3ui, &ecb_glProgramUniform3ui},
  {5709, "glProgramUniform3uiv", NULL, &weglProgramUniform3uiv, &ecb_glProgramUniform3uiv},
  {5710, "glProgramUniform4i", NULL, &weglProgramUniform4i, &ecb_glProgramUniform4i},
  {5711, "glProgramUniform4iv", NULL, &weglProgramUniform4iv, &ecb_glProgramUniform4iv},
  {5712, "glProgramUniform4f", NULL, &weglProgramUniform4f, &ecb_glProgramUniform4f},
  {5713, "glProgramUniform4fv", NULL, &weglProgramUniform4fv, &ecb_glProgramUniform4fv},
  {5714, "glProgramUniform4d", NULL, &weglProgramUniform4d, &ecb_glProgramUniform4d},
  {5715, "glProgramUniform4dv", NULL, &weglProgramUniform4dv, &ecb_glProgramUniform4dv},
  {5716, "glProgramUniform4ui", NULL, &weglProgramUniform4ui, &ecb_glProgramUniform4ui},
  {5717, "glProgramUniform4uiv", NULL, &weglProgramUniform4uiv, &ecb_glProgramUniform4uiv},
  {5718, "glProgramUniformMatrix2fv", NULL, &weglProgramUniformMatrix2fv, &ecb_glProgramUniformMatrix2fv},
  {5719, "glProgramUniformMatrix3fv", NULL, &weglProgramUniformMatrix3fv, &ecb_glProgramUniformMatrix3fv},
  {5720, "glProgramUniformMatrix4fv", NULL, &weglProgramUniformMatrix4fv, &ecb_glProgramUniformMatrix4fv},
  {5721, "glProgramUniformMatrix2dv", NULL, &weglProgramUniformMatrix2dv, &ecb_glProgramUniformMatrix2dv},
  {5722, "glProgramUniformMatrix3dv", NULL, &weglProgramUniformMatrix3dv, &ecb_glProgramUniformMatrix3dv},
  {5723, "glProgramUniformMatrix4dv", NULL, &weglProgramUniformMatrix4dv, &ecb_glProgramUniformMatrix4dv},
  {5724, "glProgramUniformMatrix2x3fv", NULL, &weglProgramUniformMatrix2x3fv, &ecb_glProgramUniformMatrix2x3fv},
  {5725, "glProgramUniformMatrix3x2fv", NULL, &weglProgramUniformMatrix3x2fv, &ecb_glProgramUniformMatrix3x2fv},
  {5726, "glProgramUniformMatrix2x4fv", NULL, &weglProgramUniformMatrix2x4fv, &ecb_glProgramUniformMatrix2x4fv},
  {5727, "glProgramUniformMatrix4x2fv", NULL, &weglProgramUniformMatrix4x2fv, &ecb_glProgramUniformMatrix4x2fv},
  {5728, "glProgramUniformMatrix3x4fv", NULL, &weglProgramUniformMatrix3x4fv, &ecb_glProgramUniformMatrix3x4fv},
  {5729, "glProgramUniformMatrix4x3fv", NULL, &weglProgramUniformMatrix4x3fv, &ecb_glProgramUniformMatrix4x3fv},
  {5730, "glProgramUniformMatrix2x3dv", NULL, &weglProgramUniformMatrix2x3dv, &ecb_glProgramUniformMatrix2x3dv},
  {5731, "glProgramUniformMatrix3x2dv", NULL, &weglProgramUniformMatrix3x2dv, &ecb_glProgramUniformMatrix3x2dv},
  {5732, "glProgramUniformMatrix2x4dv", NULL, &weglProgramUniformMatrix2x4dv, &ecb_glProgramUniformMatrix2x4dv},
  {5733, "glProgramUniformMatrix4x2dv", NULL, &weglProgramUniformMatrix4x2dv, &ecb_glProgramUniformMatrix4x2dv},
  {5734, "glProgramUniformMatrix3x4dv", NULL, &weglProgramUniformMatrix3x4dv, &ecb_glProgramUniformMatrix3x4dv},
  {5735, "glProgramUniformMatrix4x3dv", NULL, &weglProgramUniformMatrix4x3dv, &ecb_glProgramUniformMatrix4x3dv},
  {5736, "glValidateProgramPipeline", NULL, &weglValidateProgramPipeline, &ecb_glValidateProgramPipeline},
  {5737, "glGetProgramPipelineInfoLog", NULL, &weglGetProgramPipelineInfoLog, &ecb_glGetProgramPipelineInfoLog},
  {5738, "glVertexAttribL1d", NULL, &weglVertexAttribL1d, &ecb_glVertexAttribL1d},
  {5739, "glVertexAttribL2d", NULL, &weglVertexAttribL2d, &ecb_glVertexAttribL2d},
  {5740, "glVertexAttribL3d", NULL, &weglVertexAttribL3d, &ecb_glVertexAttribL3d},
  {5741, "glVertexAttribL4d", NULL, &weglVertexAttribL4d, &ecb_glVertexAttribL4d},
  {5742, "glVertexAttribLPointer", NULL, &weglVertexAttribLPointer, &ecb_glVertexAttribLPointer},
  {5743, "unused", NULL, NULL, NULL},
  {5744, "glGetVertexAttribLdv", NULL, &weglGetVertexAttribLdv, &ecb_glGetVertexAttribLdv},
  {5745, "glViewportArrayv", NULL, &weglViewportArrayv, &ecb_glViewportArrayv},
  {5746, "glViewportIndexedf", NULL, &weglViewportIndexedf, &ecb_glViewportIndexedf},
  {5747, "glViewportIndexedfv", NULL, &weglViewportIndexedfv, &ecb_glViewportIndexedfv},
  {5748, "glScissorArrayv", NULL, &weglScissorArrayv, &ecb_glScissorArrayv},
  {5749, "glScissorIndexed", NULL, &weglScissorIndexed, &ecb_glScissorIndexed},
  {5750, "glScissorIndexedv", NULL, &weglScissorIndexedv, &ecb_glScissorIndexedv},
  {5751, "glDepthRangeArrayv", NULL, &weglDepthRangeArrayv, &ecb_glDepthRangeArrayv},
  {5752, "glDepthRangeIndexed", NULL, &weglDepthRangeIndexed, &ecb_glDepthRangeIndexed},
  {5753, "glGetFloati_v", NULL, &weglGetFloati_v, &ecb_glGetFloati_v},
  {5754, "glGetDoublei_v", NULL, &weglGetDoublei_v, &ecb_glGetDoublei_v},
  {5755, "glDrawArraysInstancedBaseInstance", NULL, &weglDrawArraysInstancedBaseInstance, &ecb_glDrawArraysInstancedBaseInstance},
  {5756, "glDrawElementsInstancedBaseInstance", NULL, &weglDrawElementsInstancedBaseInstance, &ecb_glDrawElementsInstancedBaseInstance},
  {5757, "unused", NULL, NULL, NULL},
  {5758, "glDrawElementsInstancedBaseVertexBaseInstance", NULL, &weglDrawElementsInstancedBaseVertexBaseInstance, &ecb_glDrawElementsInstancedBaseVertexBaseInstance},
  {5759, "unused", NULL, NULL, NULL},
  {5760, "glGetInternalformativ", NULL, &weglGetInternalformativ, &ecb_glGetInternalformativ},
  {5761, "glBindImageTexture", NULL, &weglBindImageTexture, &ecb_glBindImageTexture},
  {5762, "glMemoryBarrier", NULL, &weglMemoryBarrier, &ecb_glMemoryBarrier},
  {5763, "glTexStorage1D", NULL, &weglTexStorage1D, &ecb_glTexStorage1D},
  {5764, "glTexStorage2D", NULL, &weglTexStorage2D, &ecb_glTexStorage2D},
  {5765, "glTexStorage3D", NULL, &weglTexStorage3D, &ecb_glTexStorage3D},
  {5766, "glDrawTransformFeedbackInstanced", NULL, &weglDrawTransformFeedbackInstanced, &ecb_glDrawTransformFeedbackInstanced},
  {5767, "glDrawTransformFeedbackStreamInstanced", NULL, &weglDrawTransformFeedbackStreamInstanced, &ecb_glDrawTransformFeedbackStreamInstanced},
  {5768, "glClearBufferData", NULL, &weglClearBufferData, &ecb_glClearBufferData},
  {5769, "unused", NULL, NULL, NULL},
  {5770, "glClearBufferSubData", NULL, &weglClearBufferSubData, &ecb_glClearBufferSubData},
  {5771, "unused", NULL, NULL, NULL},
  {5772, "glDispatchCompute", NULL, &weglDispatchCompute, &ecb_glDispatchCompute},
  {5773, "glDispatchComputeIndirect", NULL, &weglDispatchComputeIndirect, &ecb_glDispatchComputeIndirect},
  {5774, "glCopyImageSubData", NULL, &weglCopyImageSubData, &ecb_glCopyImageSubData},
  {5775, "glFramebufferParameteri", NULL, &weglFramebufferParameteri, &ecb_glFramebufferParameteri},
  {5776, "glGetFramebufferParameteriv", NULL, &weglGetFramebufferParameteriv, &ecb_glGetFramebufferParameteriv},
  {5777, "glGetInternalformati64v", NULL, &weglGetInternalformati64v, &ecb_glGetInternalformati64v},
  {5778, "glInvalidateTexSubImage", NULL, &weglInvalidateTexSubImage, &ecb_glInvalidateTexSubImage},
  {5779, "glInvalidateTexImage", NULL, &weglInvalidateTexImage, &ecb_glInvalidateTexImage},
  {5780, "glInvalidateBufferSubData", NULL, &weglInvalidateBufferSubData, &ecb_glInvalidateBufferSubData},
  {5781, "glInvalidateBufferData", NULL, &weglInvalidateBufferData, &ecb_glInvalidateBufferData},
  {5782, "glInvalidateFramebuffer", NULL, &weglInvalidateFramebuffer, &ecb_glInvalidateFramebuffer},
  {5783, "glInvalidateSubFramebuffer", NULL, &weglInvalidateSubFramebuffer, &ecb_glInvalidateSubFramebuffer},
  {5784, "glMultiDrawArraysIndirect", NULL, &weglMultiDrawArraysIndirect, &ecb_glMultiDrawArraysIndirect},
  {5785, "unused", NULL, NULL, NULL},
  {5786, "glGetProgramInterfaceiv", NULL, &weglGetProgramInterfaceiv, &ecb_glGetProgramInterfaceiv},
  {5787, "glGetProgramResourceIndex", NULL, &weglGetProgramResourceIndex, &ecb_glGetProgramResourceIndex},
  {5788, "glGetProgramResourceName", NULL, &weglGetProgramResourceName, &ecb_glGetProgramResourceName},
  {5789, "glGetProgramResourceLocation", NULL, &weglGetProgramResourceLocation, &ecb_glGetProgramResourceLocation},
  {5790, "glGetProgramResourceLocationIndex", NULL, &weglGetProgramResourceLocationIndex, &ecb_glGetProgramResourceLocationIndex},
  {5791, "glShaderStorageBlockBinding", NULL, &weglShaderStorageBlockBinding, &ecb_glShaderStorageBlockBinding},
  {5792, "glTexBufferRange", NULL, &weglTexBufferRange, &ecb_glTexBufferRange},
  {5793, "glTexStorage2DMultisample", NULL, &weglTexStorage2DMultisample, &ecb_glTexStorage2DMultisample},
  {5794, "glTexStorage3DMultisample", NULL, &weglTexStorage3DMultisample, &ecb_glTexStorage3DMultisample},
  {5795, "glTextureView", NULL, &weglTextureView, &ecb_glTextureView},
  {5796, "glBindVertexBuffer", NULL, &weglBindVertexBuffer, &ecb_glBindVertexBuffer},
  {5797, "glVertexAttribFormat", NULL, &weglVertexAttribFormat, &ecb_glVertexAttribFormat},
  {5798, "glVertexAttribIFormat", NULL, &weglVertexAttribIFormat, &ecb_glVertexAttribIFormat},
  {5799, "glVertexAttribLFormat", NULL, &weglVertexAttribLFormat, &ecb_glVertexAttribLFormat},
  {5800, "glVertexAttribBinding", NULL, &weglVertexAttribBinding, &ecb_glVertexAttribBinding},
  {5801, "glVertexBindingDivisor", NULL, &weglVertexBindingDivisor, &ecb_glVertexBindingDivisor},
  {5802, "glDebugMessageControl", "glDebugMessageControlARB", &weglDebugMessageControl, &ecb_glDebugMessageControl},
  {5803, "glDebugMessageInsert", "glDebugMessageInsertARB", &weglDebugMessageInsert, &ecb_glDebugMessageInsert},
  {5804, "glGetDebugMessageLog", "glGetDebugMessageLogARB", &weglGetDebugMessageLog, &ecb_glGetDebugMessageLog},
  {5805, "glPushDebugGroup", NULL, &weglPushDebugGroup, &ecb_glPushDebugGroup},
  {5806, "glPopDebugGroup", NULL, &weglPopDebugGroup, &ecb_glPopDebugGroup},
  {5807, "glObjectPtrLabel", NULL, &weglObjectPtrLabel, &ecb_glObjectPtrLabel},
  {5808, "unused", NULL, NULL, NULL},
  {5809, "glBufferStorage", NULL, &weglBufferStorage, &ecb_glBufferStorage},
  {5810, "unused", NULL, NULL, NULL},
  {5811, "glClearTexImage", NULL, &weglClearTexImage, &ecb_glClearTexImage},
  {5812, "unused", NULL, NULL, NULL},
  {5813, "glClearTexSubImage", NULL, &weglClearTexSubImage, &ecb_glClearTexSubImage},
  {5814, "unused", NULL, NULL, NULL},
  {5815, "glBindBuffersBase", NULL, &weglBindBuffersBase, &ecb_glBindBuffersBase},
  {5816, "glBindBuffersRange", NULL, &weglBindBuffersRange, &ecb_glBindBuffersRange},
  {5817, "glBindTextures", NULL, &weglBindTextures, &ecb_glBindTextures},
  {5818, "glBindSamplers", NULL, &weglBindSamplers, &ecb_glBindSamplers},
  {5819, "glBindImageTextures", NULL, &weglBindImageTextures, &ecb_glBindImageTextures},
  {5820, "glBindVertexBuffers", NULL, &weglBindVertexBuffers, &ecb_glBindVertexBuffers},
  {5821, "glClipControl", NULL, &weglClipControl, &ecb_glClipControl},
  {5822, "glCreateTransformFeedbacks", NULL, &weglCreateTransformFeedbacks, &ecb_glCreateTransformFeedbacks},
  {5823, "glTransformFeedbackBufferBase", NULL, &weglTransformFeedbackBufferBase, &ecb_glTransformFeedbackBufferBase},
  {5824, "glTransformFeedbackBufferRange", NULL, &weglTransformFeedbackBufferRange, &ecb_glTransformFeedbackBufferRange},
  {5825, "glCreateBuffers", NULL, &weglCreateBuffers, &ecb_glCreateBuffers},
  {5826, "glFlushMappedNamedBufferRange", NULL, &weglFlushMappedNamedBufferRange, &ecb_glFlushMappedNamedBufferRange},
  {5827, "glCreateFramebuffers", NULL, &weglCreateFramebuffers, &ecb_glCreateFramebuffers},
  {5828, "glCreateRenderbuffers", NULL, &weglCreateRenderbuffers, &ecb_glCreateRenderbuffers},
  {5829, "glCreateTextures", NULL, &weglCreateTextures, &ecb_glCreateTextures},
  {5830, "glTextureBuffer", NULL, &weglTextureBuffer, &ecb_glTextureBuffer},
  {5831, "glTextureBufferRange", NULL, &weglTextureBufferRange, &ecb_glTextureBufferRange},
  {5832, "glCompressedTextureSubImage1D", NULL, &weglCompressedTextureSubImage1D, &ecb_glCompressedTextureSubImage1D},
  {5833, "unused", NULL, NULL, NULL},
  {5834, "glCompressedTextureSubImage2D", NULL, &weglCompressedTextureSubImage2D, &ecb_glCompressedTextureSubImage2D},
  {5835, "unused", NULL, NULL, NULL},
  {5836, "glCompressedTextureSubImage3D", NULL, &weglCompressedTextureSubImage3D, &ecb_glCompressedTextureSubImage3D},
  {5837, "unused", NULL, NULL, NULL},
  {5838, "glGenerateTextureMipmap", NULL, &weglGenerateTextureMipmap, &ecb_glGenerateTextureMipmap},
  {5839, "glBindTextureUnit", NULL, &weglBindTextureUnit, &ecb_glBindTextureUnit},
  {5840, "glCreateVertexArrays", NULL, &weglCreateVertexArrays, &ecb_glCreateVertexArrays},
  {5841, "glDisableVertexArrayAttrib", NULL, &weglDisableVertexArrayAttrib, &ecb_glDisableVertexArrayAttrib},
  {5842, "glEnableVertexArrayAttrib", NULL, &weglEnableVertexArrayAttrib, &ecb_glEnableVertexArrayAttrib},
  {5843, "glVertexArrayElementBuffer", NULL, &weglVertexArrayElementBuffer, &ecb_glVertexArrayElementBuffer},
  {5844, "glVertexArrayVertexBuffer", NULL, &weglVertexArrayVertexBuffer, &ecb_glVertexArrayVertexBuffer},
  {5845, "glVertexArrayVertexBuffers", NULL, &weglVertexArrayVertexBuffers, &ecb_glVertexArrayVertexBuffers},
  {5846, "glVertexArrayAttribBinding", NULL, &weglVertexArrayAttribBinding, &ecb_glVertexArrayAttribBinding},
  {5847, "glVertexArrayAttribFormat", NULL, &weglVertexArrayAttribFormat, &ecb_glVertexArrayAttribFormat},
  {5848, "glVertexArrayAttribIFormat", NULL, &weglVertexArrayAttribIFormat, &ecb_glVertexArrayAttribIFormat},
  {5849, "glVertexArrayAttribLFormat", NULL, &weglVertexArrayAttribLFormat, &ecb_glVertexArrayAttribLFormat},
  {5850, "glVertexArrayBindingDivisor", NULL, &weglVertexArrayBindingDivisor, &ecb_glVertexArrayBindingDivisor},
  {5851, "glCreateSamplers", NULL, &weglCreateSamplers, &ecb_glCreateSamplers},
  {5852, "glCreateProgramPipelines", NULL, &weglCreateProgramPipelines, &ecb_glCreateProgramPipelines},
  {5853, "glCreateQueries", NULL, &weglCreateQueries, &ecb_glCreateQueries},
  {5854, "glGetQueryBufferObjecti64v", NULL, &weglGetQueryBufferObjecti64v, &ecb_glGetQueryBufferObjecti64v},
  {5855, "glGetQueryBufferObjectiv", NULL, &weglGetQueryBufferObjectiv, &ecb_glGetQueryBufferObjectiv},
  {5856, "glGetQueryBufferObjectui64v", NULL, &weglGetQueryBufferObjectui64v, &ecb_glGetQueryBufferObjectui64v},
  {5857, "glGetQueryBufferObjectuiv", NULL, &weglGetQueryBufferObjectuiv, &ecb_glGetQueryBufferObjectuiv},
  {5858, "glMemoryBarrierByRegion", NULL, &weglMemoryBarrierByRegion, &ecb_glMemoryBarrierByRegion},
  {5859, "glGetGraphicsResetStatus", "glGetGraphicsResetStatusARB", &weglGetGraphicsResetStatus, &ecb_glGetGraphicsResetStatus},
  {5860, "glTextureBarrier", NULL, &weglTextureBarrier, &ecb_glTextureBarrier},
  {5861, "glMultiDrawArraysIndirectCount", "glMultiDrawArraysIndirectCountARB", &weglMultiDrawArraysIndirectCount, &ecb_glMultiDrawArraysIndirectCount},
  {5862, "unused", NULL, NULL, NULL},
  {5863, "glPolygonOffsetClamp", NULL, &weglPolygonOffsetClamp, &ecb_glPolygonOffsetClamp},
  {5864, "glPrimitiveBoundingBoxARB", NULL, &weglPrimitiveBoundingBoxARB, &ecb_glPrimitiveBoundingBoxARB},
  {5865, "glMakeTextureHandleResidentARB", NULL, &weglMakeTextureHandleResidentARB, &ecb_glMakeTextureHandleResidentARB},
  {5866, "glMakeTextureHandleNonResidentARB", NULL, &weglMakeTextureHandleNonResidentARB, &ecb_glMakeTextureHandleNonResidentARB},
  {5867, "glGetImageHandleARB", NULL, &weglGetImageHandleARB, &ecb_glGetImageHandleARB},
  {5868, "glMakeImageHandleResidentARB", NULL, &weglMakeImageHandleResidentARB, &ecb_glMakeImageHandleResidentARB},
  {5869, "glMakeImageHandleNonResidentARB", NULL, &weglMakeImageHandleNonResidentARB, &ecb_glMakeImageHandleNonResidentARB},
  {5870, "glUniformHandleui64ARB", NULL, &weglUniformHandleui64ARB, &ecb_glUniformHandleui64ARB},
  {5871, "glProgramUniformHandleui64ARB", NULL, &weglProgramUniformHandleui64ARB, &ecb_glProgramUniformHandleui64ARB},
  {5872, "glIsTextureHandleResidentARB", NULL, &weglIsTextureHandleResidentARB, &ecb_glIsTextureHandleResidentARB},
  {5873, "glIsImageHandleResidentARB", NULL, &weglIsImageHandleResidentARB, &ecb_glIsImageHandleResidentARB},
  {5874, "glDispatchComputeGroupSizeARB", NULL, &weglDispatchComputeGroupSizeARB, &ecb_glDispatchComputeGroupSizeARB},
  {5875, "glProgramStringARB", NULL, &weglProgramStringARB, &ecb_glProgramStringARB},
  {5876, "glBindProgramARB", NULL, &weglBindProgramARB, &ecb_glBindProgramARB},
  {5877, "glDeleteProgramsARB", NULL, &weglDeleteProgramsARB, &ecb_glDeleteProgramsARB},
  {5878, "glGenProgramsARB", NULL, &weglGenProgramsARB, &ecb_glGenProgramsARB},
  {5879, "glProgramEnvParameter4dARB", NULL, &weglProgramEnvParameter4dARB, &ecb_glProgramEnvParameter4dARB},
  {5880, "glProgramEnvParameter4dvARB", NULL, &weglProgramEnvParameter4dvARB, &ecb_glProgramEnvParameter4dvARB},
  {5881, "glProgramEnvParameter4fARB", NULL, &weglProgramEnvParameter4fARB, &ecb_glProgramEnvParameter4fARB},
  {5882, "glProgramEnvParameter4fvARB", NULL, &weglProgramEnvParameter4fvARB, &ecb_glProgramEnvParameter4fvARB},
  {5883, "glProgramLocalParameter4dARB", NULL, &weglProgramLocalParameter4dARB, &ecb_glProgramLocalParameter4dARB},
  {5884, "glProgramLocalParameter4dvARB", NULL, &weglProgramLocalParameter4dvARB, &ecb_glProgramLocalParameter4dvARB},
  {5885, "glProgramLocalParameter4fARB", NULL, &weglProgramLocalParameter4fARB, &ecb_glProgramLocalParameter4fARB},
  {5886, "glProgramLocalParameter4fvARB", NULL, &weglProgramLocalParameter4fvARB, &ecb_glProgramLocalParameter4fvARB},
  {5887, "glGetProgramEnvParameterdvARB", NULL, &weglGetProgramEnvParameterdvARB, &ecb_glGetProgramEnvParameterdvARB},
  {5888, "glGetProgramEnvParameterfvARB", NULL, &weglGetProgramEnvParameterfvARB, &ecb_glGetProgramEnvParameterfvARB},
  {5889, "glGetProgramLocalParameterdvARB", NULL, &weglGetProgramLocalParameterdvARB, &ecb_glGetProgramLocalParameterdvARB},
  {5890, "glGetProgramLocalParameterfvARB", NULL, &weglGetProgramLocalParameterfvARB, &ecb_glGetProgramLocalParameterfvARB},
  {5891, "glGetProgramStringARB", NULL, &weglGetProgramStringARB, &ecb_glGetProgramStringARB},
  {5892, "glFramebufferTextureFaceARB", NULL, &weglFramebufferTextureFaceARB, &ecb_glFramebufferTextureFaceARB},
  {5893, "glUniform1i64ARB", NULL, &weglUniform1i64ARB, &ecb_glUniform1i64ARB},
  {5894, "glUniform2i64ARB", NULL, &weglUniform2i64ARB, &ecb_glUniform2i64ARB},
  {5895, "glUniform3i64ARB", NULL, &weglUniform3i64ARB, &ecb_glUniform3i64ARB},
  {5896, "glUniform4i64ARB", NULL, &weglUniform4i64ARB, &ecb_glUniform4i64ARB},
  {5897, "glUniform1i64vARB", NULL, &weglUniform1i64vARB, &ecb_glUniform1i64vARB},
  {5898, "glUniform2i64vARB", NULL, &weglUniform2i64vARB, &ecb_glUniform2i64vARB},
  {5899, "glUniform3i64vARB", NULL, &weglUniform3i64vARB, &ecb_glUniform3i64vARB},
  {5900, "glUniform4i64vARB", NULL, &weglUniform4i64vARB, &ecb_glUniform4i64vARB},
  {5901, "glUniform1ui64ARB", NULL, &weglUniform1ui64ARB, &ecb_glUniform1ui64ARB},
  {5902, "glUniform2ui64ARB", NULL, &weglUniform2ui64ARB, &ecb_glUniform2ui64ARB},
  {5903, "glUniform3ui64ARB", NULL, &weglUniform3ui64ARB, &ecb_glUniform3ui64ARB},
  {5904, "glUniform4ui64ARB", NULL, &weglUniform4ui64ARB, &ecb_glUniform4ui64ARB},
  {5905, "glUniform1ui64vARB", NULL, &weglUniform1ui64vARB, &ecb_glUniform1ui64vARB},
  {5906, "glUniform2ui64vARB", NULL, &weglUniform2ui64vARB, &ecb_glUniform2ui64vARB},
  {5907, "glUniform3ui64vARB", NULL, &weglUniform3ui64vARB, &ecb_glUniform3ui64vARB},
  {5908, "glUniform4ui64vARB", NULL, &weglUniform4ui64vARB, &ecb_glUniform4ui64vARB},
  {5909, "glGetUniformi64vARB", NULL, &weglGetUniformi64vARB, &ecb_glGetUniformi64vARB},
  {5910, "glGetUniformui64vARB", NULL, &weglGetUniformui64vARB, &ecb_glGetUniformui64vARB},
  {5911, "glProgramUniform1i64ARB", NULL, &weglProgramUniform1i64ARB, &ecb_glProgramUniform1i64ARB},
  {5912, "glProgramUniform2i64ARB", NULL, &weglProgramUniform2i64ARB, &ecb_glProgramUniform2i64ARB},
  {5913, "glProgramUniform3i64ARB", NULL, &weglProgramUniform3i64ARB, &ecb_glProgramUniform3i64ARB},
  {5914, "glProgramUniform4i64ARB", NULL, &weglProgramUniform4i64ARB, &ecb_glProgramUniform4i64ARB},
  {5915, "glProgramUniform1i64vARB", NULL, &weglProgramUniform1i64vARB, &ecb_glProgramUniform1i64vARB},
  {5916, "glProgramUniform2i64vARB", NULL, &weglProgramUniform2i64vARB, &ecb_glProgramUniform2i64vARB},
  {5917, "glProgramUniform3i64vARB", NULL, &weglProgramUniform3i64vARB, &ecb_glProgramUniform3i64vARB},
  {5918, "glProgramUniform4i64vARB", NULL, &weglProgramUniform4i64vARB, &ecb_glProgramUniform4i64vARB},
  {5919, "glProgramUniform1ui64ARB", NULL, &weglProgramUniform1ui64ARB, &ecb_glProgramUniform1ui64ARB},
  {5920, "glProgramUniform2ui64ARB", NULL, &weglProgramUniform2ui64ARB, &ecb_glProgramUniform2ui64ARB},
  {5921, "glProgramUniform3ui64ARB", NULL, &weglProgramUniform3ui64ARB, &ecb_glProgramUniform3ui64ARB},
  {5922, "glProgramUniform4ui64ARB", NULL, &weglProgramUniform4ui64ARB, &ecb_glProgramUniform4ui64ARB},
  {5923, "glProgramUniform1ui64vARB", NULL, &weglProgramUniform1ui64vARB, &ecb_glProgramUniform1ui64vARB},
  {5924, "glProgramUniform2ui64vARB", NULL, &weglProgramUniform2ui64vARB, &ecb_glProgramUniform2ui64vARB},
  {5925, "glProgramUniform3ui64vARB", NULL, &weglProgramUniform3ui64vARB, &ecb_glProgramUniform3ui64vARB},
  {5926, "glProgramUniform4ui64vARB", NULL, &weglProgramUniform4ui64vARB, &ecb_glProgramUniform4ui64vARB},
  {5927, "glColorTable", NULL, &weglColorTable, &ecb_glColorTable},
  {5928, "unused", NULL, NULL, NULL},
  {5929, "glColorTableParameterfv", NULL, &weglColorTableParameterfv, &ecb_glColorTableParameterfv},
  {5930, "glColorTableParameteriv", NULL, &weglColorTableParameteriv, &ecb_glColorTableParameteriv},
  {5931, "glCopyColorTable", NULL, &weglCopyColorTable, &ecb_glCopyColorTable},
  {5932, "glGetColorTable", NULL, &weglGetColorTable, &ecb_glGetColorTable},
  {5933, "glGetColorTableParameterfv", NULL, &weglGetColorTableParameterfv, &ecb_glGetColorTableParameterfv},
  {5934, "glGetColorTableParameteriv", NULL, &weglGetColorTableParameteriv, &ecb_glGetColorTableParameteriv},
  {5935, "glColorSubTable", NULL, &weglColorSubTable, &ecb_glColorSubTable},
  {5936, "unused", NULL, NULL, NULL},
  {5937, "glCopyColorSubTable", NULL, &weglCopyColorSubTable, &ecb_glCopyColorSubTable},
  {5938, "glConvolutionFilter1D", NULL, &weglConvolutionFilter1D, &ecb_glConvolutionFilter1D},
  {5939, "unused", NULL, NULL, NULL},
  {5940, "glConvolutionFilter2D", NULL, &weglConvolutionFilter2D, &ecb_glConvolutionFilter2D},
  {5941, "unused", NULL, NULL, NULL},
  {5942, "glConvolutionParameterf", NULL, &weglConvolutionParameterf, &ecb_glConvolutionParameterf},
  {5943, "glConvolutionParameterfv", NULL, &weglConvolutionParameterfv, &ecb_glConvolutionParameterfv},
  {5944, "glConvolutionParameteri", NULL, &weglConvolutionParameteri, &ecb_glConvolutionParameteri},
  {5945, "glConvolutionParameteriv", NULL, &weglConvolutionParameteriv, &ecb_glConvolutionParameteriv},
  {5946, "glCopyConvolutionFilter1D", NULL, &weglCopyConvolutionFilter1D, &ecb_glCopyConvolutionFilter1D},
  {5947, "glCopyConvolutionFilter2D", NULL, &weglCopyConvolutionFilter2D, &ecb_glCopyConvolutionFilter2D},
  {5948, "glGetConvolutionFilter", NULL, &weglGetConvolutionFilter, &ecb_glGetConvolutionFilter},
  {5949, "glGetConvolutionParameterfv", NULL, &weglGetConvolutionParameterfv, &ecb_glGetConvolutionParameterfv},
  {5950, "glGetConvolutionParameteriv", NULL, &weglGetConvolutionParameteriv, &ecb_glGetConvolutionParameteriv},
  {5951, "glSeparableFilter2D", NULL, &weglSeparableFilter2D, &ecb_glSeparableFilter2D},
  {5952, "unused", NULL, NULL, NULL},
  {5953, "glGetHistogram", NULL, &weglGetHistogram, &ecb_glGetHistogram},
  {5954, "glGetHistogramParameterfv", NULL, &weglGetHistogramParameterfv, &ecb_glGetHistogramParameterfv},
  {5955, "glGetHistogramParameteriv", NULL, &weglGetHistogramParameteriv, &ecb_glGetHistogramParameteriv},
  {5956, "glGetMinmax", NULL, &weglGetMinmax, &ecb_glGetMinmax},
  {5957, "glGetMinmaxParameterfv", NULL, &weglGetMinmaxParameterfv, &ecb_glGetMinmaxParameterfv},
  {5958, "glGetMinmaxParameteriv", NULL, &weglGetMinmaxParameteriv, &ecb_glGetMinmaxParameteriv},
  {5959, "glHistogram", NULL, &weglHistogram, &ecb_glHistogram},
  {5960, "glMinmax", NULL, &weglMinmax, &ecb_glMinmax},
  {5961, "glResetHistogram", NULL, &weglResetHistogram, &ecb_glResetHistogram},
  {5962, "glResetMinmax", NULL, &weglResetMinmax, &ecb_glResetMinmax},
  {5963, "glCurrentPaletteMatrixARB", NULL, &weglCurrentPaletteMatrixARB, &ecb_glCurrentPaletteMatrixARB},
  {5964, "glMatrixIndexubvARB", NULL, &weglMatrixIndexubvARB, &ecb_glMatrixIndexubvARB},
  {5965, "glMatrixIndexusvARB", NULL, &weglMatrixIndexusvARB, &ecb_glMatrixIndexusvARB},
  {5966, "glMatrixIndexuivARB", NULL, &weglMatrixIndexuivARB, &ecb_glMatrixIndexuivARB},
  {5967, "glSampleCoverageARB", NULL, &weglSampleCoverageARB, &ecb_glSampleCoverageARB},
  {5968, "glMaxShaderCompilerThreadsARB", NULL, &weglMaxShaderCompilerThreadsARB, &ecb_glMaxShaderCompilerThreadsARB},
  {5969, "glEvaluateDepthValuesARB", NULL, &weglEvaluateDepthValuesARB, &ecb_glEvaluateDepthValuesARB},
  {5970, "glDeleteObjectARB", NULL, &weglDeleteObjectARB, &ecb_glDeleteObjectARB},
  {5971, "glGetHandleARB", NULL, &weglGetHandleARB, &ecb_glGetHandleARB},
  {5972, "glDetachObjectARB", NULL, &weglDetachObjectARB, &ecb_glDetachObjectARB},
  {5973, "glCreateShaderObjectARB", NULL, &weglCreateShaderObjectARB, &ecb_glCreateShaderObjectARB},
  {5974, "glShaderSourceARB", NULL, &weglShaderSourceARB, &ecb_glShaderSourceARB},
  {5975, "glCompileShaderARB", NULL, &weglCompileShaderARB, &ecb_glCompileShaderARB},
  {5976, "glCreateProgramObjectARB", NULL, &weglCreateProgramObjectARB, &ecb_glCreateProgramObjectARB},
  {5977, "glAttachObjectARB", NULL, &weglAttachObjectARB, &ecb_glAttachObjectARB},
  {5978, "glLinkProgramARB", NULL, &weglLinkProgramARB, &ecb_glLinkProgramARB},
  {5979, "glUseProgramObjectARB", NULL, &weglUseProgramObjectARB, &ecb_glUseProgramObjectARB},
  {5980, "glValidateProgramARB", NULL, &weglValidateProgramARB, &ecb_glValidateProgramARB},
  {5981, "glGetObjectParameterfvARB", NULL, &weglGetObjectParameterfvARB, &ecb_glGetObjectParameterfvARB},
  {5982, "glGetObjectParameterivARB", NULL, &weglGetObjectParameterivARB, &ecb_glGetObjectParameterivARB},
  {5983, "glGetInfoLogARB", NULL, &weglGetInfoLogARB, &ecb_glGetInfoLogARB},
  {5984, "glGetAttachedObjectsARB", NULL, &weglGetAttachedObjectsARB, &ecb_glGetAttachedObjectsARB},
  {5985, "glGetUniformLocationARB", NULL, &weglGetUniformLocationARB, &ecb_glGetUniformLocationARB},
  {5986, "glGetActiveUniformARB", NULL, &weglGetActiveUniformARB, &ecb_glGetActiveUniformARB},
  {5987, "glGetUniformfvARB", NULL, &weglGetUniformfvARB, &ecb_glGetUniformfvARB},
  {5988, "glGetUniformivARB", NULL, &weglGetUniformivARB, &ecb_glGetUniformivARB},
  {5989, "glGetShaderSourceARB", NULL, &weglGetShaderSourceARB, &ecb_glGetShaderSourceARB},
  {5990, "glDeleteNamedStringARB", NULL, &weglDeleteNamedStringARB, &ecb_glDeleteNamedStringARB},
  {5991, "glCompileShaderIncludeARB", NULL, &weglCompileShaderIncludeARB, &ecb_glCompileShaderIncludeARB},
  {5992, "glIsNamedStringARB", NULL, &weglIsNamedStringARB, &ecb_glIsNamedStringARB},
  {5993, "glBufferPageCommitmentARB", NULL, &weglBufferPageCommitmentARB, &ecb_glBufferPageCommitmentARB},
  {5994, "glTexPageCommitmentARB", NULL, &weglTexPageCommitmentARB, &ecb_glTexPageCommitmentARB},
  {5995, "glGetCompressedTexImageARB", NULL, &weglGetCompressedTexImageARB, &ecb_glGetCompressedTexImageARB},
  {5996, "glLoadTransposeMatrixfARB", NULL, &weglLoadTransposeMatrixfARB, &ecb_glLoadTransposeMatrixfARB},
  {5997, "glLoadTransposeMatrixdARB", NULL, &weglLoadTransposeMatrixdARB, &ecb_glLoadTransposeMatrixdARB},
  {5998, "glMultTransposeMatrixfARB", NULL, &weglMultTransposeMatrixfARB, &ecb_glMultTransposeMatrixfARB},
  {5999, "glMultTransposeMatrixdARB", NULL, &weglMultTransposeMatrixdARB, &ecb_glMultTransposeMatrixdARB},
  {6000, "glWeightbvARB", NULL, &weglWeightbvARB, &ecb_glWeightbvARB},
  {6001, "glWeightsvARB", NULL, &weglWeightsvARB, &ecb_glWeightsvARB},
  {6002, "glWeightivARB", NULL, &weglWeightivARB, &ecb_glWeightivARB},
  {6003, "glWeightfvARB", NULL, &weglWeightfvARB, &ecb_glWeightfvARB},
  {6004, "glWeightdvARB", NULL, &weglWeightdvARB, &ecb_glWeightdvARB},
  {6005, "glWeightubvARB", NULL, &weglWeightubvARB, &ecb_glWeightubvARB},
  {6006, "glWeightusvARB", NULL, &weglWeightusvARB, &ecb_glWeightusvARB},
  {6007, "glWeightuivARB", NULL, &weglWeightuivARB, &ecb_glWeightuivARB},
  {6008, "glVertexBlendARB", NULL, &weglVertexBlendARB, &ecb_glVertexBlendARB},
  {6009, "glGetBufferParameterivARB", NULL, &weglGetBufferParameterivARB, &ecb_glGetBufferParameterivARB},
  {6010, "glBindAttribLocationARB", NULL, &weglBindAttribLocationARB, &ecb_glBindAttribLocationARB},
  {6011, "glGetActiveAttribARB", NULL, &weglGetActiveAttribARB, &ecb_glGetActiveAttribARB},
  {6012, "glGetAttribLocationARB", NULL, &weglGetAttribLocationARB, &ecb_glGetAttribLocationARB},
  {6013, "glBlendBarrierKHR", NULL, &weglBlendBarrierKHR, &ecb_glBlendBarrierKHR},
  {6014, "glMaxShaderCompilerThreadsKHR", NULL, &weglMaxShaderCompilerThreadsKHR, &ecb_glMaxShaderCompilerThreadsKHR},
  {6015, "glDepthBoundsEXT", NULL, &weglDepthBoundsEXT, &ecb_glDepthBoundsEXT},
  { -1, NULL, NULL, NULL, NULL}};
