/*
 *  Warning: Do *not* edit this file. It was automatically
 *  generated by 'make_tables'.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif /* HAVE_CONFIG_H */
#include "export.h"
#include "sys.h"
#include "erl_vm.h"
#include "global.h"
#include "erl_process.h"
#include "bif.h"
#include "erl_bif_table.h"

const ErtsUBif erts_u_bifs[] = {
  {abs_1, BIF_abs_1},
  {element_2, BIF_element_2},
  {float_1, BIF_float_1},
  {hd_1, BIF_hd_1},
  {length_1, BIF_length_1},
  {list_to_atom_1, BIF_list_to_atom_1},
  {node_1, BIF_node_1},
  {node_0, BIF_node_0},
  {round_1, BIF_round_1},
  {self_0, BIF_self_0},
  {size_1, BIF_size_1},
  {tl_1, BIF_tl_1},
  {trunc_1, BIF_trunc_1},
  {and_2, BIF_and_2},
  {or_2, BIF_or_2},
  {xor_2, BIF_xor_2},
  {not_1, BIF_not_1},
  {sgt_2, BIF_sgt_2},
  {sge_2, BIF_sge_2},
  {slt_2, BIF_slt_2},
  {sle_2, BIF_sle_2},
  {seq_2, BIF_seq_2},
  {seqeq_2, BIF_seqeq_2},
  {sneq_2, BIF_sneq_2},
  {sneqeq_2, BIF_sneqeq_2},
  {splus_2, BIF_splus_2},
  {sminus_2, BIF_sminus_2},
  {stimes_2, BIF_stimes_2},
  {div_2, BIF_div_2},
  {intdiv_2, BIF_intdiv_2},
  {rem_2, BIF_rem_2},
  {bor_2, BIF_bor_2},
  {band_2, BIF_band_2},
  {bxor_2, BIF_bxor_2},
  {bsl_2, BIF_bsl_2},
  {bsr_2, BIF_bsr_2},
  {bnot_1, BIF_bnot_1},
  {sminus_1, BIF_sminus_1},
  {splus_1, BIF_splus_1},
  {is_atom_1, BIF_is_atom_1},
  {is_list_1, BIF_is_list_1},
  {is_tuple_1, BIF_is_tuple_1},
  {is_float_1, BIF_is_float_1},
  {is_integer_1, BIF_is_integer_1},
  {is_number_1, BIF_is_number_1},
  {is_pid_1, BIF_is_pid_1},
  {is_port_1, BIF_is_port_1},
  {is_reference_1, BIF_is_reference_1},
  {is_binary_1, BIF_is_binary_1},
  {is_function_1, BIF_is_function_1},
  {is_function_2, BIF_is_function_2},
  {is_record_2, BIF_is_record_2},
  {is_record_3, BIF_is_record_3},
  {is_boolean_1, BIF_is_boolean_1},
  {list_to_existing_atom_1, BIF_list_to_existing_atom_1},
  {is_bitstring_1, BIF_is_bitstring_1},
  {tuple_size_1, BIF_tuple_size_1},
  {byte_size_1, BIF_byte_size_1},
  {bit_size_1, BIF_bit_size_1},
  {binary_to_atom_2, BIF_binary_to_atom_2},
  {binary_to_existing_atom_2, BIF_binary_to_existing_atom_2},
  {binary_part_2, BIF_binary_part_2},
  {binary_part_3, BIF_binary_part_3},
  {is_map_1, BIF_is_map_1},
  {map_size_1, BIF_map_size_1},
  {floor_1, BIF_floor_1},
  {ceil_1, BIF_ceil_1},
  {map_get_2, BIF_map_get_2},
  {is_map_key_2, BIF_is_map_key_2},
  {min_2, BIF_min_2},
  {max_2, BIF_max_2},
  {NULL, -1}
};
