/*
 *  Warning: Do not edit this file.
 *  Auto-generated by 'beam_makeops'.
 */

void emit_aligned_label(const ArgLabel&, const ArgWord&);
void emit_allocate(const ArgWord&, const ArgWord&);
void emit_allocate_heap(const ArgWord&, const ArgWord&, const ArgWord&);
void emit_append_cons(const ArgWord&, const ArgSource&);
void emit_apply(const ArgWord&);
void emit_apply_last(const ArgWord&, const ArgWord&);
void emit_badmatch(const ArgSource&);
void emit_badrecord(const ArgSource&);
void emit_bif_bit_size(const ArgWord&, const ArgLabel&, const ArgSource&, const ArgRegister&);
void emit_bif_byte_size(const ArgWord&, const ArgLabel&, const ArgSource&, const ArgRegister&);
void emit_bif_element(const ArgLabel&, const ArgSource&, const ArgSource&, const ArgRegister&);
void emit_bif_hd(const ArgSource&, const ArgRegister&);
void emit_bif_is_eq_exact(const ArgRegister&, const ArgSource&, const ArgRegister&);
void emit_bif_is_ge(const ArgSource&, const ArgSource&, const ArgRegister&);
void emit_bif_is_lt(const ArgSource&, const ArgSource&, const ArgRegister&);
void emit_bif_is_map_key(const ArgWord&, const ArgLabel&, const ArgSource&, const ArgSource&, const ArgRegister&);
void emit_bif_is_ne_exact(const ArgRegister&, const ArgSource&, const ArgRegister&);
void emit_bif_map_get(const ArgLabel&, const ArgSource&, const ArgSource&, const ArgRegister&);
void emit_bif_map_size(const ArgLabel&, const ArgSource&, const ArgRegister&);
void emit_bif_max(const ArgSource&, const ArgSource&, const ArgRegister&);
void emit_bif_min(const ArgSource&, const ArgSource&, const ArgRegister&);
void emit_bif_node(const ArgLabel&, const ArgRegister&, const ArgRegister&);
void emit_bif_tuple_size(const ArgWord&, const ArgLabel&, const ArgRegister&, const ArgRegister&);
void emit_bs_get_integer2(const ArgLabel&, const ArgRegister&, const ArgWord&, const ArgSource&, const ArgWord&, const ArgWord&, const ArgRegister&);
void emit_bs_get_tail(const ArgRegister&, const ArgRegister&, const ArgWord&);
void emit_bs_init_writable();
void emit_bs_set_position(const ArgRegister&, const ArgRegister&);
void emit_bs_test_tail2(const ArgLabel&, const ArgRegister&, const ArgWord&);
void emit_bs_test_unit(const ArgLabel&, const ArgRegister&, const ArgWord&);
void emit_build_stacktrace();
void emit_call_bif(const ArgWord&);
void emit_call_bif_mfa(const ArgAtom&, const ArgAtom&, const ArgWord&);
void emit_call_error_handler();
void emit_call_light_bif(const ArgWord&, const ArgExport&);
void emit_call_nif(const ArgWord&, const ArgWord&, const ArgWord&);
void emit_case_end(const ArgSource&);
void emit_catch(const ArgYRegister&, const ArgLabel&);
void emit_catch_end(const ArgYRegister&);
void emit_continue_exit();
void emit_deallocate(const ArgWord&);
void emit_empty_func_line();
void emit_executable_line(const ArgWord&, const ArgWord&);
void emit_fconv(const ArgSource&, const ArgFRegister&);
void emit_fload(const ArgSource&, const ArgFRegister&);
void emit_fstore(const ArgFRegister&, const ArgRegister&);
void emit_func_line(const ArgWord&);
void emit_get_hd(const ArgRegister&, const ArgRegister&);
void emit_get_list(const ArgRegister&, const ArgRegister&, const ArgRegister&);
void emit_get_tl(const ArgRegister&, const ArgRegister&);
void emit_get_tuple_element_swap(const ArgSource&, const ArgWord&, const ArgRegister&, const ArgRegister&);
void emit_get_two_tuple_elements(const ArgSource&, const ArgWord&, const ArgRegister&, const ArgRegister&);
void emit_i_apply();
void emit_i_apply_fun();
void emit_i_apply_fun_last(const ArgWord&);
void emit_i_apply_fun_only();
void emit_i_apply_last(const ArgWord&);
void emit_i_apply_only();
void emit_i_band(const ArgSource&, const ArgSource&, const ArgLabel&, const ArgRegister&);
void emit_i_bif1(const ArgSource&, const ArgLabel&, const ArgWord&, const ArgRegister&);
void emit_i_bif2(const ArgSource&, const ArgSource&, const ArgLabel&, const ArgWord&, const ArgRegister&);
void emit_i_bif3(const ArgSource&, const ArgSource&, const ArgSource&, const ArgLabel&, const ArgWord&, const ArgRegister&);
void emit_i_bnot(const ArgLabel&, const ArgSource&, const ArgRegister&);
void emit_i_bor(const ArgLabel&, const ArgSource&, const ArgSource&, const ArgRegister&);
void emit_i_breakpoint_trampoline();
void emit_i_bs_create_bin(const ArgLabel&, const ArgWord&, const ArgWord&, const ArgRegister&, const Span<const ArgVal>&);
void emit_i_bs_get_bin_and_tail(const ArgRegister&, const ArgLabel&, const ArgWord&, const ArgRegister&, const ArgWord&, const ArgRegister&, const ArgRegister&);
void emit_i_bs_get_binary2(const ArgRegister&, const ArgLabel&, const ArgWord&, const ArgSource&, const ArgWord&, const ArgRegister&);
void emit_i_bs_get_binary_all2(const ArgRegister&, const ArgLabel&, const ArgWord&, const ArgWord&, const ArgRegister&);
void emit_i_bs_get_float2(const ArgRegister&, const ArgLabel&, const ArgWord&, const ArgSource&, const ArgWord&, const ArgRegister&);
void emit_i_bs_get_position(const ArgRegister&, const ArgRegister&);
void emit_i_bs_get_utf16(const ArgRegister&, const ArgLabel&, const ArgWord&, const ArgRegister&);
void emit_i_bs_get_utf8(const ArgRegister&, const ArgLabel&, const ArgRegister&);
void emit_i_bs_match(const ArgLabel&, const ArgRegister&, const Span<const ArgVal>&);
void emit_i_bs_match_string(const ArgRegister&, const ArgLabel&, const ArgWord&, const ArgBytePtr&);
void emit_i_bs_match_test_heap(const ArgLabel&, const ArgRegister&, const ArgWord&, const ArgWord&, const Span<const ArgVal>&);
void emit_i_bs_skip_bits2(const ArgRegister&, const ArgSource&, const ArgLabel&, const ArgWord&);
void emit_i_bs_skip_utf16(const ArgRegister&, const ArgLabel&, const ArgWord&);
void emit_i_bs_skip_utf8(const ArgRegister&, const ArgLabel&);
void emit_i_bs_start_match3(const ArgRegister&, const ArgWord&, const ArgLabel&, const ArgRegister&);
void emit_i_bs_validate_unicode_retract(const ArgLabel&, const ArgSource&, const ArgRegister&);
void emit_i_bsl(const ArgSource&, const ArgSource&, const ArgLabel&, const ArgRegister&);
void emit_i_bsr(const ArgSource&, const ArgSource&, const ArgLabel&, const ArgRegister&);
void emit_i_bxor(const ArgLabel&, const ArgSource&, const ArgSource&, const ArgRegister&);
void emit_i_call(const ArgLabel&);
void emit_i_call_ext(const ArgExport&);
void emit_i_call_ext_last(const ArgExport&, const ArgWord&);
void emit_i_call_ext_only(const ArgExport&);
void emit_i_call_fun(const ArgWord&);
void emit_i_call_fun2(const ArgVal&, const ArgWord&, const ArgRegister&);
void emit_i_call_fun2_last(const ArgVal&, const ArgWord&, const ArgRegister&, const ArgWord&);
void emit_i_call_fun_last(const ArgWord&, const ArgWord&);
void emit_i_call_last(const ArgLabel&, const ArgWord&);
void emit_i_call_on_load_function();
void emit_i_call_only(const ArgLabel&);
void emit_i_call_trace_return();
void emit_i_debug_breakpoint();
void emit_i_debug_line(const ArgWord&, const ArgWord&, const ArgWord&);
void emit_i_div_rem(const ArgSource&, const ArgSource&, const ArgLabel&, const ArgRegister&, const ArgRegister&);
void emit_i_fadd(const ArgFRegister&, const ArgFRegister&, const ArgFRegister&);
void emit_i_fdiv(const ArgFRegister&, const ArgFRegister&, const ArgFRegister&);
void emit_i_fmul(const ArgFRegister&, const ArgFRegister&, const ArgFRegister&);
void emit_i_fnegate(const ArgFRegister&, const ArgFRegister&);
void emit_i_fsub(const ArgFRegister&, const ArgFRegister&, const ArgFRegister&);
void emit_i_func_info(const ArgWord&, const ArgAtom&, const ArgAtom&, const ArgWord&);
void emit_i_func_label(const ArgLabel&);
void emit_i_generic_breakpoint();
void emit_i_get(const ArgSource&, const ArgRegister&);
void emit_i_get_hash(const ArgConstant&, const ArgWord&, const ArgRegister&);
void emit_i_get_map_element(const ArgLabel&, const ArgRegister&, const ArgRegister&, const ArgRegister&);
void emit_i_get_map_element_hash(const ArgLabel&, const ArgRegister&, const ArgConstant&, const ArgWord&, const ArgRegister&);
void emit_i_get_map_elements(const ArgLabel&, const ArgSource&, const ArgWord&, const Span<const ArgVal>&);
void emit_i_get_tuple_element(const ArgSource&, const ArgWord&, const ArgRegister&);
void emit_i_hibernate();
void emit_i_int_div(const ArgLabel&, const ArgSource&, const ArgSource&, const ArgRegister&);
void emit_i_is_tagged_tuple(const ArgLabel&, const ArgSource&, const ArgWord&, const ArgAtom&);
void emit_i_is_tagged_tuple_ff(const ArgLabel&, const ArgLabel&, const ArgSource&, const ArgWord&, const ArgAtom&);
void emit_i_is_tuple(const ArgLabel&, const ArgSource&);
void emit_i_is_tuple_of_arity(const ArgLabel&, const ArgSource&, const ArgWord&);
void emit_i_is_tuple_of_arity_ff(const ArgLabel&, const ArgLabel&, const ArgSource&, const ArgWord&);
void emit_i_jump_on_val(const ArgSource&, const ArgVal&, const ArgWord&, const ArgWord&, const Span<const ArgVal>&);
void emit_i_lambda_error(const ArgWord&);
void emit_i_lambda_trampoline(const ArgLambda&, const ArgLabel&, const ArgWord&, const ArgWord&);
void emit_i_length(const ArgLabel&, const ArgWord&, const ArgRegister&);
void emit_i_length_setup(const ArgLabel&, const ArgWord&, const ArgSource&);
void emit_i_line_breakpoint_trampoline();
void emit_i_load_nif();
void emit_i_loop_rec(const ArgLabel&);
void emit_i_m_div(const ArgLabel&, const ArgSource&, const ArgSource&, const ArgRegister&);
void emit_i_make_fun3(const ArgLambda&, const ArgRegister&, const ArgWord&, const ArgWord&, const Span<const ArgVal>&);
void emit_i_minus(const ArgSource&, const ArgSource&, const ArgLabel&, const ArgRegister&);
void emit_i_move(const ArgSource&, const ArgRegister&);
void emit_i_mul_add(const ArgLabel&, const ArgSource&, const ArgSource&, const ArgSource&, const ArgSource&, const ArgRegister&);
void emit_i_new_small_map_lit(const ArgRegister&, const ArgWord&, const ArgLiteral&, const ArgWord&, const Span<const ArgVal>&);
void emit_i_nif_padding();
void emit_i_perf_counter();
void emit_i_plus(const ArgSource&, const ArgSource&, const ArgLabel&, const ArgRegister&);
void emit_i_rem(const ArgSource&, const ArgSource&, const ArgLabel&, const ArgRegister&);
void emit_i_rem_div(const ArgSource&, const ArgSource&, const ArgLabel&, const ArgRegister&, const ArgRegister&);
void emit_i_return_to_trace();
void emit_i_select_tuple_arity(const ArgRegister&, const ArgLabel&, const ArgWord&, const Span<const ArgVal>&);
void emit_i_select_val_bins(const ArgSource&, const ArgVal&, const ArgWord&, const Span<const ArgVal>&);
void emit_i_select_val_lins(const ArgSource&, const ArgVal&, const ArgWord&, const Span<const ArgVal>&);
void emit_i_test_arity(const ArgLabel&, const ArgSource&, const ArgWord&);
void emit_i_test_yield();
void emit_i_trim(const ArgWord&);
void emit_i_unary_minus(const ArgSource&, const ArgLabel&, const ArgRegister&);
void emit_i_validate(const ArgWord&);
void emit_i_yield();
void emit_if_end();
void emit_init_yregs(const ArgWord&, const Span<const ArgVal>&);
void emit_int_code_end();
void emit_is_atom(const ArgLabel&, const ArgSource&);
void emit_is_binary(const ArgLabel&, const ArgSource&);
void emit_is_bitstring(const ArgLabel&, const ArgSource&);
void emit_is_boolean(const ArgLabel&, const ArgSource&);
void emit_is_eq(const ArgLabel&, const ArgSource&, const ArgSource&);
void emit_is_eq_exact(const ArgLabel&, const ArgSource&, const ArgSource&);
void emit_is_float(const ArgLabel&, const ArgSource&);
void emit_is_function(const ArgLabel&, const ArgRegister&);
void emit_is_function2(const ArgLabel&, const ArgSource&, const ArgSource&);
void emit_is_ge(const ArgLabel&, const ArgSource&, const ArgSource&);
void emit_is_ge_ge(const ArgLabel&, const ArgLabel&, const ArgRegister&, const ArgConstant&, const ArgConstant&);
void emit_is_ge_lt(const ArgLabel&, const ArgLabel&, const ArgRegister&, const ArgConstant&, const ArgConstant&);
void emit_is_in_range(const ArgLabel&, const ArgLabel&, const ArgRegister&, const ArgConstant&, const ArgConstant&);
void emit_is_int_ge(const ArgLabel&, const ArgRegister&, const ArgConstant&);
void emit_is_int_in_range(const ArgLabel&, const ArgRegister&, const ArgConstant&, const ArgConstant&);
void emit_is_integer(const ArgLabel&, const ArgSource&);
void emit_is_list(const ArgLabel&, const ArgSource&);
void emit_is_lt(const ArgLabel&, const ArgSource&, const ArgSource&);
void emit_is_map(const ArgLabel&, const ArgSource&);
void emit_is_ne(const ArgLabel&, const ArgSource&, const ArgSource&);
void emit_is_ne_exact(const ArgLabel&, const ArgSource&, const ArgSource&);
void emit_is_nil(const ArgLabel&, const ArgRegister&);
void emit_is_nonempty_list(const ArgLabel&, const ArgRegister&);
void emit_is_nonempty_list_get_hd(const ArgLabel&, const ArgRegister&, const ArgRegister&);
void emit_is_nonempty_list_get_list(const ArgLabel&, const ArgRegister&, const ArgRegister&, const ArgRegister&);
void emit_is_nonempty_list_get_tl(const ArgLabel&, const ArgRegister&, const ArgRegister&);
void emit_is_number(const ArgLabel&, const ArgSource&);
void emit_is_pid(const ArgLabel&, const ArgSource&);
void emit_is_port(const ArgLabel&, const ArgSource&);
void emit_is_reference(const ArgLabel&, const ArgSource&);
void emit_jump(const ArgLabel&);
void emit_label(const ArgLabel&);
void emit_line(const ArgWord&);
void emit_load_tuple_ptr(const ArgSource&);
void emit_loop_rec_end(const ArgLabel&);
void emit_move_two_words(const ArgSource&, const ArgRegister&, const ArgSource&, const ArgRegister&);
void emit_new_map(const ArgRegister&, const ArgWord&, const ArgWord&, const Span<const ArgVal>&);
void emit_nif_start();
void emit_node(const ArgRegister&);
void emit_nofail_bif1(const ArgSource&, const ArgWord&, const ArgRegister&);
void emit_nofail_bif2(const ArgSource&, const ArgSource&, const ArgWord&, const ArgRegister&);
void emit_normal_exit();
void emit_on_load();
void emit_put_cons(const ArgSource&, const ArgSource&);
void emit_put_tuple2(const ArgRegister&, const ArgWord&, const Span<const ArgVal>&);
void emit_raise(const ArgSource&, const ArgSource&);
void emit_raw_raise();
void emit_recv_marker_bind(const ArgRegister&, const ArgRegister&);
void emit_recv_marker_clear(const ArgRegister&);
void emit_recv_marker_reserve(const ArgRegister&);
void emit_recv_marker_use(const ArgRegister&);
void emit_remove_message();
void emit_return();
void emit_return_trace();
void emit_self(const ArgRegister&);
void emit_send();
void emit_set_tuple_element(const ArgSource&, const ArgRegister&, const ArgWord&);
void emit_store_cons(const ArgWord&, const ArgRegister&);
void emit_swap(const ArgRegister&, const ArgRegister&);
void emit_system_limit_body();
void emit_test_heap(const ArgWord&, const ArgWord&);
void emit_timeout();
void emit_timeout_locked();
void emit_trace_jump(const ArgWord&);
void emit_try_case(const ArgYRegister&);
void emit_try_case_end(const ArgSource&);
void emit_try_end(const ArgYRegister&);
void emit_try_end_deallocate(const ArgWord&);
void emit_update_map_assoc(const ArgSource&, const ArgRegister&, const ArgWord&, const ArgWord&, const Span<const ArgVal>&);
void emit_update_map_exact(const ArgSource&, const ArgLabel&, const ArgRegister&, const ArgWord&, const ArgWord&, const Span<const ArgVal>&);
void emit_update_record(const ArgAtom&, const ArgWord&, const ArgSource&, const ArgRegister&, const ArgWord&, const Span<const ArgVal>&);
void emit_update_record_in_place(const ArgWord&, const ArgSource&, const ArgRegister&, const ArgWord&, const Span<const ArgVal>&);
void emit_wait_locked(const ArgLabel&);
void emit_wait_timeout_locked(const ArgSource&, const ArgLabel&);
void emit_wait_timeout_unlocked(const ArgSource&, const ArgLabel&);
void emit_wait_unlocked(const ArgLabel&);
