/*
 * %CopyrightBegin%
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Copyright Ericsson 2021. All Rights Reserved.
 * Copyright Ericsson AB 2021-2025. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * %CopyrightEnd%
 */

#ifndef PRIM_FILE_NIF_DYNCALL_H
#define PRIM_FILE_NIF_DYNCALL_H

#include "erl_nif.h"

enum prim_file_nif_dyncall_op {
    prim_file_nif_dyncall_dup,
};

/* Common operation header */
struct prim_file_nif_dyncall {
    enum prim_file_nif_dyncall_op op;
    int result; /* 0 for success */
};

/* Return a dup:ed file descriptor as binary data in 'handle' */
struct prim_file_nif_dyncall_dup {
    /* struct prim_file_nif_dyncall */
    enum prim_file_nif_dyncall_op op;
    int result; /* 0 for success, errno for failure */
    ErlNifEvent handle;
};


#endif /* #ifdef PRIM_FILE_NIF_DYNCALL_H */
