/*
 *  Warning: Do *not* edit this file. It was automatically
 *  generated by 'make_tables'.
 */

#ifndef __ERL_ATOM_TABLE_H__
#define __ERL_ATOM_TABLE_H__
extern char* erl_atom_names[];

#define am_false make_atom(0)
#define am_true make_atom(1)
#define am_Underscore make_atom(2)
#define am_Noname make_atom(3)
#define am_EOT make_atom(4)
#define am_Empty make_atom(5)
#define am_infinity make_atom(6)
#define am_timeout make_atom(7)
#define am_normal make_atom(8)
#define am_call make_atom(9)
#define am_return make_atom(10)
#define am_throw make_atom(11)
#define am_error make_atom(12)
#define am_exit make_atom(13)
#define am_undefined make_atom(14)
#define am_nocatch make_atom(15)
#define am_undefined_function make_atom(16)
#define am_undefined_lambda make_atom(17)
#define am_nil make_atom(18)
#define am_no make_atom(19)
#define am_none make_atom(20)
#define am_DOWN make_atom(21)
#define am_UP make_atom(22)
#define am_EXIT make_atom(23)
#define am_abandoned make_atom(24)
#define am_abort make_atom(25)
#define am_abs_path make_atom(26)
#define am_absoluteURI make_atom(27)
#define am_ac make_atom(28)
#define am_access make_atom(29)
#define am_active make_atom(30)
#define am_active_tasks make_atom(31)
#define am_active_tasks_all make_atom(32)
#define am_alias make_atom(33)
#define am_alive make_atom(34)
#define am_all make_atom(35)
#define am_all_but_first make_atom(36)
#define am_all_names make_atom(37)
#define am_alloc_info make_atom(38)
#define am_alloc_sizes make_atom(39)
#define am_allocated make_atom(40)
#define am_allocated_areas make_atom(41)
#define am_allocator make_atom(42)
#define am_allocator_sizes make_atom(43)
#define am_alloc_util_allocators make_atom(44)
#define am_allow_passive_connect make_atom(45)
#define am_already_exists make_atom(46)
#define am_already_loaded make_atom(47)
#define am_amd64 make_atom(48)
#define am_anchored make_atom(49)
#define am_and make_atom(50)
#define am_andalso make_atom(51)
#define am_andthen make_atom(52)
#define am_any make_atom(53)
#define am_anycrlf make_atom(54)
#define am_apply make_atom(55)
#define am_args make_atom(56)
#define am_arg0 make_atom(57)
#define am_arity make_atom(58)
#define am_asn1 make_atom(59)
#define am_async make_atom(60)
#define am_async_dist make_atom(61)
#define am_asynchronous make_atom(62)
#define am_atom make_atom(63)
#define am_atom_used make_atom(64)
#define am_attributes make_atom(65)
#define am_auto make_atom(66)
#define am_auto_connect make_atom(67)
#define am_await_exit make_atom(68)
#define am_await_microstate_accounting_modifications make_atom(69)
#define am_await_port_send_result make_atom(70)
#define am_await_proc_exit make_atom(71)
#define am_await_result make_atom(72)
#define am_await_sched_wall_time_modifications make_atom(73)
#define am_awaiting_load make_atom(74)
#define am_awaiting_unload make_atom(75)
#define am_backtrace make_atom(76)
#define am_backtrace_depth make_atom(77)
#define am_badarg make_atom(78)
#define am_badarith make_atom(79)
#define am_badarity make_atom(80)
#define am_badfile make_atom(81)
#define am_badfun make_atom(82)
#define am_badkey make_atom(83)
#define am_badmap make_atom(84)
#define am_badmatch make_atom(85)
#define am_badrecord make_atom(86)
#define am_badsig make_atom(87)
#define am_badopt make_atom(88)
#define am_badtype make_atom(89)
#define am_bad_map_iterator make_atom(90)
#define am_bag make_atom(91)
#define am_band make_atom(92)
#define am_big make_atom(93)
#define am_bif_handle_signals_return make_atom(94)
#define am_bif_return_trap make_atom(95)
#define am_binary make_atom(96)
#define am_binary_copy_trap make_atom(97)
#define am_binary_find_trap make_atom(98)
#define am_binary_longest_prefix_trap make_atom(99)
#define am_binary_longest_suffix_trap make_atom(100)
#define am_binary_to_list_continue make_atom(101)
#define am_binary_to_term_trap make_atom(102)
#define am_block make_atom(103)
#define am_block_normal make_atom(104)
#define am_blocked make_atom(105)
#define am_blocked_normal make_atom(106)
#define am_bm make_atom(107)
#define am_bnot make_atom(108)
#define am_bor make_atom(109)
#define am_bxor make_atom(110)
#define am_break_ignored make_atom(111)
#define am_breakpoint make_atom(112)
#define am_bsl make_atom(113)
#define am_bsr make_atom(114)
#define am_bsr_anycrlf make_atom(115)
#define am_bsr_unicode make_atom(116)
#define am_build_flavor make_atom(117)
#define am_build_type make_atom(118)
#define am_busy make_atom(119)
#define am_busy_dist_port make_atom(120)
#define am_busy_limits_port make_atom(121)
#define am_busy_limits_msgq make_atom(122)
#define am_busy_port make_atom(123)
#define am_call_count make_atom(124)
#define am_call_error_handler make_atom(125)
#define am_call_memory make_atom(126)
#define am_call_time make_atom(127)
#define am_call_trace_return make_atom(128)
#define am_caller make_atom(129)
#define am_caller_line make_atom(130)
#define am_capture make_atom(131)
#define am_case_clause make_atom(132)
#define am_caseless make_atom(133)
#define am_catch make_atom(134)
#define am_catchlevel make_atom(135)
#define am_cause make_atom(136)
#define am_cd make_atom(137)
#define am_cdr make_atom(138)
#define am_cflags make_atom(139)
#define am_CHANGE make_atom(140)
#define am_characters_to_binary_int make_atom(141)
#define am_characters_to_list_int make_atom(142)
#define am_check_gc make_atom(143)
#define am_clear make_atom(144)
#define am_clock_service make_atom(145)
#define am_close make_atom(146)
#define am_closed make_atom(147)
#define am_code make_atom(148)
#define am_command make_atom(149)
#define am_commandv make_atom(150)
#define am_compact make_atom(151)
#define am_compat_rel make_atom(152)
#define am_compile make_atom(153)
#define am_complete make_atom(154)
#define am_compressed make_atom(155)
#define am_config_h make_atom(156)
#define am_convert_time_unit make_atom(157)
#define am_connect make_atom(158)
#define am_connected make_atom(159)
#define am_connection_closed make_atom(160)
#define am_connection_id make_atom(161)
#define am_const make_atom(162)
#define am_context_switches make_atom(163)
#define am_continue_exit make_atom(164)
#define am_control make_atom(165)
#define am_copy make_atom(166)
#define am_copy_literals make_atom(167)
#define am_counters make_atom(168)
#define am_count make_atom(169)
#define am_cpu make_atom(170)
#define am_cpu_timestamp make_atom(171)
#define am_cr make_atom(172)
#define am_crlf make_atom(173)
#define am_creation make_atom(174)
#define am_current_function make_atom(175)
#define am_current_location make_atom(176)
#define am_current_stacktrace make_atom(177)
#define am_data make_atom(178)
#define am_debug_flags make_atom(179)
#define am_debugger_event make_atom(180)
#define am_decentralized_counters make_atom(181)
#define am_decimals make_atom(182)
#define am_default make_atom(183)
#define am_debug_hash_fixed_number_of_locks make_atom(184)
#define am_delay_trap make_atom(185)
#define am_demonitor make_atom(186)
#define am_deterministic make_atom(187)
#define am_dictionary make_atom(188)
#define am_dirty_bif_exception make_atom(189)
#define am_dirty_bif_result make_atom(190)
#define am_dirty_bif_trap make_atom(191)
#define am_dirty_cpu make_atom(192)
#define am_dirty_cpu_schedulers_online make_atom(193)
#define am_dirty_execution make_atom(194)
#define am_dirty_io make_atom(195)
#define am_dirty_nif_exception make_atom(196)
#define am_dirty_nif_finalizer make_atom(197)
#define am_disable_trace make_atom(198)
#define am_disabled make_atom(199)
#define am_discard make_atom(200)
#define am_dist make_atom(201)
#define am_dist_cmd make_atom(202)
#define am_dist_ctrl_put_data make_atom(203)
#define am_dist_ctrlr make_atom(204)
#define am_dist_data make_atom(205)
#define am_dist_spawn_init make_atom(206)
#define am_Div make_atom(207)
#define am_div make_atom(208)
#define am_dmonitor_node make_atom(209)
#define am_DollarDollar make_atom(210)
#define am_DollarUnderscore make_atom(211)
#define am_dollar_endonly make_atom(212)
#define am_dotall make_atom(213)
#define am_driver make_atom(214)
#define am_driver_options make_atom(215)
#define am_dsend_continue_trap make_atom(216)
#define am_duplicate_bag make_atom(217)
#define am_duplicated make_atom(218)
#define am_dupnames make_atom(219)
#define am_dynamic_node_name make_atom(220)
#define am_einval make_atom(221)
#define am_emu_flavor make_atom(222)
#define am_emu_type make_atom(223)
#define am_emulator make_atom(224)
#define am_enable_trace make_atom(225)
#define am_enabled make_atom(226)
#define am_endian make_atom(227)
#define am_entry make_atom(228)
#define am_env make_atom(229)
#define am_ensure_at_least make_atom(230)
#define am_ensure_exactly make_atom(231)
#define am_eof make_atom(232)
#define am_eol make_atom(233)
#define am_Eq make_atom(234)
#define am_Eqeq make_atom(235)
#define am_erl_erts_errors make_atom(236)
#define am_erl_init make_atom(237)
#define am_erl_kernel_errors make_atom(238)
#define am_erl_stdlib_errors make_atom(239)
#define am_erl_tracer make_atom(240)
#define am_erlang make_atom(241)
#define am_erl_signal_server make_atom(242)
#define am_error_handler make_atom(243)
#define am_error_info make_atom(244)
#define am_error_logger make_atom(245)
#define am_error_only make_atom(246)
#define am_erts_code_purger make_atom(247)
#define am_erts_debug make_atom(248)
#define am_erts_dflags make_atom(249)
#define am_erts_internal make_atom(250)
#define am_ets make_atom(251)
#define am_ets_info_binary make_atom(252)
#define am_ETS_TRANSFER make_atom(253)
#define am_exact_reductions make_atom(254)
#define am_exception_from make_atom(255)
#define am_exception_trace make_atom(256)
#define am_exclusive make_atom(257)
#define am_exit_status make_atom(258)
#define am_exited make_atom(259)
#define am_existing make_atom(260)
#define am_existing_processes make_atom(261)
#define am_existing_ports make_atom(262)
#define am_exiting make_atom(263)
#define am_export make_atom(264)
#define am_exports make_atom(265)
#define am_extended make_atom(266)
#define am_external make_atom(267)
#define am_extra make_atom(268)
#define am_fcgi make_atom(269)
#define am_fd make_atom(270)
#define am_first make_atom(271)
#define am_firstline make_atom(272)
#define am_flags make_atom(273)
#define am_flush make_atom(274)
#define am_flush_monitor_messages make_atom(275)
#define am_flush_timeout make_atom(276)
#define am_force make_atom(277)
#define am_format_bs_fail make_atom(278)
#define am_format_cpu_topology make_atom(279)
#define am_free make_atom(280)
#define am_fullsweep_after make_atom(281)
#define am_function make_atom(282)
#define am_function_counters make_atom(283)
#define am_functions make_atom(284)
#define am_function_clause make_atom(285)
#define am_garbage_collect make_atom(286)
#define am_garbage_collecting make_atom(287)
#define am_garbage_collection make_atom(288)
#define am_garbage_collection_info make_atom(289)
#define am_gc_major_end make_atom(290)
#define am_gc_major_start make_atom(291)
#define am_gc_max_heap_size make_atom(292)
#define am_gc_minor_end make_atom(293)
#define am_gc_minor_start make_atom(294)
#define am_Ge make_atom(295)
#define am_generational make_atom(296)
#define am_get_all_trap make_atom(297)
#define am_get_internal_state_blocked make_atom(298)
#define am_get_seq_token make_atom(299)
#define am_get_size make_atom(300)
#define am_get_tail make_atom(301)
#define am_get_tcw make_atom(302)
#define am_gather_gc_info_result make_atom(303)
#define am_gather_io_bytes make_atom(304)
#define am_gather_microstate_accounting_result make_atom(305)
#define am_gather_sched_wall_time_result make_atom(306)
#define am_gather_system_check_result make_atom(307)
#define am_getting_linked make_atom(308)
#define am_getting_unlinked make_atom(309)
#define am_global make_atom(310)
#define am_Gt make_atom(311)
#define am_grun make_atom(312)
#define am_group_leader make_atom(313)
#define am_handle make_atom(314)
#define am_have_dt_utag make_atom(315)
#define am_heap_block_size make_atom(316)
#define am_heap_size make_atom(317)
#define am_heap_sizes make_atom(318)
#define am_heap_type make_atom(319)
#define am_heart_port make_atom(320)
#define am_heir make_atom(321)
#define am_hidden make_atom(322)
#define am_hide make_atom(323)
#define am_high make_atom(324)
#define am_http make_atom(325)
#define am_httph make_atom(326)
#define am_https make_atom(327)
#define am_http_response make_atom(328)
#define am_http_request make_atom(329)
#define am_http_header make_atom(330)
#define am_http_eoh make_atom(331)
#define am_http_error make_atom(332)
#define am_http_bin make_atom(333)
#define am_httph_bin make_atom(334)
#define am_id make_atom(335)
#define am_if_clause make_atom(336)
#define am_ignore make_atom(337)
#define am_in make_atom(338)
#define am_in_exiting make_atom(339)
#define am_inactive make_atom(340)
#define am_include_shared_binaries make_atom(341)
#define am_incomplete make_atom(342)
#define am_inconsistent make_atom(343)
#define am_index make_atom(344)
#define am_info make_atom(345)
#define am_info_msg make_atom(346)
#define am_info_trap make_atom(347)
#define am_inherit make_atom(348)
#define am_init make_atom(349)
#define am_initial_call make_atom(350)
#define am_inplace make_atom(351)
#define am_input make_atom(352)
#define am_integer make_atom(353)
#define am_internal make_atom(354)
#define am_internal_error make_atom(355)
#define am_instruction_counts make_atom(356)
#define am_invalid make_atom(357)
#define am_is_constant make_atom(358)
#define am_is_seq_trace make_atom(359)
#define am_iterator make_atom(360)
#define am_io make_atom(361)
#define am_iolist_size_continue make_atom(362)
#define am_iolist_to_iovec_continue make_atom(363)
#define am_iodata make_atom(364)
#define am_iovec make_atom(365)
#define am_keypos make_atom(366)
#define am_kill make_atom(367)
#define am_killed make_atom(368)
#define am_kill_ports make_atom(369)
#define am_known make_atom(370)
#define am_label make_atom(371)
#define am_large_heap make_atom(372)
#define am_last_calls make_atom(373)
#define am_latin1 make_atom(374)
#define am_ldflags make_atom(375)
#define am_Le make_atom(376)
#define am_legacy make_atom(377)
#define am_lf make_atom(378)
#define am_line make_atom(379)
#define am_line_breakpoint make_atom(380)
#define am_line_counters make_atom(381)
#define am_line_delimiter make_atom(382)
#define am_line_length make_atom(383)
#define am_linked_in_driver make_atom(384)
#define am_links make_atom(385)
#define am_list make_atom(386)
#define am_list_to_bitstring_continue make_atom(387)
#define am_little make_atom(388)
#define am_loaded make_atom(389)
#define am_load_cancelled make_atom(390)
#define am_load_failure make_atom(391)
#define am_local make_atom(392)
#define am_logger make_atom(393)
#define am_long_gc make_atom(394)
#define am_long_message_queue make_atom(395)
#define am_long_schedule make_atom(396)
#define am_low make_atom(397)
#define am_Lt make_atom(398)
#define am_machine make_atom(399)
#define am_magic_ref make_atom(400)
#define am_major make_atom(401)
#define am_match make_atom(402)
#define am_match_limit make_atom(403)
#define am_match_limit_recursion make_atom(404)
#define am_match_spec make_atom(405)
#define am_match_spec_result make_atom(406)
#define am_max make_atom(407)
#define am_maximum make_atom(408)
#define am_max_heap_size make_atom(409)
#define am_mbuf_size make_atom(410)
#define am_md5 make_atom(411)
#define am_memory make_atom(412)
#define am_memory_internal make_atom(413)
#define am_memory_types make_atom(414)
#define am_message make_atom(415)
#define am_message_queue_data make_atom(416)
#define am_message_queue_len make_atom(417)
#define am_messages make_atom(418)
#define am_merge_trap make_atom(419)
#define am_meta make_atom(420)
#define am_meta_match_spec make_atom(421)
#define am_micro_seconds make_atom(422)
#define am_microsecond make_atom(423)
#define am_microstate_accounting make_atom(424)
#define am_milli_seconds make_atom(425)
#define am_millisecond make_atom(426)
#define am_min make_atom(427)
#define am_min_heap_size make_atom(428)
#define am_min_bin_vheap_size make_atom(429)
#define am_minor make_atom(430)
#define am_minor_version make_atom(431)
#define am_Minus make_atom(432)
#define am_MinusMinus make_atom(433)
#define am_module make_atom(434)
#define am_module_info make_atom(435)
#define am_monitored_by make_atom(436)
#define am_monitor make_atom(437)
#define am_monitor_nodes make_atom(438)
#define am_monitors make_atom(439)
#define am_monotonic make_atom(440)
#define am_monotonic_timestamp make_atom(441)
#define am_more make_atom(442)
#define am_multi_scheduling make_atom(443)
#define am_multiline make_atom(444)
#define am_nano_seconds make_atom(445)
#define am_nanosecond make_atom(446)
#define am_name make_atom(447)
#define am_named_table make_atom(448)
#define am_namelist make_atom(449)
#define am_native make_atom(450)
#define am_native_addresses make_atom(451)
#define am_need_gc make_atom(452)
#define am_Neq make_atom(453)
#define am_Neqeq make_atom(454)
#define am_net_kernel make_atom(455)
#define am_net_kernel_terminated make_atom(456)
#define am_never_utf make_atom(457)
#define am_new make_atom(458)
#define am_new_index make_atom(459)
#define am_new_processes make_atom(460)
#define am_new_ports make_atom(461)
#define am_new_uniq make_atom(462)
#define am_newline make_atom(463)
#define am_nifs make_atom(464)
#define am_nomatch make_atom(465)
#define am_no_auto_capture make_atom(466)
#define am_noconnect make_atom(467)
#define am_noconnection make_atom(468)
#define am_node make_atom(469)
#define am_node_type make_atom(470)
#define am_nodedown make_atom(471)
#define am_nodedown_reason make_atom(472)
#define am_nodeup make_atom(473)
#define am_noeol make_atom(474)
#define am_noproc make_atom(475)
#define am_normal_exit make_atom(476)
#define am_nosuspend make_atom(477)
#define am_no_fail make_atom(478)
#define am_no_float make_atom(479)
#define am_no_integer make_atom(480)
#define am_no_network make_atom(481)
#define am_no_start_optimize make_atom(482)
#define am_not_suspended make_atom(483)
#define am_not make_atom(484)
#define am_not_a_list make_atom(485)
#define am_not_loaded make_atom(486)
#define am_not_loaded_by_this_process make_atom(487)
#define am_not_pending make_atom(488)
#define am_not_owner make_atom(489)
#define am_not_purged make_atom(490)
#define am_notalive make_atom(491)
#define am_notbol make_atom(492)
#define am_noteol make_atom(493)
#define am_notempty make_atom(494)
#define am_notempty_atstart make_atom(495)
#define am_notify make_atom(496)
#define am_notsup make_atom(497)
#define am_nouse_stdio make_atom(498)
#define am_nul make_atom(499)
#define am_off_heap make_atom(500)
#define am_offset make_atom(501)
#define am_ok make_atom(502)
#define am_old_heap_block_size make_atom(503)
#define am_old_heap_size make_atom(504)
#define am_on_heap make_atom(505)
#define am_on_load make_atom(506)
#define am_open make_atom(507)
#define am_open_error make_atom(508)
#define am_opt make_atom(509)
#define am_or make_atom(510)
#define am_ordered_set make_atom(511)
#define am_orelse make_atom(512)
#define am_os_pid make_atom(513)
#define am_os_type make_atom(514)
#define am_os_version make_atom(515)
#define am_out make_atom(516)
#define am_out_exited make_atom(517)
#define am_out_exiting make_atom(518)
#define am_output make_atom(519)
#define am_outstanding_system_requests_limit make_atom(520)
#define am_overlapped_io make_atom(521)
#define am_owner make_atom(522)
#define am_packet make_atom(523)
#define am_packet_size make_atom(524)
#define am_parallelism make_atom(525)
#define am_parent make_atom(526)
#define am_Plus make_atom(527)
#define am_PlusPlus make_atom(528)
#define am_pause make_atom(529)
#define am_pending make_atom(530)
#define am_pending_driver make_atom(531)
#define am_pending_process make_atom(532)
#define am_pending_purge_lambda make_atom(533)
#define am_pending_reload make_atom(534)
#define am_permanent make_atom(535)
#define am_pid make_atom(536)
#define am_port make_atom(537)
#define am_ports make_atom(538)
#define am_port_count make_atom(539)
#define am_port_limit make_atom(540)
#define am_port_op make_atom(541)
#define am_positive make_atom(542)
#define am_position make_atom(543)
#define am_prepare make_atom(544)
#define am_prepare_on_load make_atom(545)
#define am_print make_atom(546)
#define am_priority make_atom(547)
#define am_priority_messages make_atom(548)
#define am_private make_atom(549)
#define am_private_append make_atom(550)
#define am_process make_atom(551)
#define am_processes make_atom(552)
#define am_processes_used make_atom(553)
#define am_process_count make_atom(554)
#define am_process_display make_atom(555)
#define am_DollarProcessLabel make_atom(556)
#define am_process_limit make_atom(557)
#define am_process_dump make_atom(558)
#define am_procs make_atom(559)
#define am_proc_sig make_atom(560)
#define am_profile make_atom(561)
#define am_protected make_atom(562)
#define am_protection make_atom(563)
#define am_ptab_list_continue make_atom(564)
#define am_public make_atom(565)
#define am_put_common_trap make_atom(566)
#define am_queue_size make_atom(567)
#define am_raw make_atom(568)
#define am_re make_atom(569)
#define am_re_exported_pattern make_atom(570)
#define am_re_pattern make_atom(571)
#define am_re_run_trap make_atom(572)
#define am_read_concurrency make_atom(573)
#define am_ready_error make_atom(574)
#define am_ready_input make_atom(575)
#define am_ready_output make_atom(576)
#define am_reason make_atom(577)
#define am_receive make_atom(578)
#define am_recent_size make_atom(579)
#define am_reductions make_atom(580)
#define am_refc make_atom(581)
#define am_register make_atom(582)
#define am_registered_name make_atom(583)
#define am_reload make_atom(584)
#define am_rem make_atom(585)
#define am_reply make_atom(586)
#define am_reply_demonitor make_atom(587)
#define am_reply_tag make_atom(588)
#define am_report_errors make_atom(589)
#define am_reset make_atom(590)
#define am_reset_seq_trace make_atom(591)
#define am_restart make_atom(592)
#define am_resume make_atom(593)
#define am_return_from make_atom(594)
#define am_return_to make_atom(595)
#define am_return_to_trace make_atom(596)
#define am_return_trace make_atom(597)
#define am_reuse make_atom(598)
#define am_run_process make_atom(599)
#define am_run_queue make_atom(600)
#define am_run_queue_lengths make_atom(601)
#define am_run_queue_lengths_all make_atom(602)
#define am_runnable make_atom(603)
#define am_runnable_ports make_atom(604)
#define am_runnable_procs make_atom(605)
#define am_running make_atom(606)
#define am_running_ports make_atom(607)
#define am_running_procs make_atom(608)
#define am_runtime make_atom(609)
#define am_safe make_atom(610)
#define am_save_calls make_atom(611)
#define am_sbct make_atom(612)
#define am_scheduler make_atom(613)
#define am_scheduler_id make_atom(614)
#define am_scheduler_wall_time make_atom(615)
#define am_scheduler_wall_time_all make_atom(616)
#define am_schedulers_online make_atom(617)
#define am_scheme make_atom(618)
#define am_scientific make_atom(619)
#define am_scope make_atom(620)
#define am_second make_atom(621)
#define am_seconds make_atom(622)
#define am_send make_atom(623)
#define am_send_to_non_existing_process make_atom(624)
#define am_sensitive make_atom(625)
#define am_sequential_tracer make_atom(626)
#define am_sequential_trace_token make_atom(627)
#define am_serial make_atom(628)
#define am_session make_atom(629)
#define am_set make_atom(630)
#define am_set_cpu_topology make_atom(631)
#define am_set_on_first_link make_atom(632)
#define am_set_on_first_spawn make_atom(633)
#define am_set_on_link make_atom(634)
#define am_set_on_spawn make_atom(635)
#define am_set_seq_token make_atom(636)
#define am_set_tcw make_atom(637)
#define am_set_tcw_fake make_atom(638)
#define am_short make_atom(639)
#define am_shutdown make_atom(640)
#define am_sighup make_atom(641)
#define am_signed make_atom(642)
#define am_sigterm make_atom(643)
#define am_sigusr1 make_atom(644)
#define am_sigusr2 make_atom(645)
#define am_sigill make_atom(646)
#define am_sigchld make_atom(647)
#define am_sigabrt make_atom(648)
#define am_sigalrm make_atom(649)
#define am_sigstop make_atom(650)
#define am_sigint make_atom(651)
#define am_sigsegv make_atom(652)
#define am_sigtstp make_atom(653)
#define am_sigcont make_atom(654)
#define am_sigquit make_atom(655)
#define am_sigwinch make_atom(656)
#define am_siginfo make_atom(657)
#define am_silent make_atom(658)
#define am_size make_atom(659)
#define am_skip make_atom(660)
#define am_slots make_atom(661)
#define am_spawn_executable make_atom(662)
#define am_spawn_driver make_atom(663)
#define am_spawn_init make_atom(664)
#define am_spawn_reply make_atom(665)
#define am_spawn_request make_atom(666)
#define am_spawn_request_yield make_atom(667)
#define am_spawn_service make_atom(668)
#define am_spawned make_atom(669)
#define am_ssl_tls make_atom(670)
#define am_stack_size make_atom(671)
#define am_start make_atom(672)
#define am_status make_atom(673)
#define am_stderr_to_stdout make_atom(674)
#define am_stop make_atom(675)
#define am_stream make_atom(676)
#define am_strict_monotonic make_atom(677)
#define am_strict_monotonic_timestamp make_atom(678)
#define am_success_only make_atom(679)
#define am_sunrm make_atom(680)
#define am_suspend make_atom(681)
#define am_suspended make_atom(682)
#define am_suspending make_atom(683)
#define am_system make_atom(684)
#define am_system_flag_scheduler_wall_time make_atom(685)
#define am_system_limit make_atom(686)
#define am_system_version make_atom(687)
#define am_system_architecture make_atom(688)
#define am_table make_atom(689)
#define am_table_type make_atom(690)
#define am_tag make_atom(691)
#define am_term_to_binary_trap make_atom(692)
#define am_this make_atom(693)
#define am_thread_pool_size make_atom(694)
#define am_threads make_atom(695)
#define am_time_offset make_atom(696)
#define am_timeout_value make_atom(697)
#define am_Times make_atom(698)
#define am_timestamp make_atom(699)
#define am_total make_atom(700)
#define am_total_active_tasks make_atom(701)
#define am_total_active_tasks_all make_atom(702)
#define am_total_heap_size make_atom(703)
#define am_total_run_queue_lengths make_atom(704)
#define am_total_run_queue_lengths_all make_atom(705)
#define am_tpkt make_atom(706)
#define am_trace make_atom(707)
#define am_traced make_atom(708)
#define am_trace_control_word make_atom(709)
#define am_trace_status make_atom(710)
#define am_tracer make_atom(711)
#define am_trap_exit make_atom(712)
#define am_trim make_atom(713)
#define am_trim_all make_atom(714)
#define am_try_clause make_atom(715)
#define am_type make_atom(716)
#define am_ucompile make_atom(717)
#define am_ucp make_atom(718)
#define am_explicit_unalias make_atom(719)
#define am_undef make_atom(720)
#define am_ungreedy make_atom(721)
#define am_unicode make_atom(722)
#define am_unregister make_atom(723)
#define am_urun make_atom(724)
#define am_use_stdio make_atom(725)
#define am_used make_atom(726)
#define am_utf8 make_atom(727)
#define am_utf16 make_atom(728)
#define am_utf32 make_atom(729)
#define am_unblock make_atom(730)
#define am_unblock_normal make_atom(731)
#define am_uniq make_atom(732)
#define am_unit make_atom(733)
#define am_unless_suspending make_atom(734)
#define am_unloaded make_atom(735)
#define am_unloaded_only make_atom(736)
#define am_unload_cancelled make_atom(737)
#define am_unsafe make_atom(738)
#define am_unsupported make_atom(739)
#define am_value make_atom(740)
#define am_version make_atom(741)
#define am_visible make_atom(742)
#define am_wait_release_literal_area_switch make_atom(743)
#define am_waiting make_atom(744)
#define am_wall_clock make_atom(745)
#define am_warning make_atom(746)
#define am_warning_msg make_atom(747)
#define am_wordsize make_atom(748)
#define am_write_concurrency make_atom(749)
#define am_x make_atom(750)
#define am_xor make_atom(751)
#define am_x86 make_atom(752)
#define am_y make_atom(753)
#define am_yes make_atom(754)
#define am_yield make_atom(755)
#define am_abs make_atom(756)
#define am_adler32 make_atom(757)
#define am_adler32_combine make_atom(758)
#define am_atom_to_list make_atom(759)
#define am_binary_to_list make_atom(760)
#define am_binary_to_term make_atom(761)
#define am_crc32 make_atom(762)
#define am_crc32_combine make_atom(763)
#define am_date make_atom(764)
#define am_delete_module make_atom(765)
#define am_display make_atom(766)
#define am_display_string make_atom(767)
#define am_element make_atom(768)
#define am_erase make_atom(769)
#define am_exit_signal make_atom(770)
#define am_external_size make_atom(771)
#define am_float make_atom(772)
#define am_float_to_list make_atom(773)
#define am_fun_info make_atom(774)
#define am_get make_atom(775)
#define am_get_keys make_atom(776)
#define am_halt make_atom(777)
#define am_phash make_atom(778)
#define am_phash2 make_atom(779)
#define am_hd make_atom(780)
#define am_integer_to_list make_atom(781)
#define am_length make_atom(782)
#define am_link make_atom(783)
#define am_list_to_atom make_atom(784)
#define am_list_to_binary make_atom(785)
#define am_list_to_float make_atom(786)
#define am_list_to_pid make_atom(787)
#define am_list_to_port make_atom(788)
#define am_list_to_ref make_atom(789)
#define am_list_to_tuple make_atom(790)
#define am_localtime make_atom(791)
#define am_localtime_to_universaltime make_atom(792)
#define am_make_ref make_atom(793)
#define am_unique_integer make_atom(794)
#define am_md5_init make_atom(795)
#define am_md5_update make_atom(796)
#define am_md5_final make_atom(797)
#define am_module_loaded make_atom(798)
#define am_function_exported make_atom(799)
#define am_monitor_node make_atom(800)
#define am_nodes make_atom(801)
#define am_now make_atom(802)
#define am_monotonic_time make_atom(803)
#define am_system_time make_atom(804)
#define am_open_port make_atom(805)
#define am_pid_to_list make_atom(806)
#define am_pre_loaded make_atom(807)
#define am_process_flag make_atom(808)
#define am_process_info make_atom(809)
#define am_put make_atom(810)
#define am_registered make_atom(811)
#define am_round make_atom(812)
#define am_self make_atom(813)
#define am_setelement make_atom(814)
#define am_spawn make_atom(815)
#define am_spawn_link make_atom(816)
#define am_split_binary make_atom(817)
#define am_statistics make_atom(818)
#define am_term_to_binary make_atom(819)
#define am_term_to_iovec make_atom(820)
#define am_time make_atom(821)
#define am_tl make_atom(822)
#define am_trunc make_atom(823)
#define am_tuple_to_list make_atom(824)
#define am_universaltime make_atom(825)
#define am_universaltime_to_localtime make_atom(826)
#define am_unlink make_atom(827)
#define am_whereis make_atom(828)
#define am_spawn_opt make_atom(829)
#define am_setnode make_atom(830)
#define am_dist_get_stat make_atom(831)
#define am_dist_ctrl_input_handler make_atom(832)
#define am_dist_ctrl_get_data make_atom(833)
#define am_dist_ctrl_get_data_notification make_atom(834)
#define am_dist_ctrl_get_opt make_atom(835)
#define am_dist_ctrl_set_opt make_atom(836)
#define am_port_info make_atom(837)
#define am_port_call make_atom(838)
#define am_port_command make_atom(839)
#define am_port_control make_atom(840)
#define am_port_close make_atom(841)
#define am_port_connect make_atom(842)
#define am_request_system_task make_atom(843)
#define am_check_process_code make_atom(844)
#define am_map_to_tuple_keys make_atom(845)
#define am_term_type make_atom(846)
#define am_map_hashmap_children make_atom(847)
#define am_time_unit make_atom(848)
#define am_perf_counter_unit make_atom(849)
#define am_is_system_process make_atom(850)
#define am_system_check make_atom(851)
#define am_dirty_process_handle_signals make_atom(852)
#define am_create_dist_channel make_atom(853)
#define am_ets_super_user make_atom(854)
#define am_dist_spawn_request make_atom(855)
#define am_no_aux_work_threads make_atom(856)
#define am_spawn_request_abandon make_atom(857)
#define am_erts_literal_area_collector make_atom(858)
#define am_release_area_switch make_atom(859)
#define am_send_copy_request make_atom(860)
#define am_port_set_data make_atom(861)
#define am_port_get_data make_atom(862)
#define am_trace_pattern make_atom(863)
#define am_trace_info make_atom(864)
#define am_trace_delivered make_atom(865)
#define am_seq_trace make_atom(866)
#define am_seq_trace_info make_atom(867)
#define am_seq_trace_print make_atom(868)
#define am_suspend_process make_atom(869)
#define am_resume_process make_atom(870)
#define am_bump_reductions make_atom(871)
#define am_math make_atom(872)
#define am_cos make_atom(873)
#define am_cosh make_atom(874)
#define am_sin make_atom(875)
#define am_sinh make_atom(876)
#define am_tan make_atom(877)
#define am_tanh make_atom(878)
#define am_acos make_atom(879)
#define am_acosh make_atom(880)
#define am_asin make_atom(881)
#define am_asinh make_atom(882)
#define am_atan make_atom(883)
#define am_atanh make_atom(884)
#define am_erf make_atom(885)
#define am_erfc make_atom(886)
#define am_exp make_atom(887)
#define am_log make_atom(888)
#define am_log2 make_atom(889)
#define am_log10 make_atom(890)
#define am_sqrt make_atom(891)
#define am_atan2 make_atom(892)
#define am_pow make_atom(893)
#define am_start_timer make_atom(894)
#define am_send_after make_atom(895)
#define am_cancel_timer make_atom(896)
#define am_read_timer make_atom(897)
#define am_make_tuple make_atom(898)
#define am_append_element make_atom(899)
#define am_system_flag make_atom(900)
#define am_system_info make_atom(901)
#define am_system_profile make_atom(902)
#define am_ref_to_list make_atom(903)
#define am_port_to_list make_atom(904)
#define am_fun_to_list make_atom(905)
#define am_is_process_alive make_atom(906)
#define am_raise make_atom(907)
#define am_is_builtin make_atom(908)
#define am__AtomAlias26 make_atom(909)
#define am_append make_atom(910)
#define am_subtract make_atom(911)
#define am_is_atom make_atom(912)
#define am_is_list make_atom(913)
#define am_is_tuple make_atom(914)
#define am_is_float make_atom(915)
#define am_is_integer make_atom(916)
#define am_is_number make_atom(917)
#define am_is_pid make_atom(918)
#define am_is_port make_atom(919)
#define am_is_reference make_atom(920)
#define am_is_binary make_atom(921)
#define am_is_function make_atom(922)
#define am_is_record make_atom(923)
#define am_match_spec_test make_atom(924)
#define am_internal_request_all make_atom(925)
#define am_delete make_atom(926)
#define am_delete_object make_atom(927)
#define am_first_lookup make_atom(928)
#define am_is_compiled_ms make_atom(929)
#define am_lookup make_atom(930)
#define am_lookup_element make_atom(931)
#define am_last make_atom(932)
#define am_last_lookup make_atom(933)
#define am_match_object make_atom(934)
#define am_member make_atom(935)
#define am_next make_atom(936)
#define am_next_lookup make_atom(937)
#define am_prev make_atom(938)
#define am_prev_lookup make_atom(939)
#define am_insert make_atom(940)
#define am_insert_new make_atom(941)
#define am_rename make_atom(942)
#define am_safe_fixtable make_atom(943)
#define am_slot make_atom(944)
#define am_update_counter make_atom(945)
#define am_select make_atom(946)
#define am_select_count make_atom(947)
#define am_select_reverse make_atom(948)
#define am_select_replace make_atom(949)
#define am_match_spec_compile make_atom(950)
#define am_match_spec_run_r make_atom(951)
#define am_os make_atom(952)
#define am_getenv make_atom(953)
#define am_putenv make_atom(954)
#define am_unsetenv make_atom(955)
#define am_getpid make_atom(956)
#define am_perf_counter make_atom(957)
#define am_erl_ddll make_atom(958)
#define am_try_load make_atom(959)
#define am_try_unload make_atom(960)
#define am_loaded_drivers make_atom(961)
#define am_format_error_int make_atom(962)
#define am_run make_atom(963)
#define am_internal_run make_atom(964)
#define am_lists make_atom(965)
#define am_reverse make_atom(966)
#define am_keymember make_atom(967)
#define am_keysearch make_atom(968)
#define am_keyfind make_atom(969)
#define am_disassemble make_atom(970)
#define am_same make_atom(971)
#define am_flat_size make_atom(972)
#define am_get_internal_state make_atom(973)
#define am_set_internal_state make_atom(974)
#define am_dist_ext_to_term make_atom(975)
#define am_instructions make_atom(976)
#define am_interpreter_size make_atom(977)
#define am_dirty make_atom(978)
#define am_lcnt_control make_atom(979)
#define am_lcnt_collect make_atom(980)
#define am_lcnt_clear make_atom(981)
#define am_hibernate make_atom(982)
#define am_warning_map make_atom(983)
#define am_get_module_info make_atom(984)
#define am_is_boolean make_atom(985)
#define am_string make_atom(986)
#define am_make_fun make_atom(987)
#define am_iolist_size make_atom(988)
#define am_iolist_to_binary make_atom(989)
#define am_list_to_existing_atom make_atom(990)
#define am_is_bitstring make_atom(991)
#define am_tuple_size make_atom(992)
#define am_byte_size make_atom(993)
#define am_bit_size make_atom(994)
#define am_list_to_bitstring make_atom(995)
#define am_bitstring_to_list make_atom(996)
#define am_update_element make_atom(997)
#define am_decode_packet make_atom(998)
#define am_characters_to_binary make_atom(999)
#define am_characters_to_list make_atom(1000)
#define am_bin_is_7bit make_atom(1001)
#define am_atom_to_binary make_atom(1002)
#define am_binary_to_atom make_atom(1003)
#define am_binary_to_existing_atom make_atom(1004)
#define am_dflag_unicode_io make_atom(1005)
#define am_give_away make_atom(1006)
#define am_setopts make_atom(1007)
#define am_load_nif make_atom(1008)
#define am_call_on_load_function make_atom(1009)
#define am_finish_after_on_load make_atom(1010)
#define am_binary_part make_atom(1011)
#define am_compile_pattern make_atom(1012)
#define am_matches make_atom(1013)
#define am_longest_common_prefix make_atom(1014)
#define am_longest_common_suffix make_atom(1015)
#define am_at make_atom(1016)
#define am_part make_atom(1017)
#define am_list_to_bin make_atom(1018)
#define am_referenced_byte_size make_atom(1019)
#define am_encode_unsigned make_atom(1020)
#define am_decode_unsigned make_atom(1021)
#define am_nif_error make_atom(1022)
#define am_prim_file make_atom(1023)
#define am_internal_name2native make_atom(1024)
#define am_internal_native2name make_atom(1025)
#define am_internal_normalize_utf8 make_atom(1026)
#define am_is_translatable make_atom(1027)
#define am_file make_atom(1028)
#define am_native_name_encoding make_atom(1029)
#define am_check_old_code make_atom(1030)
#define am_universaltime_to_posixtime make_atom(1031)
#define am_posixtime_to_universaltime make_atom(1032)
#define am_dt_put_tag make_atom(1033)
#define am_dt_get_tag make_atom(1034)
#define am_dt_get_tag_data make_atom(1035)
#define am_dt_spread_tag make_atom(1036)
#define am_dt_restore_tag make_atom(1037)
#define am_dt_prepend_vm_tag_data make_atom(1038)
#define am_dt_append_vm_tag_data make_atom(1039)
#define am_finish_loading make_atom(1040)
#define am_insert_element make_atom(1041)
#define am_delete_element make_atom(1042)
#define am_integer_to_binary make_atom(1043)
#define am_float_to_binary make_atom(1044)
#define am_binary_to_float make_atom(1045)
#define am_printable_range make_atom(1046)
#define am_inspect make_atom(1047)
#define am_is_map make_atom(1048)
#define am_map_size make_atom(1049)
#define am_maps make_atom(1050)
#define am_find make_atom(1051)
#define am_from_list make_atom(1052)
#define am_is_key make_atom(1053)
#define am_keys make_atom(1054)
#define am_merge make_atom(1055)
#define am_remove make_atom(1056)
#define am_update make_atom(1057)
#define am_values make_atom(1058)
#define am_cmp_term make_atom(1059)
#define am_take make_atom(1060)
#define am_fun_info_mfa make_atom(1061)
#define am_map_info make_atom(1062)
#define am_is_process_executing_dirty make_atom(1063)
#define am_check_dirty_process_code make_atom(1064)
#define am_purge_module make_atom(1065)
#define am_split make_atom(1066)
#define am_size_shared make_atom(1067)
#define am_copy_shared make_atom(1068)
#define am_has_prepared_code_on_load make_atom(1069)
#define am_floor make_atom(1070)
#define am_ceil make_atom(1071)
#define am_fmod make_atom(1072)
#define am_set_signal make_atom(1073)
#define am_iolist_to_iovec make_atom(1074)
#define am_get_dflags make_atom(1075)
#define am_new_connection make_atom(1076)
#define am_map_next make_atom(1077)
#define am_gather_alloc_histograms make_atom(1078)
#define am_gather_carrier_info make_atom(1079)
#define am_map_get make_atom(1080)
#define am_is_map_key make_atom(1081)
#define am_internal_delete_all make_atom(1082)
#define am_internal_select_delete make_atom(1083)
#define am_persistent_term make_atom(1084)
#define am_erase_persistent_terms make_atom(1085)
#define am_atomics_new make_atom(1086)
#define am_atomics make_atom(1087)
#define am_add make_atom(1088)
#define am_add_get make_atom(1089)
#define am_exchange make_atom(1090)
#define am_compare_exchange make_atom(1091)
#define am_counters_new make_atom(1092)
#define am_counters_get make_atom(1093)
#define am_counters_add make_atom(1094)
#define am_counters_put make_atom(1095)
#define am_counters_info make_atom(1096)
#define am_spawn_system_process make_atom(1097)
#define am_ets_lookup_binary_info make_atom(1098)
#define am_ets_raw_first make_atom(1099)
#define am_ets_raw_next make_atom(1100)
#define am_abort_pending_connection make_atom(1101)
#define am_get_creation make_atom(1102)
#define am_prepare_loading make_atom(1103)
#define am_beamfile_chunk make_atom(1104)
#define am_beamfile_module_md5 make_atom(1105)
#define am_unalias make_atom(1106)
#define am_from_keys make_atom(1107)
#define am_binary_to_integer make_atom(1108)
#define am_list_to_integer make_atom(1109)
#define am_term_to_string make_atom(1110)
#define am_coverage_support make_atom(1111)
#define am_get_coverage_mode make_atom(1112)
#define am_get_coverage make_atom(1113)
#define am_reset_coverage make_atom(1114)
#define am_set_coverage_mode make_atom(1115)
#define am_trace_session_create make_atom(1116)
#define am_trace_session_destroy make_atom(1117)
#define am_erts_trace_cleaner make_atom(1118)
#define am_check make_atom(1119)
#define am_send_trace_clean_signal make_atom(1120)
#define am_system_monitor make_atom(1121)
#define am_processes_next make_atom(1122)
#define am_get_debug_info make_atom(1123)
#define am_erl_debugger make_atom(1124)
#define am_supported make_atom(1125)
#define am_instrumentations make_atom(1126)
#define am_toggle_instrumentations make_atom(1127)
#define am_breakpoints make_atom(1128)
#define am_notify_breakpoint_hit make_atom(1129)
#define am_stack_frames make_atom(1130)
#define am_peek_stack_frame_slot make_atom(1131)
#define am_xregs_count make_atom(1132)
#define am_peek_xreg make_atom(1133)
#define am_unaligned_bitstring make_atom(1134)
#define am_import make_atom(1135)
#define am_put_new make_atom(1136)
#endif
